/*------------------------------------------------------------------------*/
/* Copyright (C) 2015 - 2016 Armin Biere, Johannes Kepler University Linz */
/*------------------------------------------------------------------------*/

// header files for 'usage'

#include <iostream>		// 'std::cout' etc.
#include <cstdio>		// for 'sprintf' in next 'OPT'
#include <climits>		// for 'INT_MAX' as default in 'opts.h'

/*------------------------------------------------------------------------*/

#define NL std::endl

static void usage () {
  std::cout
  << "usage: splatz [ <option> ... ] [ <dimacs> ]" << NL
  << "where <option> is one of the following short options:" << NL
  << NL
  << "-h                     print this command line option summary" << NL
  << "-e                     print embedded option list (for 'cnfdd')" << NL
  << "-r                     print option value ranges (for 'cnfuzz')" << NL
  << "--version              print version" << NL
  << NL                     
  << "-n                     do not print the witness 'v ...' lines" << NL
  << "-v                     increase verbosity level" << NL
#ifdef LOGGING		     // './configure.sh -l'
  << "-l                     enable very verbose logging" << NL
#endif
#ifdef EXPENSIVECHECKS       // './configure.sh -e'
  << "-c                     enable expensive checking" << NL
#endif
  << NL
  << "-t <drat-trace>        DRUP/DRAT tracing to specified file" << NL
  << "                       (default '<stdout>' with '--trace' only)" << NL
  << NL
  << "-o <output-file>       output DIMACS file" << NL
  << "                       (dumped after each simplification)" << NL
  << NL
  << "Also <option> can be one of the following long options" << NL
  << NL;
  char line[200];
#define OPT(NAME,DEF,MIN,MAX,DESCRIPTION) \
  sprintf (line, "--%-20s %s [%d]", #NAME, DESCRIPTION, DEF); \
  std::cout << line << NL;
#include "opts.h"
#undef OPT
  std::cout << NL
  << "Long options '--<long>=<val>' can be used in the following" << NL
  << "form '--<long>', using value '1', and '--no-<long>', which" << NL
  << "sets the value to '0'." << NL
  << NL
  << "Finally <dimacs> is a path to a CNF in DIMACS format.  If it" << NL
  << "has a 'gz', 'bz2' or '7z' suffix, then a pipe is opened and" << NL
  << "the file is decompressed with an appropriate helper function." << NL
  << std::flush;
}

/*------------------------------------------------------------------------*/

// Remaining standard external header files

#include <algorithm>	// 'std::sort' and 'std::swap'
#include <cassert>	// 'assert'
#include <cstdlib>	// 'malloc' etc.
#include <cstring>	// 'memset', 'strcpy'
#include <iomanip>	// 'std::setprecision'
#include <stdio.h>	// 'popen', 'fopen', 'getc'
#include <fstream>

/*------------------------------------------------------------------------*/

// More system specific C style external header files

extern "C" {
#include <signal.h>		// 'signal'
#include <sys/time.h>		// 'getrusage'
#include <sys/resource.h>	// 'getrusage'
#include <sys/types.h>		// 'stat'
#include <sys/stat.h>		// 'stat'
#include <unistd.h>		// 'stat'
#include <ctype.h>		// 'isdigit'
};

/*------------------------------------------------------------------------*/

// Online forward DRUP checking for testing, fuzzing and delta debugging.
// This simplifies debugging, since no external scripts are necessary.  In
// contrast to Lingeling this solver has its own native tracer which does
// not rely on an external library such as DRUPLIG.  But in principle the
// DRUPLIG tracer can also be used (should generate identical traces).  To
// include this code use './configure.sh --druplig', which checks for proper
// existence of the DRUPLIG code in '../druplig'.  Per default
// 'opts::druplig', and thus online checking, is enabled.  Tracing through
// DRUPLIG can be enabled by setting 'opts::drupligtrace'.

#ifdef DRUPLIG
extern "C" {
#include "../druplig/druplig.h"
};
static Druplig * druplig;
#endif

/*------------------------------------------------------------------------*/

// Here comes inclusion of our single internal header file.

#include "config.h"

/*========================================================================*/

// We start with global compile time constants.

// MSB should be zero (for clausal reasons) and (redundantly)
// should be smaller than UINT_MAX (used as invalid index in 'collect...').

static const int LDMAXCIDX = 31;

// Clause size might be stored in a bit-field of this number of bits.
// In the current version we use a simple unsigned word ('sizebits').

static const int LDMAXSIZE = 30;

// One bit, more specifically bit 'MSB-1', should be ignored (because of
// bit-stuffing in 'Rsn'). Note, variable indices and literals are signed,
// e.g., of type 'int'.  So there is one additional bit needed to store the
// sign of a literal, which in turn means that 'LDMAXVAR+1' is the actual
// width of a literal.  Further related to the comments for 'sll'.

static const int LDMAXVAR = 30;

/*========================================================================*/

#ifdef LOGGING
#define LOGENABLED opts::log
#define LOG(ARG) \
do { \
  if (!opts::log) break; \
  std::cout << "c LOG " << level << ' ' << ARG << NL << std::flush; \
} while (0)
#else
#define LOG(ARG) do { } while (0)
#define LOGENABLED 0
#endif

/*------------------------------------------------------------------------*/

#define MSG(LEVEL,NAME,COUNT,ARG) \
do { \
  if (!LOGENABLED && opts::verbose < LEVEL) break; \
  std::cout << "c [" \
       << NAME \
       << '-' \
       << COUNT \
       << "] " \
       << ARG \
       << NL << std::flush; \
} while (0)

/*------------------------------------------------------------------------*/

#define COVER(COND) \
do { \
  if (!(COND)) break; \
  std::cerr << "splatz: " \
       << __FILE__ << ": " \
       << __LINE__ << ": " \
       << __FUNCTION__ << ": " \
       << "Coverage `" \
       << #COND \
       << "' hit." << NL << std::flush; \
  abort (); \
} while (0)

/*------------------------------------------------------------------------*/

#define DIE(MSG) \
do { \
  std::cout << std::flush; \
  std::cerr << "*** splatz: " << MSG << NL << std::flush; \
  exit (1); \
} while (0)

/*------------------------------------------------------------------------*/

static bool is_power_of_two (int64_t n) { return !(n & (n-1)); }

static bool at_most_two_bits_set (int64_t n) {
  if (!n) return true;
  int64_t t = n & (n-1);
  if (!t) return true;
  return is_power_of_two (t);
}

/*------------------------------------------------------------------------*/

// Almost all heap memory allocation is done explicitly (except for things
// like internal stacks in 'sort' or heap memory used for I/O streams).
// This allows us to determine the current and maximum allocated heap
// memory.  We can also reuse 'realloc', which might be useful for
// reallocating stacks aka 'vector' resp. 'sector'.  However, we saw some
// performance regression in using these wrappers together with our
// 'sector' template. See also the comments there.

static size_t allocated;		// currently allocated
static size_t maxbytes;			// maximally allocated

/*------------------------------------------------------------------------*/

// Place clauses in a contiguous memory 'arena' to optimize access pattern
// based on the order of the occurrence of clauses in watch lists.  We
// put clauses and watcher stacks into this arena, but for the low-level
// allocation functions in the next section all pointers in the arena are
// the same.  Allocating clauses and watches gives around 5% to 8% more
// propagations per second (each!) even considering the overhead for copying
// and moving clauses in 'copy_and_move_arena'.  This trick was triggered by
// an EMail from Norbert Manthey as reaction to a blog post by Mate Soos,
// which showed that MiniSAT was doing the first (moving clauses there) all
// the time (and Mate did not do this in CryptoMiniSAT).  Many thanks to
// Norbert to share this information (just before the SAT Competition 2016
// deadline ;-).  However, I had to debug it for a day.  Anyhow, see the
// following functions 'is_in_arena' and then 'copy_and_move_arena'.

static struct { char * start, * end; } arena;

/*------------------------------------------------------------------------*/

static double average (double a, double b) { return b ? a/b : 0; }

static double percent (double a, double b) { return 100.0*average (a, b); }

static void inc_bytes (size_t bytes) {
  allocated += bytes;
  if (allocated > maxbytes) maxbytes = allocated;
}

static void dec_bytes (size_t bytes) {
  assert (allocated >= bytes);
  allocated -= bytes;
}

static void deallocate_arena () {
  size_t bytes = arena.end - arena.start;
  dec_bytes (bytes);
  free (arena.start);
  arena.end = arena.start = 0;
}

static bool is_in_arena (void * ptr) {
  return arena.start <= ptr && ptr < arena.end;
}

static void * allocate (size_t bytes) {
  if (!bytes) return 0;
  void * res = malloc (bytes);
  if (!res) DIE ("failed to allocate " << bytes << " bytes");
  inc_bytes (bytes);
  return res;
}

static void deallocate (void * ptr, size_t bytes) {
  if (is_in_arena (ptr)) return;			// Ignore arena mem.
  if (!bytes) { assert (!ptr); return; }
  dec_bytes (bytes);
  free (ptr);
}

static void *
reallocate (void * ptr, size_t old_bytes, size_t new_bytes) {
  void * res;
  if (is_in_arena (ptr)) {		// Uh, move it away from arena.
    res = allocate (new_bytes);
    size_t bytes_to_copy = std::min (old_bytes, new_bytes);
    memcpy (res, ptr, bytes_to_copy);
  } else {
    dec_bytes (old_bytes);
    res = realloc (ptr, new_bytes);
    inc_bytes (new_bytes);
    if (!res) DIE ("failed to reallocate " << new_bytes << " bytes");
  }
  return res;
}

// Allocate but do not initialize elements.
//
#define ALLOC(T,P,N) \
do { \
  assert (sizeof *(P) == sizeof (T)); \
  size_t BYTES = (N) * sizeof *(P); \
  (P) = (T *) allocate (BYTES); \
} while (0)

// Allocate and zero initialize elements.
//
#define CALLOC(T,P,N) \
do { \
  assert (sizeof *(P) == sizeof (T)); \
  size_t BYTES = (N) * sizeof *(P); \
  (P) = (T *) allocate (BYTES); \
  memset ((P), 0, BYTES); \
} while (0)

// Reallocate using 'realloc'. This assumes copy-as-is for 'T' works.
//
#define REALLOC(T,P,O,N) \
do { \
  assert (sizeof *(P) == sizeof (T)); \
  size_t OLD_BYTES = (O) * sizeof *(P); \
  size_t NEW_BYTES = (N) * sizeof *(P); \
  (P) = (T *) reallocate ((void*)(P), OLD_BYTES, NEW_BYTES); \
} while (0)

// Allocate and call default constructor to initialize each element
// using a 'placement new' which calls the default constructor of 'T'.
//
#define NEW(T,P,N) \
do { \
  ALLOC (T, P, N); \
  (P) = new (P) T[N]; \
} while (0)

// This does not call the destructor of the elements.
//
#define DEALLOC(T,P,N) \
do { \
  assert (sizeof *(P) == sizeof (T)); \
  size_t BYTES = (N) * sizeof *(P); \
  deallocate ((P), BYTES); \
} while (0)

/*------------------------------------------------------------------------*/
// We only need one type of container, actually something like
// 'std::vector'.  Default is to use our own 'home-brewn' version defined
// below.  But for performance regression testing we make it compatible to
// the 'std::vector' template class and if this configured will actually
// use the 'std::vector', which however will give inaccurate memory stats.

#ifdef VECTOR // './configure --vector' or './configure -v'

#include <vector>
#define sector std::vector
template<class T> static void release (sector<T> & s) { s = sector<T> (); }

#else

// Our own small 'vector' class, which only needs 16 bytes versus 32 bytes
// needed by 'std::vector' (on a 64-bit machine).  It further only has
// implemented those 'std::vector' functions which we use (with 'unsigned'
// 'size ()' and indices instead of 'size_t').  Using a count-based design
// versus top- and end-pointers should also be faster, since we iterate
// those stacks with indices and not with iterators anyhow.  Iterating over
// indices might arguably be slower in general but is easier to understand.
// The main reason we are using this is to be able to maintain an accurate
// count of allocated heap memory. The results of some testing seems to
// indicate that the more tight integration of the memory allocator in
// 'std::vector' does give a small advantage though, in contrast to
// our approch which goes through 'malloc', 'free' and 'realloc' even
// though sizes of the allocated blocks are known.

// After adding 'arena' based clause allocation, 'sector' became faster
// than 'std::vector', for which we can only partially use 'arena' based
// allocation, since we can not keep watcher stacks in the arena.

// WARNING: 'sector' relies on 'DEALLOC' and does not call 'T' destructors.
// So do not use this template if 'T' has a real destructor.  Similar in
// spirit, but probably much harder to see is that the default copying
// constructor of 'sector' simply copies 'as-is', which is in contrast
// to the copying constructor of 'std::vector'. The STL version does not end
// up sharing the allocated array 'a'.  In order to achieve STL semantics
// you need to explicitly use the 'copy' member function.  This issue occurs
// during initializing the 'watches' array which has 'sector' elements.  In
// that scenario the 'copy-as-is' semantics is intended. Further down,
// where 'clause' is copied to 'unshrunken', the actual copying semantics of
// 'std::vector' is meant, and we have to copy explicitly, instead of just
// using 'unshrunken = clause' as in STL semantics would suffice.

template<class T> class sector {
  T * a;
  unsigned n, s;
  void enlarge () {
    if (s == (1u<<31)) DIE ("too many elements on small stack");
    const unsigned t = s ? (s<<1) : 1;
    REALLOC (T, a, s, t);
    s = t;
  }
public:
  sector () : a (0), n (0), s (0) { }
  void release () { if (s) { DEALLOC (T, a, s); a = 0; n = s = 0; } }
  ~sector () { release (); }
  void push_back (const T & t) { if (n == s) enlarge (); a[n++] = t; }
  T & operator [] (unsigned i) { assert (i < n); return a[i]; }
  const T & operator [] (unsigned i) const { assert (i < n); return a[i]; }
  void resize (unsigned i) { assert (i <= n); n = i; }
  void clear () { resize (0); }
  bool empty () const { return !n; }
  unsigned size () const { return n; }
  unsigned capacity () const { return s; }
  const T & back () const { assert (n > 0); return a[n-1]; }
  void pop_back () { assert (n > 0); n--; }
  T * begin () { return a; }			// to support 'sort'
  T * end () { return a + n; }			// to support 'sort'

  // Function not in 'std::vector' used for 'copy_and_move_arena'.

  void copy_and_move_to (void * p) {
    memcpy (p, a, n * sizeof (T));
    DEALLOC (T, a, s);
    a = (T*) p;
  }
};

template<class T> static void release (sector<T> & s) { s.release (); }

#endif // of '#else' case of '#ifdef VECTOR'

/*------------------------------------------------------------------------*/

template<class S, class T>
static void remove (S & s, const T & t) {
  unsigned j = 0;
  for (unsigned i = 0; i < s.size (); i++) {
    T o = s[i];
    if (o != t) s[j++] = o;
  }
  s.resize (j);
}

template<class S, class T>
static void copy (S & dst, const T & src) {
  dst.clear ();
  for (unsigned i = 0; i < src.size (); i++)
    dst.push_back (src[i]);
}

template<class S>
static void reverse (S & s, unsigned l = 0) {
  assert (l <= s.size ());
  if (s.empty ()) return;
  unsigned r = s.size () - 1;
  while (l < r) std::swap (s[l++], s[r--]);
}

/*------------------------------------------------------------------------*/

// This fixed size vector class is used for static allocation of 'profiling'
// data. This supports printing (and sorting) profiling data even after all
// heap memory has been reclaimed for leak checking.

template<class T, unsigned s> class fector {
  T a[s];
  unsigned n;
public:
  fector () : n (0) { }
  void push_back (const T & t) { assert (n < s); a[n++] = t; }
  T & operator [] (unsigned i) { assert (i < n); return a[i < s ? i : 0]; }
  bool empty () const { return !n; }
  unsigned size () const { return n; }
  const T & back () const { assert (n > 0); return a[n-1]; }
  void pop_back () { assert (n > 0); n--; }
};

/*------------------------------------------------------------------------*/

struct Cls;		// forward declaration, e.g., used in 'Rsn'

// The reason for assigning a variable.  In essence this could be either a
// decision, a binary clause or a larger non-binary clause.  For top-level
// assigned variables, either due to a top-level unit, due to substituting
// or eliminating a variable, the reason becomes invalid.  In this case we
// use a decision reason as default, which also is used for unassigned
// variables.  Thus the fixed embedded reasons in the 'Var' struct are only
// valid for assigned active varaibles.

class Rsn {

  unsigned word;	// data, assume 'sizeof (word) == 4', e.g., 32 bits

public:

  // decision reason
  //
  Rsn () : word (0) { }
  bool decision () const { return !word; }

  // binary reason, e.g., a single antecedent literal (LSB 0)
  //
  Rsn (int l) : word (l) {		// see also 'sll' comments below
    assert (l);
    assert (l != INT_MIN);
    assert (abs (l) < (1<<LDMAXVAR));
    assert (!(word & (1u<<31)) == !(word & (1u<<30))); // bit 30 == bit 31
    word <<= 1;
  }
  int lit () const { assert (binary ()); return ((int)word) >> 1; }
  bool binary () const { return word && !(word & 1u); }

  // actual clause antecedent as reason (LSB 1)
  //
  Rsn (Cls * c, unsigned cidx);		// see below, needs 'clauses'
  Cls * cls () const;			// see below, needs 'clauses'
  bool clausal () const { return word & 1u; }	// !decision && !binary

  unsigned cidx () const { assert (clausal ()); return word>>1; }
  void update (unsigned cidx) { word = (cidx << 1) | 1; }
};

struct Unit : public Rsn { };
struct Decision : public Rsn { };
struct Substituted : public Rsn { };
struct Eliminated : public Rsn { };

/*------------------------------------------------------------------------*/

struct Lnk {		// support for VMTF, e.g., see our SAT'15 paper
  unsigned stamp;	// decision queue enqueued stamp
  int prev;		// decision queue previous link
  int next;		// decision queue next link
  Lnk () : stamp (0), prev (0), next (0) { }
};

// While searching for the first UIP clause, we mark variables resolved into
// the resolvent, which eventually will give the learned clause.  Those are
// marked with 'SEEN'.  The DFS for clause minimization tries to remove some
// of those original literals and marks those that canot be removed
// (recursively) as 'POISON'.  See the description of 'minimize' below.
// Otherwise we also use '0' and '1' for the corresponding tag field of a
// variable for instancing during touching variables in 'elim' and 'block'
// (through 'touch_clause').  By default the tag is zero, which means not
// seen or not touched depending on the context.  We use a bit-field of two
// bits below to save some space.  Thus the symbol 'SeenTag' should be
// considered to be a comment.  We also do not use 'NOTSEEN' explicitly.

enum SeenTag {
  NOTSEEN = 0,	// just a comment not really used
  SEEN = 1,	// see 'first_uip' and 'minimize' for more details
  POISON = 2,	// for 'SEEN' and 'POISON'
};

// Variables are either active, in which case they take part in search, or
// inactive, because they either became a top-level unit are eliminated
// during variable-elimination in 'elim' or substituted by equivalent
// literal substitution in 'decompose'.  In those cases they are fixed to a
// constant value but marked as inactive.  Their actual value will be
// determined by 'extend_model' by going over the 'extend' stack.  As with
// 'SeenTag' the symbol 'VarStatus' is not used and just a comment, which
// also applies to 'ACTIVE'.

enum VarStatus {
  ACTIVE = 0,
  FIXED = 1,
  ELIMINATED = 2,
  SUBSTITUTED = 3
};

struct Var {
  unsigned level;
  unsigned tag:2;	// SeenTag or just 0/1
  unsigned status:2;	// StatusTag
  unsigned pbb:2;	// potential backbone (see 'sweep_backbone')
  unsigned lca:1;	// least-common-ancestor mark bit 
  			// (see 'on_the_fly_subsumption')
  Rsn reason;
  Var () : 
    level (0), tag (0), status (0), pbb (3), lca (0),
    reason (Decision ())
  { }
  bool decision () const { return reason.decision (); }
};

/*------------------------------------------------------------------------*/

static const unsigned CLSLITS = 2;	// Used for 'Cls.lits]' below.

struct Cls {
  unsigned redundant:1;	// learned so not irredundant (original)
  unsigned garbage:1;	// collect during 'reduce_move' or 'flush_clauses'
  unsigned traced:1;	// deletion already traced
  unsigned nontriv:1;	// has a non trivial resolvent (in 'bce_var')
  unsigned blocked:1;	// left part of blocked clause decomposition
  unsigned notblocked:1;// right part of blocked clause decomposition
  unsigned watched:1;	// for 'watch_bcd_clauses'

  unsigned sizebits;	// hide because it may be (it was) a bit-field
   			// and should be different from 'size' anyhow

  int lits[CLSLITS];	// Actually of 'sizebits' literals and thus 'embeds'
  			// the literals directly into the same memory
			// (hopefully cache line) which avoids another
			// memory access every time a clause is visited.

  // Provide partial support for some 'std::vector' style functions, which
  // are used for template based generic access to literals on a
  // 'sector<int>' stack or a clause of type 'Cls &', e.g., 'trace_...' or
  // 'remove', 'copy', 'reverse' above.

  unsigned size () const { return sizebits; }

  size_t bytes () const;

  const int & operator [] (unsigned i) const {
    assert (i < sizebits); return lits[i];
  }
};

// Redundant clauses need some more data, mainly used to determine whether a
// clause is considered to less useful and remove it during 'reduce'.  To
// avoid cluttering irredundant and original clauses with this data, which
// takes up space, we simply prefix redundant clause with this data.  See
// also 'cls2red'.

struct Red {
  unsigned resolved;	// Resolved last at this conflict (index).
  unsigned active:1;	// Resolved since last 'reduce'.
  unsigned glue:31;	// Also called LBD or glucose level as in the
  			// the seminal IJCAI'09 paper by
			// Gilles Audemard and Laurent Simon.
  Cls cls;		// The actual common clause data.
};

// So the actual memory layout of a clause looks like this:
//         ................
//         : resolved     : Red  (only exists if 'redundant' is true)
//         : active       :
// actual  : glue         :
//  -----> +--------------+
// clause  | redundant    | Cls starts here (always exists)
// pointer | garbage      | 
//         | ...          | (other bits)
//         | ...          |
//         | size         | (assumption 'size >= 2')
//         | lits[0]      |
//         | ...          | ('size' embedded literals)
//         | lits[size-1] |
//         +--------------+

/*------------------------------------------------------------------------*/

struct Frame {		// Frames on control stack, one per decision level.
  int decision;		// The actual decision literal of this level.
  unsigned pulled;	// Num lits seen / resolved with this level.
  Frame (int d) : decision (d), pulled (0) { }
};

/*------------------------------------------------------------------------*/

struct Queue {
  int first;		// First variable in decision queue.
  int last;		// Last variable in decision queue.
  int searched;		// Last searched variable in decision queue.
  unsigned stamp;	// Enqueued time stamp (see SAT'15 paper).
  Queue () : first (0), last (0), searched (0), stamp (0) { }

  void restamp ();

  void enqueue (int idx);
  void dequeue (int idx);
};

/*------------------------------------------------------------------------*/

// Exponential moving averages.  See our 'POS'15' paper for more details.

class EMA {
  int64_t v;		// Fixed point with 32 bit fractional part.
  unsigned short s;	// Actual smoothing factor 2^(-s).
  unsigned short t;	// Current factor until initialized.
  unsigned d;		// Delay until next 't' increase.
public:

  void init (int i) {
    assert (0 <= i), assert (i < 32), s = i;
    v = t = d = 0;
  }

  int64_t val () const { return v; }

  int nval () const { return (v+(1ll<<31)) >> 32; }	// round

  // Our representation for 64-bit EMAs are fixed point numbers with 32-bit
  // fractional and non-fractional part.  This function turns such a fixed
  // point number into 'double'.  This is only used for messages.

  double norm () const { return v / (double)(((int64_t)1)<<32); }

  // Interpret an exponential moving average 'ema' as a prediction, then
  // updating means to adjust it by a fraction 'f' of the error 'val - ema'
  // between the new value 'val' and the current prediction 'ema':
  //
  //   ema += (val - ema) * f
  //
  // And this would be the actual code for a (not deterministically
  // portable) floating point implementation.  Instead we use a 64 bit fixed
  // point representation with a fractional part of 32 bits, as in Donald
  // Knuth's implementation of our 'agility' measure (see our SAT'08
  // paper and Don's TAOCP SAT chapter on 'agility').
  //
  // For a fixed point implementation we restrict 'f' to '2^{-shift}', which
  // gives the following code.  See our POS'15 paper for more details too.

  void update (int y) {

    int shift = t;

    // This initialization should increment 't' until it reaches 's'.
    // For each 't' we wait for '2^t' updates before incrementing it again
    // The idea is that the resulting 'shift' follows the sequence:
    //
    //  0,1,1,2,2,2,2,4,4,4,4,4,4,4,4,8,8, ..., s, s, ... (forever 's')
    //
    // which needs '2^(s)-1' updates before 's' is reached.  We did not
    // derive this formally, but observed it during logging.  This
    // is also not published yet, e.g., was not in POS'15.

    if (t < s && !d--) d = (1u << ++t) - 1;

    assert (0 <= shift), assert (shift <= 32);
    v -= v >> shift;
    // (1) Beware of left shift of 'signed int' (see comment to 'sll').
    // (2) Only gcc with c11++ support enabled also supports 'uint64_t'.
    v += ((unsigned long long)(int64_t)y) << (32 - shift);
  }

  // We often compare EMAs against certain other metrics or counters and
  // based on the outcome of this comparison trigger certain events.  Often
  // there is a margin value included, which makes sure that the other metric
  // is really well above or below the EMA.  This margin values are given in
  // percent of the EMA, where '100%' means no margin, '120%', means '20%'
  // above the EMA, and '80%' means '20%' below the EMA.

  int cmpi64 (int64_t m, int margin) const {
    assert (0 < margin);
    int64_t a = v/100;
    int64_t b = m/margin;
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
  }

  int cmpema (const EMA & other, int margin) const {
    return cmpi64 (other.val (), margin);
  }

  int cmpi32 (int m, int margin) const {
    return cmpi64 ((int64_t)(((unsigned int64_t)m)<<32), margin);
  }

  int cmpu32 (unsigned m, int margin) const {
    assert (m < (unsigned) INT_MAX);
    return cmpi32 ((int) m, margin);
  }
};

/*------------------------------------------------------------------------*/

class RNG {			// Random Number Generator
  unsigned z, w;
public:
  void seed (unsigned s) { w = s; z = ~w; }
  RNG () { seed (0); }
  unsigned rand () {
    z = 36969 * (z & 65535) + (z >> 16);
    w = 18000 * (w & 65535) + (w >> 16);
    unsigned res = z + w;
    return res;
  }
};

/*------------------------------------------------------------------------*/

typedef signed char Val;	// -1=false,0=unassigned,1=true

/*------------------------------------------------------------------------*/

// We merge blockling literals for both binary and non binary clauses with
// the clause reference (clause indices / offset into 'clauses') in one
// contiguous stack. This restricts the number of clauses (and literals) to
// 2^31, which is not a severe restriction I think.  In contrast to
// Lingeling no ternary clauses are store in these watcher stacks nor do we
// include information about redundancy or glue of a class.  We consider the
// latter actually now to be a sub-optimal design decision for Lingeling.
// More on such design trade-offs can be found in the slides of our POS'14
// presentation.

typedef sector<unsigned> Wchs;	// watched literal stacks

/*------------------------------------------------------------------------*/

static sector<unsigned> * bcdwatches;	// Watches for blocked clauses.

struct Blocked {
  int blit;				// The blocking literal.
  unsigned cidx;
  Blocked (int b, unsigned c) : blit (b), cidx (c) { }
};

typedef sector<Blocked> BlockedSet;

/*------------------------------------------------------------------------*/

typedef int64_t BPV;		// 64-bit parallel variant of 'Val'

static const BPV bpfalse = 0;
static const BPV bptrue = ~bpfalse;

static BPV * bpvals;		// similar to 'vals' - used in 'sweep'

/*------------------------------------------------------------------------*/

// Partition with literals for sweeping.  Each variable occurs only in one
// partition, so this is in essence a partition of signed variables.

static const unsigned PARTLITS = 2;	// as 'CLSLITS' for 'Cls'

struct Part {
  unsigned size:31;
  unsigned garbage:1;
  int lits[PARTLITS];		// actually of size 'size', see 'Cls' too.
};

/*------------------------------------------------------------------------*/

static const char * output;	// name of output file

/*------------------------------------------------------------------------*/

namespace opts {

// Define all options statically.

#define OPT(NAME,DEF,MIN,MAX,DESCRIPTION) \
static int NAME;
#include "opts.h"
#undef OPT

// Initialize them to their default value.  Force to use an explicit
// initialization function to always check 'opts.h' ranges.  This speeds-up
// finding bugs due to incorrect ranges in 'opts.h'.

static void init () {
#define OPT(NAME,DEF,MIN,MAX,DESCRIPTION) \
   assert (MIN <= DEF), assert (DEF <= MAX); /* check 'opts.h' ranges */ \
   NAME = DEF;
#include "opts.h"
#undef OPT
}

}; // end of 'namespace opts'

// TODO: Using a namespace for statically allocated options was first pretty
// handy (in order to overlay option names with other symbols), but now
// makes it hard to make the library reentrant (without static variables).
// The same applies to 'profs' below.  So it would be good to turn this into
// a statically allocated struct/class, which then would simplify to make
// this library reentrant.

/*------------------------------------------------------------------------*/

struct Prof {		// profiling data
  const char * name;	// should match profiled function if it exists
  double time;		// accumulated time
  double start;		// entry time for computing updates to 'time'

  Prof (const char * n) : name (n), time (0), start (0) { }
};

namespace profs {		// profiling without external tools

#define PROF(NAME,LEVEL) \
static Prof NAME(#NAME);	// static allocation for fast access
#include "profs.h"
#undef PROF

#define PROF(NAME,LEVEL) 1+
static const int NP =	// static const counter of profs
#include "profs.h"
0;
#undef PROF

static fector<Prof*,NP> active;	// stack of currently profiled functions
static fector<Prof*,NP> table;	// table of all enabled profiled functions

static void init () {
#define PROF(NAME,LEVEL) \
  if (LEVEL > opts::profile)	/* check whether to profile function */ \
    NAME.name = 0;		/* thus disabled */ \
  else table.push_back (&NAME);	// store for sorting in 'profile ()'
#include "profs.h"		// to avoid additional allocation later
#undef  PROF
}

}; // end of 'namespace profs'

/*------------------------------------------------------------------------*/

static int maxvar;		// maximum variable index
static int maxcls;		// maximum original number clauses
static int maxlit;		// signed literal encoded as positive int
static int eliminated;		// number of eliminated variables
static int substituted;		// number of substituted variables
static int fixed;		// number of fixed root level assigned vars

/*------------------------------------------------------------------------*/

static Var * vars;
static Val * vals;		// assignment -1=false,0=unassigned,1=true
static Val * marks;		// mark flags used similar to 'vals'
static Val * phases;		// previous assignment (old copy of 'vals')

/*------------------------------------------------------------------------*/

static sector<Cls*> clauses;		// clause index to clause (map)
static sector<unsigned> * watches;	// maps literals to watches

/*------------------------------------------------------------------------*/

static Lnk * lnks;		// decision queue links

static Queue queue;		// the actual decision queue header

/*------------------------------------------------------------------------*/

static bool inconsistent;	// root level empty clause found/learned
static Cls * conflict;		// partial assignment inconsistent
static Cls binary_conflict;	// Static binary conflicting clause.
static unsigned conflictcidx;	// Need this for double OTFS.

/*------------------------------------------------------------------------*/

static unsigned level;		// decision level, e.g. 'control.size ()'
static sector<Frame> control;	// decisions and 'pulled' literals
static sector<int> trail;	// assigned variable stack
static unsigned next;		// propagation BFS index (in 'trail')
static unsigned active;		// number of remaining active variables

/*------------------------------------------------------------------------*/

static sector<int> tagged;	// tagged during first UIP search & minimize
static sector<int> lcas;	// tagged least-common-ancestors
static sector<unsigned> levels;	// pulled levels during analyze
static sector<unsigned> work;	// keep a global working stack/map around
static sector<unsigned> resched;// keep a global rescheduling stack around
static sector<int> probes;	// for failed literal probing
static sector<Part*> parts;	// partition classes in sweep
static sector<int> part;	// one partition class
static unsigned partlo, parthi;	// working queue in 'parts'
static int nextbb;		// next back bone to try

/*------------------------------------------------------------------------*/

// Statistics section.

static int64_t additionalbins;	// additional binary clauses
static int64_t additionalunits;	// additional root level learned units
static int64_t backbonechecks;	// number of backbone checks during 'sweep'
static int64_t backbones;	// found backbones during 'sweep'
static int64_t bcdresolutions;	// during blocked clause decomposition
static int64_t blocked;		// eliminated blocked clauses
static int64_t blockresolutions;// during blocked clause elimination
static int64_t blocks;		// blocked clause eliminations
static int64_t blockvars;	// calls to 'block_var'
static int64_t conflicts;	// number of conflicts
static int64_t copymovedarenas;	// copye and moved arenas
static int64_t decisions;	// number of decisions
static int64_t decompositions;	// into strongly connected components
static int64_t decos;		// decision only learned clauses
static int64_t eliminations;	// bounded variable eliminations (BVE)
static int64_t elimresolutions;	// during variable elimination
static int64_t equivchecks;	// number of equiv checks in 'sweep'
static int64_t equivs;		// successfull equiv checks in 'sweep'
static int64_t failedlitprobes;	// failed literal probes
static int64_t failedlits;	// failed literals during 'probe'
static int64_t flushedclauses;	// number of collected clauses during GC
static int64_t flushedlits;	// number of flushed literals during GC
static int64_t gcs;		// number of garbage collections (GC)
static int64_t getimes;		// sanity check for profiling overhead
static int64_t glueupdates;	// updated glue values
static int64_t hbrs;		// hyper binary resolutions (on level 1)
static int64_t hbrsubs;		// self-subsuming  HBRs
static int64_t implieschecks;	// number of implies checks in 'sweep'
static int64_t implies;		// successfull implies checks in 'sweep'
static int64_t iterations;	// learned units
static int64_t learnedbins;	// learned binary clauses
static int64_t learnedlits;	// actual learned literals
static int64_t liftedlits;	// lifted literals during 'probe'
static int64_t makeirrs;	// redundant clauses turned into irredundant
static int64_t minimizedlits;	// minimized away literals
static int64_t moved;		// moved blocked clauses
static int64_t notreusedlevels;	// how many not reused trail levels
static int64_t otfsconflicts;	// number on-the-fly subsumption conflicts
static int64_t otfsdrives;	// number on-the-fly subsumption driving
static int64_t outputs;		// calls to 'dump_to_output'
static int64_t probings;	// number of calls to 'probe'
static int64_t propagations;	// propagated assignments
static int64_t reduced;		// reduced (collected) learned clauses
static int64_t reductions;	// calls to 'reduce'
static int64_t reports;		// calls to 'report'
static int64_t restartsblocked;	// blocked skipped restarts
static int64_t restarts;	// calls to 'restart'
static int64_t restartsdelayed;	// blocked skipped restarts
static int64_t restartskipped;	// skipped = noforce + reused
static int64_t restartsnoforce;	// restarts skipped due not being forced
static int64_t restartsreused;	// restarts skipped due to complete reuse
static int64_t reusedlevels;	// how many reuse trail levels
static int64_t reusedtrails;	// how often trail was reused (partially)
static int64_t searches;	// for next decision in decision queue
static int64_t searchreds;	// reductions during search
static int64_t shrunken;	// shrunken clauses in 'subsume'
static int64_t simpconflicts;	// conflicts during simplification 
static int64_t simplifications;	// calls to 'simplify'
static int64_t simpreds;	// reductions during simplifications
static int64_t subrecent;	// eagerly subsumed recent learned clauses
static int64_t subsumechecks;	// number of subsumption checks in 'subsume'
static int64_t subsumedirr;	// irredundant subsumed clauses in 'subsume'
static int64_t subsumed;	// number of subsumed clauses in 'subsume'
static int64_t subsumedred;	// redundant subsumed clauses in 'subsume'
static int64_t subsumes;	// number of calls to 'subsume'
static int64_t sweeps;		// blocked clause decomposition sweeps

/*------------------------------------------------------------------------*/

static unsigned resolved;	// 32-bit abstraction of 'conflicts'
static bool iterating;		// force delay of 'i' report
static bool simpmode;		// set to true during 'simplify'
static int simpreason;		// 0=conflicts,1=iterations,2=learnedbins

// Remember 'simpmode' at last 'report' to trigger printing header.

static bool lastreportsimpmode;

// Partially save search space during simplification.

static struct {
  Val * phases;
  sector<int> queue;
} saved;

/*------------------------------------------------------------------------*/

static sector<int> units;	// found units in 'decompose' and 'elim'
static sector<int> bins;	// found bins in 'deduce_additional_bin...'

typedef int64_t Sig;		// signatures or bloom filters, see SAT'05

static Sig * clsigs;		// clause signatures
static unsigned szclsigs;	// size of clause signatures

int64_t lastelimres;	// last elim resolutions before elimination
int64_t lastprops;	// last propagations after simplification

/*------------------------------------------------------------------------*/

static struct {
  struct {
    int64_t restart;		// lower bound on next restart
    int64_t reduce;		// next reduce  (upper bound)
  } conflicts;
  struct {
    int64_t conflicts;
    int64_t iterations;
    int64_t learnedbins;
  } simplify;
  unsigned queuestamp;	// queue stamp limit for fuzzing and testing only
  unsigned resolved;	// resolved limit for fuzzing and testing only
  struct {
    int gc;		// number of roots (top level units) after last gc
  } fixed;
  struct {
    int64_t conflicts;
    int64_t reductions;
  } subsume;		// next subsume all clauses limits
} limits;

// Increments for (some) limits.

static struct {
  struct {
    unsigned reduce;		// current reduce interval
    unsigned subsume;		// current subsume all interval
  } conflicts;
  struct {
    int64_t conflicts;
    int64_t learnedbins;
    int64_t iterations;
  } simplify;
} inc;

class Delay {
  int cnt;				// how long to wait
  int pen;				// logarithm current interval
public:
  Delay () : cnt (0), pen (0) { }
  void update (bool success) {
    if (success) cnt = pen = 0;
    else {
      cnt = (1 << pen);
      if (pen < opts::maxpen) pen++;
    }
  }
  bool delayed () {
    if (!cnt) return false;
    cnt--;
    return true;
  }
};

static struct {
  Delay sweep, probe, elim, subsume, block, sls;
} delays;

// Only run inpreprocessors if they are triggered, e.g., something changed.

static struct {
  bool subsume;
  bool decompose;
  bool elim;
  bool block;
  bool probe;
} trigger;

// Exponential moving averages (see POS'15 paper).

static struct {

  // The following three capture averages at the point a conflict occurs:

  struct {
    EMA fast;	// fast moving glue/trail average (opts::...fastema)
    EMA slow;	// slow moving glue/trail average (opts::...slowema)
  } glue, trail;
  EMA jump;	// back jump level average (opts::jumpema)

  // These two are updated during 'bumping' a clause in 'analyze':

  struct {
    EMA glue;	// glue resolved clauses (opts::resolvedgluema)
    EMA size;	// size resolved clauses (opts::resolvedsizema)
  } resolved;	// both are used in 'reduce'
} ema;

/*------------------------------------------------------------------------*/

static sector<int> clause;	// Partial to be added clause.
static sector<int> unshrunken;	// Unshrunken copy of 'clause'.
static unsigned numirr;		// Current number of irredundant clauses.
static unsigned numred;		// Current number of redundant clauses.

/*------------------------------------------------------------------------*/

static RNG rng;

/*------------------------------------------------------------------------*/

static sector<int> extend;	// Extension stack to fix models.

#ifndef NDEBUG
static sector<int> orig;	// Original formula for model checking.
#endif

/*------------------------------------------------------------------------*/

// These are system / OS specific, but I think still Posix compliant.

static double get_process_time () {
  getimes++;
  struct rusage u;
  double res;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

static bool file_exists (const char * path) {
  struct stat buf;
  return !stat (path, &buf);
}

static size_t file_bytes (const char * path) {
  struct stat buf;
  if (stat (path, &buf)) return 0;
  return (size_t) buf.st_size;
}

static void rename_file (const char * old_path, const char * new_path) {
  (void) rename (old_path, new_path);
  // TODO some verbose and/or error messages here
}

static void delete_file (const char * path) {
  (void) unlink (path);
  // TODO some verbose and/or error messages here
}

/*------------------------------------------------------------------------*/

static double bytes2mb (size_t bytes) { return bytes / (double)(1<<20); }
static double space () { return bytes2mb (allocated); }
static double maxspace () { return bytes2mb (maxbytes); }

/*------------------------------------------------------------------------*/

static void
setoptaux (int & opt, const char * name, int val, int MIN, int MAX) {
  assert (MIN <= MAX);
  if (val < MIN) {
    std::cout << "c forcing '--" << name << "=" << val << "'"
         << " to minimum value  '--" << name << '=' << MIN << "'"
         << NL;
    val = MIN;
  } else if (val > MAX) {
    std::cout << "c forcing '--" << name << "=" << val << "'"
         << " to maximum value '--" << name << '=' << MAX << "'"
         << NL;
    val = MAX;
  }
  if (opt != val) {
    std::cout << "c setting option --" << name << "=" << val << NL;
    opt = val;
  } else {
    std::cout << "c keeping option --" << name << "=" << val << NL;
  }
}

static bool setopt (const char * name, int val, bool set = true) {
#define OPT(NAME,DEF,MIN,MAX,DESCRIPTION) \
  if (!strcmp (name, #NAME)) { \
    if (set) setoptaux (opts::NAME, #NAME, val, MIN, MAX); \
    return true; \
  }
#include "opts.h"
#undef OPT
  return false;
}

static bool parseoptaux (char * arg, bool set) {
  int l = strlen (arg);
  if (arg[0] != '-' || arg[1] != '-') return false;
  if (arg[2] == 'n' && arg[3] == 'o' && arg[4] == '-') { // --no-<option>
    LOG ("parsed long option '" << arg << "' (value 0)");
    return setopt (arg + 5, 0, set);
  }
  int i;
  for (i = 2; i < l; i++) if (arg[i] == '=') break;
  if (i == 2) return false;	// arg = '--'
  if (i == l - 1) return false;	// arg = '--..='
  if (i == l) {						// --<option>
    LOG ("parsed long option '" << arg << "' (value 1)");
    return setopt (arg + 2, 1, set);
  }
  assert (arg[i] == '=');		// terminate 'name' with '0'
  arg[i] = 0;				// that is why 'arg' canot be const
  const char * name = arg + 2;
  const char * str = arg + i + 1;
  const char * p = str;
  if (*p == '-') p++;
  if (!isdigit (*p)) return false;	// require at least one digit
  p++;
  while (isdigit (*p)) p++;		// rest should be digits too
  if (*p) return false;
  int val = atoi (str);					// --<option>=<val>
  LOG ("parsed long option '--" << name << '=' << val << "'");
  return setopt (name, val, set);
}

static bool parseopt (const char * arg, bool set = true) {
  char * work = strdup (arg);
  bool res = parseoptaux (work, true);
  free (work);
  return res;
}

static bool hasopt (const char * arg) { return parseopt (arg, false); }

/*------------------------------------------------------------------------*/

// Support fuzzing and testing with '-e' and '-r' options.

static bool hasprefix (const char * str, const char * prefix) {
  const char * p, * q = prefix;
  for (p = str; *q && *q == *p; p++, q++)
    ;
  return !*q;
}

static bool ignore_opt (const char * name) {
  if (hasprefix (name, "druplig")) return true;
  if (!strcmp (name, "profile")) return true;
  if (!strcmp (name, "trace")) return true;
  if (!strcmp (name, "verbose")) return true;
  return false;
}

static void embedded_opts () {				// '-e'
#define OPT(NAME,DEF,MIN,MAX,DESCRIPTION) \
do { \
  if (ignore_opt (#NAME)) break; \
  std::cout << "c --" << #NAME << '=' << DEF << NL; \
} while (0);
#include "opts.h"
#undef OPT
  std::cout << std::flush;
}

static void range_opts () {				// '-r'
#define OPT(NAME,DEF,MIN,MAX,DESCRIPTION) \
do { \
  if (ignore_opt (#NAME)) break; \
  std::cout << #NAME << ' ' << DEF << ' ' << MIN << ' ' << MAX << NL; \
} while (0);
#include "opts.h"
#undef OPT
  std::cout << std::flush;
}

/*------------------------------------------------------------------------*/

// Light-weight built in profiling support.  The profiling level determines
// for which functions profiling is enabled.  The default is so low such
// that the number of times 'get_process_time' is called is small enough
// that it does not cost much.  For higher levels of 'opts::profile' more
// calls to 'get_process_time' are required, which has a Heisenberg effect
// in the sense that it influences the overall running time considerably,
// which in turn has the risk to make some of the measurements (percentage
// of time spent in a certain function) inaccurate.

static void start (Prof & p) {
  if (!p.name) return;				// disabled
  profs::active.push_back (&p);
  p.start = get_process_time ();
}

static double update_time (Prof * p, double t) {
  assert (p->name);
  double delta = t - p->start;
  p->time += delta;
  return delta;
}

static void stop (Prof * expected) {
  if (!expected->name) return;			// disabled
  assert (!profs::active.empty ());
  Prof * p = profs::active.back ();
  assert (p == expected);
  profs::active.pop_back ();
  (void) update_time (p, get_process_time ());
}

static double flush_and_update_all_active_profs () {
  double t = get_process_time ();
  while (!profs::active.empty ()) {
    Prof * p = profs::active.back ();
    profs::active.pop_back ();
    update_time (p, t);
  }
  return t;
}

#define START(PROF) start (profs::PROF)
#define STOP(PROF) stop (&profs::PROF)

/*------------------------------------------------------------------------*/

// Support for printing a clause.  Using templates, as for 'release', 'copy'
// etc. did not work, so here are some explicit functions.

static inline
std::ostream & operator << (std::ostream & o, const Cls & c) {
  for (unsigned i = 0; i < c.size (); i++) {
    if (i) o << ' ';
    o << c[i];
  }
  return o;
}

static inline
std::ostream & operator << (std::ostream & o, const sector<int> & c) {
  for (unsigned i = 0; i < c.size (); i++) {
    if (i) o << ' ';
    o << c[i];
  }
  return o;
}

#ifdef LOGGING

static inline
std::ostream & operator << (std::ostream & o, const Part & p) {
  for (unsigned i = 0; i < p.size; i++) {
    if (i) o << ' ';
    o << p.lits[i];
  }
  return o;
}

static inline
std::ostream & operator << (std::ostream & o, const Rsn & r) {
  if (r.decision ()) return o << "decision";
  else if (r.binary ()) return o << "binary reason " << r.lit () << " ...";
  else return o << "non-binary reason " << *r.cls ();
}

#endif

/*------------------------------------------------------------------------*/

// During the derivation of the 'first UIP' clause, particularly for
// 'on_the_fly_subsumption', and then for 'minimize' it is useful to show
// the current resolvent while logging is enabled.  Otherwise the actual
// resolvent is not needed.

#ifdef LOGGING
static sector<int> resolvent;
#endif
#ifdef LOGGING
#define RESOLVE(CODE) \
  do { CODE; } while (0)
static void log_initial_resolvent () {
  LOG ("initial resolvent " << resolvent);
}
static void update_and_log_resolvent (int lit) {
  remove (resolvent, -lit);
  LOG ("resolvent " << resolvent);
}
#else
#define RESOLVE(CODE) do { } while (0)
#define log_initial_resolvent() do { } while (0)
#define update_and_log_resolvent(LIT) do { } while (0)
#endif

/*------------------------------------------------------------------------*/

// This whole section is about low-level memory management of clauses.  Note
// that the whole 'arena' idea is hidden in the even lower-level actual
// memory handling functions implemented above.

static void
init_clause_header (Cls * c, bool redundant, unsigned size, unsigned glue) {
  assert (size > 1);
  memset (c, 0, sizeof (Cls) - CLSLITS * sizeof (int));
  c->redundant = redundant;
  assert (size < (1u<<LDMAXSIZE));
  c->sizebits = size;
  assert (c->size () == size);
}

static void init_clause_lits (Cls * c, unsigned size) {
  assert (clause.size () == size);
  for (unsigned i = 0; i < size; i++) c->lits[i] = clause[i];
}

static void
init_clause (Cls * c, bool redundant, unsigned size, unsigned glue = 0) {
  init_clause_header (c, redundant, size, glue);
  init_clause_lits (c, size);
  if (redundant) numred++; else numirr++;
}

static void init_redundant_clause (Cls * c, unsigned size, unsigned glue) {
  init_clause (c, true, size, glue);
}

static void init_irredundant_clause (Cls * c, unsigned size) {
  init_clause (c, false, size);
}

static Red * cls2red (Cls * c) {	// See the 'Red' vs 'Cls' layout.
  Red * r;
  assert (c->redundant);
  char * p = (char *) c;

  // This is dangerous because of alignment but 'clang --analyze'
  // did not like our previous version, which was as follows:
  //
  //  (ptrdiff_t)&((Red*)0)->cls'
  //
  p -= sizeof (Red) - sizeof (Cls);

  r = (Red*) p;
  assert (&r->cls == c);
  return r;
}

static size_t bytes_clause (bool redundant, unsigned size) {
  assert (size >= CLSLITS);
  size_t res = (size - CLSLITS) * sizeof (int);
  if (redundant) res += sizeof (Red);
  else res += sizeof (Cls);
  return res;
}

inline size_t Cls::bytes () const {
  return bytes_clause (redundant, size ());
}

static Cls * allocate_redundant_clause (unsigned glue) {
  unsigned size = clause.size ();
  size_t bytes = bytes_clause (true, size);
  Red * r = (Red*) allocate (bytes);
  assert (glue < (1u<<31));		// 'r->glue' bit-field of 31 bits
  r->glue = glue;
  assert (resolved);
  r->resolved = resolved;
  r->active = 1;
  Cls * c = &r->cls;
  init_redundant_clause (c, size, glue);
  assert (cls2red (c) == r);
  return c;
}

static Cls * allocate_irredundant_clause (unsigned size) {
  size_t bytes = bytes_clause (false, size);
  Cls * c = (Cls *) allocate (bytes);
  init_irredundant_clause (c, size);
  return c;
}

static void dealloc_redundant_clause (Cls * c) {
  assert (c->redundant);
  assert (numred > 0), numred--;
  deallocate (cls2red (c), c->bytes ());
}

static void dealloc_irredundant_clause (Cls * c) {
  assert (!c->redundant);
  assert (numirr > 0), numirr--;
  deallocate (c, c->bytes ());
}

static void dealloc_clause (Cls * c) {
  if (c->redundant) dealloc_redundant_clause (c);
  else dealloc_irredundant_clause (c);
}

// Reallocate the actual memory used by the clause.

static Cls *
realloc_clause (Cls * c, unsigned old_size, unsigned new_size) {
  assert (1 < new_size);
  assert (new_size < old_size);
  assert (c->size () == old_size);
  size_t old_bytes = bytes_clause (c->redundant, old_size);
  size_t new_bytes = bytes_clause (c->redundant, new_size);
  if (c->redundant) {
    Red * r = cls2red (c);
    r = (Red*) reallocate (r, old_bytes, new_bytes);
    if (new_size <= r->glue) {
      r->glue = new_size-1;	// only place were we update the glue
      glueupdates++;
    }
    c = &r->cls;
  } else {
    c = (Cls*) reallocate (c, old_bytes, new_bytes);
  }
  c->sizebits = new_size;
  assert (c->sizebits == new_size);
  return c;
}

/*------------------------------------------------------------------------*/

// These partitions are used for keeping a set of equivalent literal
// candidates during SAT sweeping (see also our LPAR paper in 2013).

static size_t bytes_part (unsigned size) {
  size_t res = sizeof (Part);
  assert (size > 1);
  res += (size - PARTLITS) * sizeof (int);
  return res;
}

static Part * alloc_part (const sector<int> & lits) {
  unsigned size = lits.size ();
  Part * p = (Part *) allocate (bytes_part (size));
  p->size = size;
  p->garbage = false;
  for (unsigned i = 0; i < size; i++) p->lits[i] = lits[i];
  return p;
}

static Part * realloc_part (Part * p, int remove) {
  assert (p->size > 2), assert (!p->garbage);
  unsigned old_size = p->size;
  unsigned new_size = 0;
  for (unsigned j = 0; j < old_size; j++) {
    int lit = p->lits[j];
    if (lit != remove) p->lits[new_size++] = lit;
  }
  assert (new_size + 1 == old_size);
  p = (Part*) reallocate (p, bytes_part (old_size), bytes_part (new_size));
  p->size = new_size;
  return p;
}

static void dealloc_part (Part * p) {
  deallocate (p, bytes_part (p->size));
}

/*------------------------------------------------------------------------*/

static int remaining_variables () {
  int res = maxvar;
  res -= fixed;			// root level assigned units
  res -= eliminated;		// eliminated variables
  res -= substituted;		// equivalent substituted variables
  assert (res >= 0);
  return res;
}

// The number of 'remaining_variables' (not root assigned, not eliminated,
// not substituted), which are assigned.  This is used in computing the
// actual 'trail height' ignoring root assigned eliminated or substituted
// variables.  Otherwise relative changes in trail height are meaningless.
// Note that 'active' does not necessarily include roots, i.e., variables
// fixed at root level 0, before a transition from decision level 0 to 1.
// See 'assign' and 'assume_decision' for more details on the latter.

static int remaining_assigned_variables () { 
  assert (trail.size () <= active);
  unsigned res = active - trail.size ();
  if (!level) assert (res >= trail.size ()), res -= trail.size ();
  return res;
}

/*------------------------------------------------------------------------*/

// Support for proof tracing.  This comes in two flavors, online through
// DRUPLIG and directly tracing to 'std::cout' as requested by the DRAT/DRUP
// format.  Both are triggered by the functions in this section.
// As discussed where 'druplig.h' is included, DRAT/DRUP is the standard
// proof trace format as for instance used in the competition, which
// can be checked by 'drat-trim' or 'drabt, while connecting to the
// 'DRUPLIG' library is useful for online testing and debugging, e.g.,
// 'DRUPLIG' raises a signal if a clause is added which is not implied,
// and thus running this flow in a symbolic debugger such as 'gdb' or
// through a delta debugger such as 'cnfdd' does not require additional
// scripting etc.  It also had and has the additional benefit to allow cross
// checking of DRUPLIG and DRAT-TRIM/DRABT semantics.  This for instance
// showed that DRUPLIG version 0.8 treated learned units lazily, while
// DRAT-TRIM and DRABT perform unit propagation eagerly after adding
// clauses. This gives different semantics w.r.t. which clauses can be
// deleted etc.

static std::ostream * tout = &std::cout;	// 'trace' output stream
static bool delete_and_close_tout;

#ifdef DRUPLIG

template<class C>
static void trace_druplig_literals (const C & c) {
  assert (druplig);
  for (unsigned i = 0; i < c.size (); i++)
    druplig_add_literal (druplig, c[i]);
}

#endif

static void trace_add_original_clause () {
  LOG ("tracing addition of original clause " << clause);
#ifdef DRUPLIG
  if (druplig)
    trace_druplig_literals (clause),
    druplig_add_original_clause (druplig);
#endif
  // DRAT/DRUP format does not include original clauses but
  // DRUPLIG needs to know them for online checking.
}

template<class C>
static void trace_add_learned_clause (const C & c) {
  LOG ("tracing addition of learned clause " << c);
#ifdef DRUPLIG
  if (druplig)
    trace_druplig_literals (c),
    druplig_check_and_add_redundant_clause (druplig);
#endif
  if (!opts::trace) return;
  (*tout) << c;
  if (c.size ()) (*tout) << ' ';
  (*tout) << '0' << NL;
}

static void trace_add_empty_learned_clause () {
  LOG ("tracing addition of empty clause");
#ifdef DRUPLIG
  if (druplig) druplig_check_and_add_redundant_clause (druplig);
#endif
  if (opts::trace) (*tout) << "0" << NL;
}

static void trace_add_unit_learned_clause (int unit) {
  LOG ("tracing addition of unit clause " << unit);
#ifdef DRUPLIG
  if (druplig)
    druplig_add_literal (druplig, unit),
    druplig_check_and_add_redundant_clause (druplig);
#endif
  if (opts::trace) (*tout) << unit << " 0" << NL;
}

static void trace_add_binary_learned_clause (int a, int b) {
  LOG ("tracing addition of binary clause " << a << ' ' << b);
#ifdef DRUPLIG
  if (druplig)
    druplig_add_literal (druplig, a),
    druplig_add_literal (druplig, b),
    druplig_check_and_add_redundant_clause (druplig);
#endif
  if (opts::trace) (*tout) << a << ' ' << b << " 0" << NL;
}

static bool already_traced (const Cls & c) { return c.traced; }
static bool already_traced (const sector<int> & s) { return false; }

template<class C>
static void trace_delete_clause (const C & c) {
  if (already_traced (c)) return;
  LOG ("tracing deletion of clause " << c);
#ifdef DRUPLIG
  if (druplig)
    trace_druplig_literals (c),
    druplig_forget_clause (druplig);
#endif
  if (!opts::trace) return;
  (*tout) << "d " << c;
  if (c.size ()) (*tout) << ' ';
  (*tout) << '0' << NL;
}

static void trace_delete_binary_learned_clause (int a, int b) {
  LOG ("tracing deletion of binary clause " << a << ' ' << b);
#ifdef DRUPLIG
  if (druplig)
    druplig_add_literal (druplig, a),
    druplig_add_literal (druplig, b),
    druplig_forget_clause (druplig);
#endif
  if (opts::trace) (*tout) << "d " << a << ' ' << b << " 0" << NL;
}

/*------------------------------------------------------------------------*/

// The idea of 'triggers' is to avoid certain costly inprocessing attempts
// if they are bound to fail, since nothing changed since the last time it
// was tried (until completion and unsuccessfully).

static void trigger_all () {
  trigger.subsume = true;
  trigger.decompose = true;
  trigger.elim = true;
  trigger.block = true;
  trigger.probe = true;
}

static void trigger_binary () {
  trigger.decompose = true;
  trigger.probe = true;
}

static void trigger_shrunken_or_rewritten (Cls * c) {
  if (!c->redundant) 
    trigger.subsume = trigger.elim = trigger.block = true;
  if (c->size () == 2) trigger_binary ();
}

static bool something_triggered () {
  if (trigger.subsume) return true;
  if (trigger.decompose) return true;
  if (trigger.elim) return true;
  if (trigger.block) return true;
  if (trigger.probe) return true;
  return false;
}

/*------------------------------------------------------------------------*/

static void init_global () {
  LOG ("init");

  // Set '%.2f' globally for 'std::cout'.

  std::cout << std::fixed << std::setprecision (2);

  // The next two limits are really set at 'init_solve'.  Not setting the
  // first to something meaningful leads to many 'Queue.restamp' calls
  // already during parsing, which then results in quadratic parsing effort
  // (even in non-debugging mode), which of course should be avoided.

  limits.queuestamp = (1u<<31);
  limits.resolved = (1u<<31);

  trigger_all ();
  opts::init ();
  init_clause_header (&binary_conflict, false, 2, 2);
  rng.seed (opts::seed);
  resolved = 1;
}

static void init_after_options_have_been_set () {
#ifdef DRUPLIG
  if (!opts::druplig) return;
  druplig = druplig_init ();
  if (opts::drupligcheck) {
    druplig_set_check (druplig, 1);
    LOG ("forward DRUP checking with DRUPLIG enabled");
  }
  if (opts::drupligtrace) {
    druplig_set_trace (druplig, stdout);
    druplig_set_flush (druplig, 1);
    LOG ("DRUP tracing with DRUPLIG enabled");
  }
  if (opts::drupligtraceorig) {
    druplig_set_traceorig (druplig, 1);
    LOG ("DRUP tracing original clauses with DRUPLIG too");
  }
#endif
}

static void init_solve () {
  limits.conflicts.restart = opts::restartint;
  limits.conflicts.reduce = opts::reduceinit;
  limits.simplify.conflicts = opts::simplifyinit;

  inc.conflicts.reduce = opts::reduceinit;
  inc.simplify.conflicts = opts::simplifyint;

  ema.glue.fast.init (opts::gluefastema);
  ema.glue.slow.init (opts::glueslowema);

  ema.trail.fast.init (opts::trailfastema);
  ema.trail.slow.init (opts::trailslowema);

  ema.jump.init (opts::jumpema);

  ema.resolved.size.init (opts::resolvedsizema);
  ema.resolved.glue.init (opts::resolvedgluema);

  // The next two limits are only used for testing, that is fuzzing and
  // delta debugging.  In production usage they should always be set to
  // the maximum value, which delays re{stamp,scor}ing as long as possible.

  limits.queuestamp = (1u << opts::queuestamplim);
  limits.resolved = (1u << opts::resolvedlim);

  if (opts::verbose)
    std::cout
    << "c initial restart interval " << limits.conflicts.restart << NL
    << "c initial reduce interval " << limits.conflicts.reduce << NL
    << "c initial reduce increment " << inc.conflicts.reduce << NL
    << "c initial simplifying increment " << inc.conflicts.reduce << NL
    << "c queue stamp limit " << limits.queuestamp << NL
    << "c resolved limit " << limits.resolved << NL
    << std::flush;

  control.push_back (0);
}

/*------------------------------------------------------------------------*/

// Get the value a literal is assigned to.  The return value is either '-1'
// so negative if the variable is assigned to false, '1' if it is assigned
// to true so positive in this case and zero otherwise, e.g., if unassigned.
// This function takes the sign of the literal into account.  A similar
// bit-parallel version is found further down as 'bpval'.

static int val (int lit) {
  assert (lit);
  assert (lit != INT_MIN);
  int idx = abs (lit);
  assert (idx <= maxvar);
  int res = vals[idx];
  if (lit < 0) res = -res;
  return res;
}

static Var & var (int lit) {
  assert (lit);
  assert (lit != INT_MIN);
  int idx = abs (lit);
  return vars[idx];
}

/*------------------------------------------------------------------------*/

// The solver reports what it is doing after some repeated events occur,
// like 'reduce', 'blocked clause elimination', etc.  Grep for 'report' to
// find all these reported events.  By default these reports are printed
// even if verbose mode is switched off.  With verbose messages enabled more
// reports are printed.

static void report_stats (char * line, int fractionals, double s) {
  char fmt[10], str[20];
  if (s < 0) s = -1; else if (s > 1e15) s = -2;
  sprintf (fmt, " %%.%df", fractionals);
  sprintf (str, fmt, s);
  strcat (line, str);
}

// We try to layout the reported numbers and the headers explicitly.  This
// really makes it much easier to add reported statistics and remove them
// again, which is something happening quite frequently during development.

static void report_header (char * rline, char * hline, const char * name) {
  int stats_start, stats_end;
  assert (rline[0]);
  assert (rline[1] == ' ');
  assert (rline[2] != ' ');
  for (stats_end = 0; rline[stats_end + 1]; stats_end++)
    ;
  for (stats_start = stats_end; rline[stats_start-1] != ' '; stats_start--)
    ;
  int stats_width = stats_end - stats_start + 1;
  int name_width = strlen (name);
  int middle = stats_start + stats_width/2;
  int round = (name_width < stats_width ? -1 : +1);
  int name_start = middle - (name_width + round)/2;
  int len = strlen (hline);
  hline[len++] = ' ';
  while (len < name_start)
    hline[len++] = ' ';
  hline[len] = 0;
  strcat (hline, name);
}

#define REPORT(NAME,FRACTIONALS,STATS) \
do { \
  report_stats (rline, FRACTIONALS, STATS); \
  if (!need_header) break; \
  report_header (rline, hline[h], NAME); \
  if (++h == num_header_lines) h = 0; \
} while (0)

static void report (char type) {

  // Upper case 'type' requires higher 'verbose' mode.

  if (isupper (type) && opts::verbose < 2) return;

  reports++;

  // Number of header lines, change if you want less or more:

  const int num_header_lines = 4;

  char rline[256], hline[num_header_lines][256];
  sprintf (rline, "c %c", type);

  const bool initial = (reports == 1);

  // Only print header initially and after an inprocessing phase.

  const bool need_header = initial || (lastreportsimpmode && !simpmode);

  lastreportsimpmode = simpmode;
  if (simpmode && !lastreportsimpmode) std::cout << 'c' << NL;

  if (need_header)
    for (int i = 0; i < num_header_lines; i++)
      strcpy (hline[i], "c");

  int h = 2;

  REPORT ("sec", 2, get_process_time ());
  REPORT ("mb", 0, space ());
  strcat (rline, " ");
  REPORT ("simplifications", 0, simplifications);
  REPORT ("reductions", 0, reductions);
  REPORT ("restarts", 0, restarts);
  REPORT ("conflicts", 0, conflicts);
  strcat (rline, " ");
  REPORT ("learned", 0, numred);
  REPORT ("glue", 0, ema.glue.slow.norm ());
  REPORT ("jump", 0, ema.jump.norm ());
  REPORT ("trail", 0,
    percent (ema.trail.slow.norm (), remaining_variables ()));
  strcat (rline, "% ");
  REPORT ("resolved-glue", 0, ema.resolved.glue.norm ());
  REPORT ("resolved-size", 0, ema.resolved.size.norm ());
  strcat (rline, " ");
  REPORT ("variables", 0, remaining_variables ());
  REPORT ("clauses", 0, numirr);

  if (need_header) {
    std::cout << 'c' << NL;
    for (int i = 0; i < num_header_lines; i++)
      std::cout << hline[i] << NL;
  } 

  if (need_header || type == '*') std::cout << 'c' << NL;
  std::cout << rline << NL << std::flush;
}

/*------------------------------------------------------------------------*/

// Support for a fast implementation of the VMTF decision heuristics using
// enqueue time stamping.  See our SAT'15 paper for more general information
// about this technique.  It is conceptually much simpler, easier to
// implement and slightly faster than VSIDS (at least in its overhead).
// There are only two more places were these stamps are used: first while
// unassigning literals during backtracking, and second while searching for
// a 'next_decision'.  Bumping variables uses these stamps too, but
// implicitly through 'enqueue' and 'dequeue'. Otherwise this is just a
// doubly linked list with 'first' as 'tail' to enqueue new elements and
// 'last' as head.

static void check_queue () {			// queue invarant
#ifdef EXPENSIVECHECKS
  if (!opts::check) return;
  int lit, prev = 0, last_unassigned = 0;
  for (lit = queue.first; lit; lit = lnks[lit].next) {
    assert (0 < lit), assert (lit <= maxvar);
    // This is is a correct doubly linked list:
    assert (lnks[lit].prev == prev);
    // Queue elements are linearly ordered by their stamps:
    if (prev) assert (lnks[prev].stamp < lnks[lit].stamp);
    if (!vals[lit]) last_unassigned = lit;
    prev = lit;
  }
  // No unassigned variable right of 'searched' towards the head 'last':
  if (queue.searched && last_unassigned)
    assert (lnks[last_unassigned].stamp <= lnks[queue.searched].stamp);
#endif
}

static Lnk & lnk (int lit) {
  assert (lit);
  assert (lit != INT_MIN);
  int idx = abs (lit);
  return lnks[idx];
}

void Queue::restamp () {
  int idx = first;
  assert (queue.stamp >= limits.queuestamp);
  stamp = 0;
  while (idx) {
    Lnk & l = lnk (idx);
    l.stamp = stamp++;
    idx = l.next;
  }

  // In order to test 'restamp_queue' it should be triggered during testing,
  // which hardly is the case if we wait until the stamp would overflow.
  // Thus we added 'limits.queuestamp' which is initialized in 'init_global'
  // and later to the final value in 'init_solve' using the option
  // 'opts::queuestamplim' (which can be fuzzed).  If the limit is too small
  // then 'restamp_queue' is triggered too often and thus we update it here
  // to avoid this situation, which is guaranteed to occur otherwise during
  // delta debugging options, i.e., when 'opts::queuestamplim' shrinks.
  // Note, that in contrast to the 'resolved' field in redundant clauses
  // 'Red', this field in 'Lnk' has a much higher chance to overflow, so it
  // is not inconceivable that a long SAT run would trigger overflow even if
  // a 64 bit integer is used in stead of the 32 'unsigned' we use.  Thus
  // this design decision to make it 32 bit saves space and also prevents
  // overflow issues, since those overflows do occur in shorter runs.

  while (limits.queuestamp < (1u<<31) &&
         limits.queuestamp/2 <= (unsigned) maxvar)
    limits.queuestamp <<= 1;

  LOG ("restamped " << queue.stamp);
  report ('Q');
}

inline void Queue::enqueue (int idx) {
  if (stamp >= limits.queuestamp) restamp ();
  assert (stamp < INT_MAX);
  assert (idx), assert (idx <= maxvar);
  LOG ("enqueue " << idx << " stamp " << stamp);
  Lnk & l = lnks[idx];
  assert (!l.prev), assert (!l.next);
  l.stamp = stamp++;	// 'Lnk.stamp' acts as 'enqueue time'
  l.prev = last;
  l.next = 0;
  if (last) lnks[last].next = idx; else first = idx;
  last = idx;
  if (!val (idx)) searched = last;
}

inline void Queue::dequeue (int idx) {
  assert (idx), assert (idx <= maxvar);
  Lnk & l = lnks[idx];
  LOG ("dequeue " << idx << " stamp " << l.stamp);
  if (l.prev) {
    Lnk & prev = lnks[l.prev];
    assert (prev.next == idx);
    prev.next = l.next;
  } else {
    assert (first == idx);
    first = l.next;
  }
  if (l.next) {
    Lnk & next = lnks[l.next];
    assert (next.prev == idx);
    next.prev = l.prev;
  } else {
    assert (last == idx);
    last = l.prev;
  }
  if (searched == idx) {
    searched = l.prev;
    if (searched) LOG ("searched dequeue " << lnk (searched).stamp);
    else LOG ("searched invalid");
  }
#ifndef NDEBUG
  l.prev = l.next = 0;
#endif
}

/*------------------------------------------------------------------------*/

static void init_after_maximum_variable_index_known () {
  if (maxvar >= (1<<LDMAXVAR))
   DIE ("can not handle " <<
     (1<<LDMAXVAR) << " = 2^" << LDMAXVAR << " variables");
  LOG ("setting up " << maxvar << " variables");
  maxlit = 2*maxvar + 1;
  active = maxvar;
  NEW (Var, vars, maxvar + 1);
  NEW (Wchs, watches, maxlit + 1);
  CALLOC (Val, vals, maxvar + 1);
  CALLOC (Val, marks, maxvar + 1);
  ALLOC (Val, phases, maxvar + 1);
  for (int i = 1; i <= maxvar; i++) phases[i] = -1;
  NEW (Lnk, lnks, maxvar + 1);
  for (int i = maxvar; i > 0; i--) queue.enqueue (i);
  queue.searched = queue.last;
  LOG ("searched enqueue " << lnk (queue.searched).stamp);
  init_after_options_have_been_set ();
}

static void reset_parts () {
  for (unsigned i = 0; i < parts.size (); i++)
    dealloc_part (parts[i]);
  release (parts);
}

static void reset () {
  if (delete_and_close_tout) { ((std::ofstream*)tout)->close (); delete tout; }
  LOG ("reset");
#ifdef DRUPLIG
  if (opts::druplig) {
    assert (druplig);
    std::cout << std::flush;
    druplig_stats (druplig, stdout);
    druplig_reset (druplig);
  }
#endif

#ifndef NDEBUG

  // As long we only have global static variables (and are non-reentrant),
  // there is no need to deallocate anything.  This will change for a
  // multi-entrant version.  In order to find memory leaks (see 'assert
  // (!allocted)' below), it is still useful to deallocte everything at
  // least in debugging mode.

  for (unsigned i = 0; i < clauses.size (); i++)
    dealloc_clause (clauses[i]);

  reset_parts ();

  DEALLOC (Var, vars, maxvar + 1);
  DEALLOC (Lnk, lnks, maxvar + 1);
  DEALLOC (Val, vals, maxvar + 1);
  DEALLOC (Val, marks, maxvar + 1);
  DEALLOC (Val, phases, maxvar + 1);

  // See lengthy comments to 'sector' above.  Need explicit deallocation.
  // We could use a 'placement delete' operator but then our 'arena'
  // optimization is hard to realize.

  for (int i = 2; i <= maxlit; i++) release (watches[i]);
  DEALLOC (sector<unsigned>, watches, maxlit + 1);

  release (clauses);
  release (control);
  release (trail);
  release (tagged);
  release (lcas);
  release (part);
  release (clause);
  release (unshrunken);
#ifndef NDEBUG		// Well redundant, but 'orig' is only defined then.
  release (orig);
#endif
  release (extend);
  release (levels);
  release (work);
  release (resched);
  release (probes);
  release (units);
  release (bins);

  RESOLVE (release (resolvent));

  deallocate_arena ();

  // This environment variable check let us run an external leak checker
  // such as 'valgrind --leak-check=yes' without recompilation.

  if (!getenv ("SPLATZLEAK")) assert (!allocated);

#endif
}

/*------------------------------------------------------------------------*/

static int sign (int lit) { return lit < 0 ? -1 : 1; }

static void mark_lit (int lit) { marks [abs (lit)] = sign (lit); }

static int marked_lit (int lit) { return marks[abs (lit)] * sign (lit); }

static void unmark_lit (int lit) { marks [ abs (lit) ] = 0; }

static bool decision (int lit) { return var (lit).decision (); }

template<class C>
static void mark_clause (const C & c) {
  const unsigned size = c.size ();
  for (unsigned i = 0; i < size; i++) mark_lit (c[i]);
}

template<class C>
static void unmark_clause (const C & c) {
  const unsigned size = c.size ();
  for (unsigned i = 0; i < size; i++) unmark_lit (c[i]);
}

/*------------------------------------------------------------------------*/

// Had to move these member functions here, because they need 'clauses'.

inline Rsn::Rsn (Cls * c, unsigned cidx) : word (cidx) {
  (void) c;
  assert (0u <= cidx), assert (cidx < clauses.size ());
  assert (clauses[cidx] == c);
  word <<= 1;
  word |= 1u;
  assert (!binary ());
}

inline Cls * Rsn::cls () const {
  return binary () ? 0 : clauses[word >> 1];
}

/*------------------------------------------------------------------------*/

// It is probably better to have literal objects for a negative and the
// corresponding positive literal allocated next to each other.  Thus we map
// literals to an unsigned value (though still using 'int') in the same way
// MiniSAT and AIGER are encoding literals as unassigned numbers, by using
// the least significant bit of the unsigned number as actual sign bit.  The
// actual variable index can be obtained by dividing the unsigned number by
// two.  This seems to be particularly important for watcher stacks which
// are accessed quite frequently.  It would be possible to map all literals
// to these numbers (as in MiniSAT and AIGER), but that makes it harder to
// follow the internals of the solver during testing and debugging, and we
// decided against it (as in Lingeling).

static int litidx (int lit) {
  assert (lit);
  assert (lit != INT_MIN);
  int res = abs (lit);
  assert (res <= maxvar);
  res *= 2;
  if (lit < 0) res++;
  assert (1 <= res);
  assert (res <= maxlit);
  return res;
}

static Wchs & wchs (int lit) { return watches[litidx (lit)]; }

/*------------------------------------------------------------------------*/

// The functions in this section checks whether two literals have common
// ancestor in the binary implication graph (BIG), which could also be
// interpreted as a dominator in BIG.

static void mark_lca (int lit) {
  LOG ("mark lca " << lit);
  Var & v = var (lit);
  assert (!v.lca);
  v.lca = true;
  lcas.push_back (lit);
}

static bool marked_lca (int lit) { return var (lit).lca; }

static int
least_common_ancestor_for_hyper_binary_resolution (int a, int b) {

  assert (level == 1);	// Only used and supported for decision level 1.

  assert (b);
  assert (val (b) < 0);
  if (!a) return b;
  assert (val (a) < 0);
  assert (lcas.empty ());

  // In this first loop we follow back the binary implication graph starting
  // from the first literal 'a' and mark all variables which can be reached
  // that way.  If a decision or a non-binary reason is found then abort.

  int lit = a;
  while (lit) {
    assert (val (lit) < 0);
    assert (var (lit).level == 1);
    assert (!marked_lca (lit));
    mark_lca (lit);
    Rsn & r = var (lit).reason;
    if (r.decision ()) break;			// abort at decision
    if (r.binary ()) lit = r.lit ();		// binary blit
    else {
      Cls * c = r.cls ();			// check actually binary
      int pred = 0;
      for (unsigned i = 0; i < c->size (); i++) {
	int other = c->lits[i];
	if (other == -lit) continue;
	assert (val (other) < 0);
	assert (other != lit);
	if (!var (other).level) continue;	// ignore fixed
	assert (var (other).level == 1);
	if (pred) { pred = 0; break; }		// non-binary
	pred = other;
      }
      lit = pred;	// abort if 'pred' remained zero
    }
  }

  // In the second loop we search backward through the binary implication
  // graph starting from the second literal 'b' and try to find the first
  // variable backward reachable from 'a' as well.

  int res = b;
  while (res) {
    assert (val (res) < 0);
    assert (var (res).level == 1);
    if (marked_lca (res)) break;
    Rsn & r = var (res).reason;
    if (r.decision ()) { res = 0; break; }	// no common ancestor
    if (r.binary ()) res = r.lit ();
    else {
      Cls * c = r.cls ();
      int pred = 0;
      for (unsigned i = 0; i < c->size (); i++) {
	int other = c->lits[i];
	if (other == -res) continue;
	assert (val (other) < 0);
	assert (other != res);
	if (!var (other).level) continue;
	assert (var (other).level == 1);
	if (pred) { pred = 0; break; }		// no common ancestor
	pred = other;
      }
      res = pred;	// could also be zero, but that is fine ...
    }
  }

  // Clean-up variables marked in first loop.

  while (!lcas.empty ()) {
    var (lcas.back ()).lca = 0;
    lcas.pop_back ();
  }

  LOG ("least common ancestor " << res << " of " << a << " and " << b);
  return res;
}

/*------------------------------------------------------------------------*/

// Beware that left shift of negative integers gives undefined behaviour
// according to the C standard and is checked/used in newer compiler
// versions (new 'LLVM/clang' versions for sure).  This issue also occurs at
// other places in this code. They should all have a reference to 'sll' in
// comments.  Our method to circumvent this issue is to cast the signed word
// to an unsigned word, peform the shift and cast it back, which is fine
// according to the C standard and does what we want if we use this for
// bit-stuffing a bit at the least-signifcant bit position in a 31 bit
// signed (two-complement) integer.

static int sll (int lit) {
  assert (-(1<<30) < lit), assert (lit < (1<<30));  // see 'LDMAXVAR'
  return (int)(((unsigned)lit)<<1);
}

/*------------------------------------------------------------------------*/

// Binary clauses only have 'blocking literals' but no references to the
// actual clause in the watcher stacks.  In Lingeling we also do it for
// ternary clauses, but wanted to avoid the complexity of that (for now).

static void watch_binary (int lit, int blit) {
  LOG ("watch binary " << lit << " blit " << blit);
  wchs (lit).push_back (sll (blit));
}

// Watches for large (that is non-binary) clauses have a blocking literal
// 'blit' and a reference to the clause in the 'clauses' stack.  Note that
// 'blocking' here has nothing to do with 'blocking' (or even 'blocking
// literal') in the context of blocked clause.  This term of blocking was
// introduced by Chu and Stuckey and caches probably satisfied literals in a
// clause to avoid actually accessing the clause during BCP.

static void watch_large (int lit, int blit, unsigned cidx) {
  LOG ("watch large " << lit <<
    " blit " << blit << " in " << *clauses[cidx]);
  Wchs & ws = wchs (lit);
  ws.push_back (sll (blit) + 1);
  ws.push_back (cidx);
}

// Update the 'blit' for a large watched clause.  This is accumulated
// expensive but unfortunately necessary since otherwise during shrinking
// clauses the current blit might get removed and thus the clause is not
// satisfied if that removed literal is true.  Since the propagation
// algorithm checks for blits to be true before checking the actual clause,
// not eagerly updating a blit might result in 'unpropagated clauses', e.g.,
// clauses which are not properly watched.  This would break important core
// invariants of CDCL at various places.

static void update_large_blit (int lit, unsigned cidx, int new_blit) {
  Wchs & ws = wchs (lit);
  const unsigned s = ws.size ();
  unsigned i = 0;
  while (i < s) {
    if (!(ws[i++] & 1)) continue;
    if (ws[i++] == cidx) break;
  }
  assert (s), assert (1 < i);
  i -= 2; assert (ws[i] & 1);
#ifdef LOGGING
  int old_blit = ws[i]; old_blit >>= 1;
  LOG ("update large " << lit << " old blit " << old_blit <<
    " to new blit " << new_blit << " in " << *clauses[cidx]);
#endif
  ws[i] = sll (new_blit) + 1;
}

static void watch (Cls * c, unsigned cidx) {
  const int lit = c->lits[0];
  const int other = c->lits[1];
  assert (clauses[cidx] == c);
  if (c->size () == 2) {
    watch_binary (lit, other);
    watch_binary (other, lit);
  } else {
    assert (c->size () > 2);
    watch_large (lit, other, cidx);
    watch_large (other, lit, cidx);
  }
}

// During 'propagation' watches are removed, which is logged with this.

static void log_unwatch (int lit, int blit, unsigned cidx) {
  LOG ("unwatch large " << lit <<
    " blit " << blit << " in " << *clauses[cidx]);
}

/*------------------------------------------------------------------------*/

static void move_max_level_lit_to_front (size_t start) {
  for (unsigned i = start + 1; i < clause.size (); i++)
    if (var (clause[start]).level < var (clause[i]).level)
      std::swap (clause[start], clause[i]);
}

// Make sure that the first two literals in the clause are those with
// maximum decision level.  Watching literals with smaller decision level
// might miss propagations.  For redundant clauses this does not have any
// correctness impact though.

static void move_max_level_lits_to_front () {
  move_max_level_lit_to_front (0);
  move_max_level_lit_to_front (1);
}

// Adding a redundant clause is split into two parts, one part peforms basic
// addition and another part which then inserts watches to clauses.  The
// first part is needed by 'move_clause', which already has watches to the
// clause (index) but needs to reallocate a clause (redundant clauses have a
// 'Red' header).

static Cls *
add_but_do_not_watch_redundant_clause (unsigned glue, unsigned & cidx) {
  Cls * c = allocate_redundant_clause (glue);
  LOG ("learned length " << c->size () << " clause " << *c);
  if (clauses.size () >= (1u<<LDMAXCIDX))
    DIE ("redundant clause reaches limit of " <<
      (1u<<LDMAXCIDX) << " = 2^" << LDMAXCIDX << " clauses");
  cidx = clauses.size ();
  clauses.push_back (c);
  return c;
}

static Cls * add_redundant_clause (unsigned glue, unsigned & cidx) {
  move_max_level_lits_to_front ();
  Cls * c = add_but_do_not_watch_redundant_clause (glue, cidx);
  watch (c, cidx);
  return c;
}

/*------------------------------------------------------------------------*/

// The following technique, originally came out of my PrecoSAT solver which
// had on-the-fly hyper binary resolution on all decision levels (by keeping
// a reference to the closest dominator in the binary implication graph of
// an implied literal).  However, it turned out that most of the binary
// clauses derived by hyper binary resolutions occur on decision level one
// anyhow and thus a faster algorithm can be used for this special case.  It
// is clear that on decision level one, whenever there is a large non-binary
// reason for an implied literal, then a hyper binary resolution is possible
// and will give a better binary reason.  So there is no need for the
// dominator check.  This gives the following procedure. For more details on
// this technique and its effects in general read our CPAIOR'13 paper or the
// thesis or papers by Hyojung Han and Fabio Somenzi (like DATE'11).

static int on_the_fly_hyper_binary_resolution (int lit) {
  LOG ("hyber binary resolution for " << lit);
  hbrs++;
  assert (level == 1);
  assert (var (lit).reason.clausal ());
  assert (opts::hbr);
  Rsn & r = var (lit).reason;
  Cls * c = r.cls ();
  assert (c->size () > 2);
  assert (!c->garbage);
  int lca = 0;
  for (unsigned i = 0; i < c->size (); i++) {
    int other = c->lits[i];
    if (other == lit) continue;
    if (!var (other).level) continue;
    lca = least_common_ancestor_for_hyper_binary_resolution (lca, other);
    if (!lca) break;
  }
  if (!lca) {
    lca = -control[1].decision;
    LOG ("negated decision " << lca << " as least common ancestor");
  } else LOG ("least common ancestor " << lca);
  assert (lca);
  assert (val (lca) < 0);
  assert (!val (lit));
  bool found = false;
  for (unsigned i = 0; !found && i < c->size (); i++)
    found = (c->lits[i] == lca);
  if (found) {
    LOG ("self-subsuming hyper-binary resolvent " << lca << ' ' << lit);
    copy (unshrunken, *c);
    unsigned cidx = r.cidx ();
    assert (clauses[cidx] == c);
    c = realloc_clause (c, c->size (), 2);
    clauses[cidx] = c;
    int pos = c->lits[1] == lit;
    assert (c->lits[pos] == lit);
    if (c->lits[!pos] != lca) {
      c->lits[!pos] = lca;
      watch_binary (lca, lit);
      update_large_blit (lit, cidx, lca);
      // This does produce a stale third 'watch', but see comments in
      // 'on_the_fly_conflicts' why this is no problem there nor here (since
      // we patched 'propagate' accordingly).  We further mix binary and
      // large watches, which should also neither be a problem as long we do
      // not use the type of one watch as indicator for the type of the
      // other watch in a clause (it breaks the invariant that only
      // non-binary clauses have a reference though).
    }
    trace_add_binary_learned_clause (lit, lca);
    trace_delete_clause (unshrunken);
    trigger_shrunken_or_rewritten (c);
    hbrsubs++;
  } else {
    LOG ("resolved hyper binary clause " << lca << ' ' << lit);
    clause.clear ();
    clause.push_back (lca);
    clause.push_back (lit);
    unsigned cidx;
    (void) add_redundant_clause (0, cidx);
    trace_add_binary_learned_clause (lca, lit);
    clause.clear ();
  }
  return lca;
}

/*------------------------------------------------------------------------*/

static bool is_fixed_lit (int lit) { return var (lit).status == FIXED; }

static bool is_elim_lit (int lit) { return var (lit).status == ELIMINATED; }

// The name of the following function is a little bit misleading.  We do
// have a separate 'active' counter of active variables.  Those 'active'
// variables do include fixed (that is root-level assigned variables aka
// units) before they are propagated and only after they completely have
// been propagated they become 'inactive' (see 'flush_root_trail').  However
// as soon the 'status' of a variable becomes non-zero we consider them as
// 'inactive'.  This is the case at the point a variable is fixed (assigned
// at decision level zero), see 'assign' and the comments for 'active'
// there.  So a more precise but much longer name would be
// 'is_active_and_not_fixed'.  One could consider a fixed but not propagated
// variable as being on its way to become inactive soon and thus treat is as
// such immediately.
//
// To summarize, there are three classes of variables:
//
//   (1) inactive for sure: eliminated or substituted, fixed and propagated
//   (2) not quite inactive yet: fixed but not propagated yet
//   (3) active: so neither eliminated, substituted, nor fixed
//
// For the first set (1) of really inactive variables we can remove the
// variable from the decision queue and further release its watches.  For
// those fixed but not propagated variables of type (2) we have to wait
// until they are flushed from the trail in 'flush_root_trail'.

static bool is_active_lit (int lit) { return !var (lit).status; }

static void release_inactive_wchs (int lit) {
  assert (!is_active_lit (lit));
  release (wchs (lit));
  release (wchs (-lit));
}

// Used as part of "equivalent literal substitution" (ELS) in 'decompose'.

static void assign_substituted (int lit) {
  LOG ("substituted " << lit);
  assert (!val (lit));
  assert (!level);
  Var & v = var (lit);
  v.level = 0;;
  v.reason = Substituted ();
  assert (!v.status);
  v.status = SUBSTITUTED;
  substituted++;
  assert (active);
  active--;
  trigger_all ();
  int idx = abs (lit);
  vals[idx] = sign (lit);
  queue.dequeue (idx);
  release_inactive_wchs (idx);
}

// Used as part of "bounded variable elimination" (BVE) in 'elim'.

static void assign_eliminated (int lit) {
  LOG ("eliminated " << lit);
  assert (!val (lit));
  assert (!level);
  Var & v = var (lit);
  v.level = 0;;
  v.reason = Eliminated ();
  assert (!v.status);
  v.status = ELIMINATED;
  eliminated++;
  assert (active);
  active--;
  trigger_all ();
  int idx = abs (lit);
  vals[idx] = sign (lit);
  queue.dequeue (idx);
  release_inactive_wchs (idx);
}

// Regular assignment of a variable (also for decisions).

static void assign (int lit, const Rsn & reason) {

#ifdef LOGGING
  if (!level) LOG ("assign " << lit << " unit");
  else LOG ("assign " << lit << " reason " << reason);
#endif

  assert (!val (lit));
  int idx = abs (lit);
  Var & v = vars[idx];
  assert (level < (1u<<30)); // 'v->level' could be bit-field of 30 bits
  v.level = level;

  if (!level) {

    v.reason = Unit ();
    assert (!v.status);
    v.status = FIXED;		// 'active' is not changed here yet
    fixed++;			// since we have to propagate first
    trigger_all ();

    // We would want to do 'active--' as well as 'dequeue (idx)', but
    // this conflicts with propagating units learned during simplifications
    // withouth watching all clauses.  We have to reconnect those clauses
    // and propagate again (setting 'next = 0').  Since the trail is used as
    // BFS queue for propagation, this has to wait.  See also comments below.
    // The same applies to 'dequeue', because 'next_decision' assumes that
    // 'all_variables_assigned' was tested, which in turn uses the 'active'
    // counter to determine that everything has been assigned.

  } else {
    
    v.reason = reason;

    if (level == 1 &&
        opts::hbr &&			// hyper binary resolution?
	v.reason.clausal () &&
	v.reason.cls ()->size () > 2 &&

	// The following limit became necessary since there are benchmarks
	// (for instance 'UR-20-10p1' from 'diagnosis') with too many
	// hyper binary resolvents (HBR) even without probing and sweeping.
	// The idea is to increase the limit for HBRs slowly only.
	// Note, it is dangerous to include 'numred' here, since most
	// of the HBRs are redundant clauses anyhow, which could
	// give a self-amplifying effect.  We do it nevertheless with small
	// weight though.

	(!opts::hbrlim || hbrs < conflicts/8 + numirr/16 + numred/32)) {

      int lca = on_the_fly_hyper_binary_resolution (lit);
      v.reason = Rsn (lca);
      LOG ("assign " << lit << " new reason " << v.reason);
    }
  }

  // Note, that we treat units as 'active' for now until they are propagated
  // completely for the first time, e.g., during 'assume_decision'.  Then
  // they can savely be popped from the trail and the number of 'active'
  // literals fixed.  See comments to 'assume_decision'.

  trail.push_back (lit);

  Val tmp = sign (lit);
  vals[idx] = tmp;

  // As observed by Knot Pipatsrisawat and Adnan Darwiche 'phase saving'
  // keeps the solver in the same search space and handles disconnected
  // compenents of the CNF gracefully. The saved phase is used to pick the
  // phase of the next decision literal (in 'next_decision' below).

  phases[idx] = tmp;
}

// For our VMTF implementation, as described in our SAT'15 paper,
// unassigning a variable might trigger moving the 'searched' so far
// reference, if it happens that the unassigned literal is to the right of
// this reference (has been enqueued at the tail of the queue thus bumped
// more recently).  This update is on the other hand not necessary for the
// first UIP, which immediately after bumping becomes assigned again.  This
// 'except'ion leads to less searches (but not dramatically) and is a new
// feature of this solver (not published before).

static void unassign (int lit, int except) {
  LOG ("unassign " << lit);
  assert (val (lit) > 0);
  int idx = abs (lit);
  assert (0 < idx), assert (idx <= maxvar);
  if (decision (idx)) {
    assert (control[var (idx).level].decision == lit);
    assert (var (idx).level + 1u == control.size ());
    control.pop_back ();
  }
  vals [idx] = 0;
  assert (0 <= except), assert (except <= maxvar);
  if (idx != except && lnks[queue.searched].stamp < lnks[idx].stamp) {
    queue.searched = idx;
    LOG ("searched unassign " << lnk (queue.searched).stamp);
  }
}

// Root level assigned variables aka units will need to be propagated and
// only then can be popped from the trail (since we never backtrack over
// them and thus never unassign root level assigned variables).  They are
// inactive in the sense that we can ignore them as soon they have at
// least once completely been propagated.  This is safely the case here if
// the assertion /*(1)*/ holds.

static void flush_root_trail () {
  assert (!level);
  if (inconsistent) return;
  if (trail.empty ()) return;
  assert (next == trail.size ());		/*(1)*/
  assert (active >= trail.size ());
  active -= trail.size ();
  for (unsigned i = 0; i < trail.size (); i++) {
    int idx = abs (trail[i]);
    release_inactive_wchs (idx);
    queue.dequeue (idx);
  }
  trail.clear ();
  next = 0;
}

/*------------------------------------------------------------------------*/
#ifndef NDEBUG

void dump (Cls * c) { std::cout << *c << NL; }

void dump () {
  std::cout << "p cnf " << maxvar << ' ' <<  clauses.size () << NL;
  for (unsigned i = 0; i < clauses.size (); i++)
    std::cout << *clauses[i] << " 0" << NL;
}

void dump_non_garbage () {
  unsigned count = 0;
  for (unsigned i = 0; i < clauses.size (); i++)
    if (!clauses[i]->garbage) count++;
  std::cout << "p cnf " << maxvar << ' ' <<  count << NL;
  for (unsigned i = 0; i < clauses.size (); i++)
    if (!clauses[i]->garbage) std::cout << *clauses[i] << " 0" << NL;
}

// Built in checker for models.  This is pretty useful for catching and
// shrinking certain bugs triggered by incorrect clause removal during
// inprocessing or violations of watching invariants during search,
// something triggered for instance by too aggressive incorrect
// clause removal techniques.  Having such a checker inside the solver makes
// delta-debugging and testing easier.

static void check_model () {
  LOG ("checking model");
  for (unsigned i = 0; i < orig.size (); i++) {
    bool satisfied = false;
    int lit, j;
    for (j = 0; (lit = orig[i + j]); j++)
      if (val (lit) > 0) satisfied = true;
    if (!satisfied) {
      std::cout << std::flush;
      std::cerr << "splatz: unsatisfied original clause:"  << NL;
      for (j = 0; (lit = orig[i + j]); j++)
	std::cerr << lit << ' ';
      std::cerr << '0' << NL << std::flush;
    }
    assert (satisfied);
    i += j;
  }
}

#endif
/*------------------------------------------------------------------------*/

static bool is_white_space (int ch) {
  return ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r';
}

static void parse_header (FILE * file) {
  char line[80];
  int ch;
  for (;;) {
    ch = getc (file);
    if (ch == 'c') {
      unsigned i = 0;
      while ((ch = getc (file)) != '\n')
	if (ch == EOF) DIE ("unexpected end-of-file in header comment");
	else if (i < sizeof line && ch != ' ') line[i++] = ch;
      if (i && i < sizeof line) line[i++] = 0, parseopt (line);
    } else if (ch != 'p')
      DIE ("expected 'p' or 'c' in header");
    else break;
  }
  if (getc (file) != ' ') DIE ("expected space after 'p'");
  while ((ch = getc (file)) == ' ')
    ;
  if (ch != 'c') DIE ("expected 'c' after 'p '");
  if (getc (file) != 'n') DIE ("expected 'n' after 'p c'");
  if (getc (file) != 'f') DIE ("expected 'f' after 'p cn'");
  if (getc (file) != ' ') DIE ("expected space after 'p cnf'");
  while ((ch = getc (file)) == ' ')
    ;
  if (!isdigit (ch)) DIE ("expected digit after 'p cnf '");
  maxvar = ch - '0';
  if (maxvar) {
    while (isdigit (ch = getc (file))) {
      int digit = ch - '0';
      if (INT_MAX/10 < maxvar || INT_MAX - digit < (maxvar *= 10))
	DIE ("maximum variable index exceeds INT_MAX");
      maxvar = maxvar + digit;
    }
  } else ch = getc (file);
  if (ch != ' ') DIE ("expected space after 'p cnf " << maxvar << "'");
  while ((ch = getc (file)) == ' ')
    ;
  if (!isdigit (ch)) DIE ("expected digit after 'p cnf '");
  maxcls = ch - '0';
  if (maxcls) {
    while (isdigit (ch = getc (file))) {
      int digit = ch - '0';
      if (INT_MAX/10 < maxcls ||
	  INT_MAX - digit < (maxcls *= 10))
	DIE ("maximum variable index exceeds INT_MAX");
      maxcls = maxcls + digit;
    }
  } else ch = getc (file);
  if (!is_white_space (ch))
    DIE ("expected white space character after 'p cnf " <<
       maxvar << ' ' << maxcls << "'");
  while (ch != '\n' && is_white_space (ch = getc (file)))
    ;
  if (ch != '\n')
    DIE ("new line missing after 'p cnf " <<
       maxvar << ' ' << maxcls << "'");

  std::cout << "c found 'p cnf "
       <<  maxvar << ' ' << maxcls
       << "' header" << NL << std::flush;
}

static void add_original_clause () {
  assert (!level);

  // First determine whether the parsed original clause is trivial, aka
  // tautological, i.e., contains two literals over the same variable with
  // different signs (phases) or is satisfied on the root level by units.
  // At the same time, determine, whether it can be shrunken, i.e., contains
  // two identical literals or a literal falsified at the root level.

  bool trivial = false, shrink = false;
  for (unsigned i = 0; i < clause.size (); i++) {
    int lit = clause[i];
         if (val (lit) < 0 || marked_lit (lit) > 0) shrink = true;
    else if (val (lit) > 0 || marked_lit (lit) < 0) trivial = true;
    else mark_lit (lit);
  }
  unmark_clause (clause);

  trace_add_original_clause ();		// DRUPLIG needs original clauses.

  // Simply ignore trivial or root level satisfied clauses.

  if (trivial) { trace_delete_clause (clause); return; }

  // Now shrink the clause if a falsified literal or duplicates have been
  // found. This can not easily be merged with the previous loop without
  // eagerly copying the original clause first, due to trace generation,
  // which has to work with the real original clause.  So this whole section
  // here is only used for proper trace generation.

  if (shrink) {

    copy (unshrunken, clause);

    unsigned j = 0;
    for (unsigned i = 0; i < clause.size (); i++) {
      int lit = clause[i];
      if (val (lit) < 0 || marked_lit (lit) > 0) continue;
      assert (!val (lit)), assert (!marked_lit (lit));
      clause[j++] = lit;
      mark_lit (lit);
    }
    clause.resize (j);
    unmark_clause (clause);

    // After the shrunken clause is added, we can delete the original
    // clause.  Actually, the semantics of how root level units and
    // duplicate literals are handled both in DRAT-TRIM and DRABT (and in
    // newer versions of DRUPLIG) would allow to swap the following two
    // lines, which means copying the clause is not necessary at all.
    // However, for cleaner semantics we keep it in this order.

    trace_add_learned_clause (clause);
    trace_delete_clause (unshrunken);
  }

  if (!clause.size ()) {
    inconsistent = true;
    LOG ("added original empty clause");
  } else if (clause.size () == 1) {
    LOG ("added original unit clause " << clause[0]);
    assign (clause[0], Decision ());
  } else {
    assert (clause.size () >= 2);
    Cls * c = allocate_irredundant_clause (clause.size ());
    LOG ("added original length " << c->size () << " clause " << *c);
    if (clauses.size () >= (1u<<LDMAXCIDX))
      DIE ("original clause reaches limit of " <<
        (1u<<LDMAXCIDX) << " = 2^" << LDMAXCIDX << " clauses");
    unsigned cidx = clauses.size ();
    clauses.push_back (c);
    watch (c, cidx);
  }
}

static void parse_body (FILE * file) {
  int parsed_clauses = 0;
  for (;;) {
    int ch = getc (file);
    if (is_white_space (ch)) continue;
    if (ch == EOF) break;
    if (ch == 'c') {
      while ((ch = getc (file)) != '\n')
	if (ch == EOF)
	  DIE ("unexpected end-of-file in body comment");
      continue;
    }
    int sign = 1;
    if (ch == '-') {
      sign = -1;
      if ((ch = getc (file)) == EOF)
	DIE ("unexpected end-of-file after '-'");
      if (!isdigit (ch))
	DIE ("expected digit after '-'");
      if (ch == '0')
	DIE ("expected non zero digit after '-'");
    } else if (!isdigit (ch))
      DIE ("expected digit or '-'");
    int lit = (ch - '0');
    if (!lit) {
      ch = getc (file);
      if (ch == 'c') ungetc (ch, file);
      else if (!is_white_space (ch))
	DIE ("expected white-space or 'c' after '0'");
    } else {
      while (isdigit (ch = getc (file))) {
	int digit = ch - '0';
	if (INT_MAX/10 < lit || INT_MAX - digit < (lit *= 10))
	  DIE ("variable index exceeds INT_MAX");
	lit = lit + digit;
      }
      lit *= sign;
    }
    if (ch == 'c') ungetc (ch, file);
    else if (!is_white_space (ch))
      DIE ("expected white-space or 'c' after literal " << lit);
    if (abs (lit) > maxvar)
      DIE ("literal " << lit <<
        " exceeds maximal variable index " << maxvar);
#ifndef NDEBUG
    if (orig.size () == (1u<<30))
      DIE ("clause to big: can not save all original literals");
    orig.push_back (lit);
#endif
    if (parsed_clauses >= maxcls)
      DIE ("parsed too many clauses");
    if (lit) {
      if (clause.size () == (1u<<LDMAXSIZE))
	DIE ("parsed clause with " <<
	  (1u<<LDMAXSIZE) << " = 2^" << LDMAXSIZE << " literals");
      clause.push_back (lit);
    } else {
      parsed_clauses++;
      add_original_clause ();
      clause.clear ();
    }
  }
  if (clause.size ()) DIE ("terminating zero missing");
  if (parsed_clauses + 1 == maxcls)
    DIE ("one clause missing");
  if (parsed_clauses < maxcls)
    DIE (maxcls - parsed_clauses << " clauses missing");
  assert (parsed_clauses == maxcls);
  std::cout
  << "c finished parsing " << parsed_clauses << " clauses"
  << " after " << get_process_time () << " seconds" << NL << std::flush;
}

static bool has_suffix (const char * path, const char * suffix) {
  if (strlen (path) < strlen (suffix)) return false;
  return !strcmp (path + strlen (path) - strlen (suffix), suffix);
}

static FILE * read_compressed  (const char * fmt, const char * path) {
  size_t bytes = strlen (fmt) + strlen (path) + 10;
  char * cmd;
  ALLOC (char, cmd, bytes);
  sprintf (cmd, fmt, path);
  FILE * res = popen (cmd, "r");
  DEALLOC (char, cmd, bytes);
  return res;
}

static void parse (const char * path) {
  START (parse);
  int closefile = 2;
  FILE * file = 0;
  if (!path) {
    file = stdin, closefile = 0;
    std::cout << "c reading from <stdin>" << NL << std::flush;
  } else {
         if (has_suffix (path, ".bz2"))
      file = read_compressed ("bzcat %s", path);
    else if (has_suffix (path, ".gz"))
      file = read_compressed ("gunzip -c %s", path);
    else if (has_suffix (path, ".7z"))
      file = read_compressed ("7z x -so %s 2>/dev/null", path);
    else file = fopen (path, "r"), closefile = 1;
    if (!file) DIE ("can not open file '" << path << "'");
    std::cout << "c reading from '" << path << "'" << NL << std::flush;
  }

  parse_header (file);

  init_after_options_have_been_set ();
  init_after_maximum_variable_index_known ();

  parse_body (file);

  if (closefile == 1) fclose (file);
  if (closefile == 2) pclose (file);
  STOP (parse);
}

/*------------------------------------------------------------------------*/

// For binary clauses we do in general not keep references (clause indices)
// in the watcher lists.  An exception occurs if a clause is shrunken to a
// binary clause, and we do not want to go through the effort and explicity
// remove the clause reference in the watch list of the one watching literal
// (see for instance 'on_the_fly_hyper_binary_resolution').  Anyhow, if a
// binary conflict occurs we do not have the actual clause but only know its
// two literals.  In order to simplify down-stream handling of conflicts,
// e.g., interpret them as simple clauses we use a 'virtual' statically
// allocated binary clause as a replacement and update its literals every
// time a binary conflict occurs.  For reasons (antecedent clauses) the
// current literal is not needed and we can get away by just storing the
// other (antecedent) literal.  This function also logs the conflict and
// updates the important number of conflicts counter.

static inline void set_binary_conflict (int l, int o) {
  conflicts++;
  if (simpmode) simpconflicts++;
  assert (!conflict);
  conflict = &binary_conflict;
  conflict->lits[0] = l;
  conflict->lits[1] = o;
  conflictcidx = UINT_MAX;
  LOG ("binary conflict " <<  *conflict);
}

static inline void set_large_conflict (Cls * c, unsigned cidx) {
  assert (c);
  assert (!conflict);
  conflicts++;
  if (simpmode) simpconflicts++;
  conflict = c;
  conflictcidx = cidx;
  LOG ("large conflict " << *conflict);
}

/*------------------------------------------------------------------------*/

// This is the core and most important function in a CDCL SAT solver where
// most of its time is usually spent.  Our variant here uses blocking
// literals (from Chu and Stuckey) to avoid dereferencing clauses if the
// blocking literal turns out to be satisfied, then omits clauses references
// for binary clauses altogether (only using the blocking literal as
// suggested by Pilarski and Hu in DATE'02), and further implements the
// literal move to front idea from Lingeling.  Our version actually comes in
// two flavors.  One in which only binary clauses are propagated
// ('skip_large' is true) and the usual one in which all clauses are
// propagated at the same time. The former is used for non-transitive hyper
// binary resolution (see our CPAIOR'13 paper).

// TODO this is not perfect yet, since 'assign' with a large clause below
// should trigger aborting the propagation if 'skiplarge' is true.  It is
// also unclear whether this specialization to avoid producing transitive
// binary clauses is really necessary.

static void propagate_literal (const int lit, const bool skiplarge) {
  assert (val (lit) < 0);
  Wchs & ws = wchs (lit);
  unsigned i = 0, j = 0;
  while (j < ws.size ()) {
    int blit = ws[i++] = ws[j++];	// by default copy watch
    if (conflict) continue;		// copy remaining watches
    const bool large = blit & 1;
    blit >>= 1;				// blocking literal 'blit'
    int u = val (blit);
    if (!large) {
      if (u > 0) continue;			// 'blit' satisfied
      if (!u) assign (blit, Rsn (lit));
      else set_binary_conflict (lit, blit);
    } else {
      const unsigned cidx = ws[i++] = ws[j++];	// copy clause index too
      if (skiplarge) continue;
      if (u > 0) continue;			// 'blit' satisfied
      Cls * c = clauses[cidx];
      if (c->garbage) continue;		// due to 'subsume_learned'
      int o = c->lits[1];		// other watched literal
      const unsigned p = (o == lit);	// position of 'lit'
      if (p) o = c->lits[0];
      else if (c->lits[0] != lit) {	// see 'on_the_fly_conflict'
	assert (otfsconflicts || otfsdrives || hbrsubs);
	log_unwatch (lit, blit, cidx);
	i -= 2;
	continue;
      }
      if (o != blit && (u = val (o)) > 0) {	// other watch true?
	ws[j-2] = sll (o) + 1;			// update 'blit'
	continue;
      }
      const unsigned t = c->size ();
      unsigned k;
      int r = lit;
      for (k = 2; k < t; k++) {
	std::swap (c->lits[k], r);	// literal-move-to-front (LMTF)
	if (val (r) >= 0) break;	// found replacement watch?
      }
      if (k < t) {
	c->lits[p] = r;			// replace 'lit' with 'r'
	watch_large (r, o, cidx);	// watch 'r' in 'c' now
	i -= 2;				// remove 'lit' watch
	log_unwatch (lit, blit, cidx);
      } else {
	while (k-- > 2) std::swap (c->lits[k], r);	// undo LMTF
	if (!u) assign (o, Rsn (c, cidx));
	else set_large_conflict (c, cidx);
      }
    }
  }
  ws.resize (i);
}

// It is faster to propagate them together of course, so this is default.

static void propagate_binary_and_large_clauses_together () {
  LOG ("propagating binary and large clauses together");
  while (!conflict && next < trail.size ()) {
    const int lit = -trail[next++];
    propagations++;
    LOG ("propagate " << -lit << " in binary and large clauses together");
    propagate_literal (lit, false);
  }
}

// On decision level 1, to avoid (as much as possible) deducing transitive
// hyper binary resolvents, we first propagate binary clauses until
// completion.  This is described in our CPAIOR'13 paper, but as discussed
// above, the current version does not achieve this completely.

static void propagate_binary_clauses_first () {
  LOG ("propagating binary clauses first");
  unsigned next2 = next;
  while (!conflict && next < trail.size ()) {
    if (next2 < trail.size ()) {
      propagations++;				// only count once
      const int lit = -trail[next2++];
      LOG ("propagate " << -lit << " in binary clauses");
      propagate_literal (lit, true);
    } else {
      const int lit = -trail[next++];
      LOG ("propagate " << -lit << " in large clauses");
      propagate_literal (lit, false);
    }
  }
}

static bool propagate () {
  assert (!conflict);
  assert (!inconsistent);
  START (propagate);
  if (level == 1 && opts::hbr) propagate_binary_clauses_first ();
  else propagate_binary_and_large_clauses_together ();
  STOP (propagate);
  return !conflict;
}

/*------------------------------------------------------------------------*/

// Backtrack or -jump to decision level 'jump'.  The second argument avoids
// a 'dequeue' and 'enqueue' operation for a literal which immediately after
// backtracking is known to be anyhow assigned again (as the 1st UIP).

static void backtrack (unsigned jump = 0, int except = 0) {
  if (jump == level) return;
  assert (jump < level);
  except = abs (except);
  LOG ("backtrack " << jump);
  unsigned i = trail.size ();
  int lit;
  while (i > 0 && var (lit = trail[i-1]).level > jump)
    unassign (lit, except), i--;
  trail.resize (i);
  if (next > i) next = i;
  level = jump;
  conflict = 0;
}

// The flag 'SEEN' means reached during conflict analysis (and clause
// minimization). This is MiniSAT terminology.

static void mark_seen (int lit) {
  Var & v = var (lit);
  assert (!v.tag);
  v.tag = SEEN;
  tagged.push_back (lit);
  LOG ("seen " << lit << " @ " << v.level);
}

// The 'poison idea' from Allen Van Gelder is used to cache the search for
// literals which can not be removed during clause minimization.  Without it
// clause minimization risks to be quadratic in the implication graph.

static void mark_poison (int lit) {
  Var & v = var (lit);
  assert (v.tag != POISON);
  if (v.tag != SEEN) tagged.push_back (lit);
  v.tag = POISON;
  LOG ("poison " << lit << " @ " << v.level);
}

// A key ingredient of modern solvers could be summarized as 'relevance based
// reasoning'.  The heuristic is to mark more recently used variables as
// more relevant.  This marking process is also called 'bumping' such a used
// variable.  In our VMTF version of this idea this amounts to checking
// whether a variable has been bumped more recently, which is stored in its
// 'stamp'. Here we get stable sorting for free, since stamps are unique.

struct bump_less_than {
  bool operator () (int a, int b) const {
    assert (a == b || lnk (a).stamp != lnk (b).stamp);
    return lnk (a).stamp < lnk (b).stamp;
  }
};

// This implements the actual 'variable move to front' (VMTF) strategy as
// originally proposed for the Siege SAT solver and then has been carefully
// extended and reengineered as described in our SAT'15 paper using stamps
// and applying VMTF to all bumped vars.

static void bump (int lit) {
  int idx = abs (lit);
  LOG ("bump " << idx);
  queue.dequeue (idx);
  queue.enqueue (idx);
}

// Clauses are 'bumped' too as soon they are resolved, which is a common
// indicator that a redundant / learned clause is important and should
// be kept and not thrown away during garbage collection of learned clauses
// in 'reduce'.

static void bump (Cls * c) {
  if (!c->redundant) return;
  Red * r = cls2red (c);

  // We want to focus on most recently resolved / bumped clauses.  The
  // 'resolved' field keeps the last time a clause was resolved (which is as
  // described above a 32 bit approximation of the 'conflicts' counter).

  r->resolved = resolved;

  // We also want to make sure that clauses used at least once during the
  // learning phase after the last reduction are kept and thus we mark them.

  r->active = 1;

  // These are moving averages of the glue and size of resolved clauses.

  ema.resolved.glue.update (r->glue);
  ema.resolved.size.update (c->size ());
}

// It is cheaper to sort 'Red *' pointers instead of sorting indices.  By
// going over the indirection of indices we first use less 'work' space,
// which probably is a minor advantage, but also get stable sorting, thus
// become deterministically portable (indepedent of the actual 'sort'
// implementation).

struct resolved_less_than {
  bool operator () (const unsigned & i, const unsigned & j) const {
    Red * r = cls2red (clauses[i]);
    Red * s = cls2red (clauses[j]);
    if (r->resolved < s->resolved) return true;
    if (r->resolved > s->resolved) return false;
    return i < j;
  }
};

// Unfortunately the 32 bit abstraction 'resolved' of 'conflicts' used in
// 'Red.resolved' might overflow (after billions of conflicts).  In this
// case we shrink the range of 'Red.cls' values maintaining the relative
// order between them.  One could get rid of this complexity by either (A)
// ignoring overflow or (B) using a complete 64 bit (8 byte) 'conflicts'
// counter in each learned clause in the 'Red' header.  Clearly (B) is
// wasteful and to commit to (A) incoporates the risk to produce strange
// behavior of the solver after a long run-time, which then is very hard to
// debug.  So we went to this additional effort and implemented 'rescoring'
// of 'resolved'.  In order to test it through fuzzing we also use a limit
// set by the option 'opts::resolvedlim', which triggers during fuzzing
// rescoring way earlier before an actual (32-bit) overflow occurs.

static void rescore_resolved () {
  assert (resolved >= limits.resolved);
  LOG ("bump resolved");
  assert (work.empty ());
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (!c->redundant) continue;
    work.push_back (i);
  }
  std::sort (work.begin (), work.end (), resolved_less_than ());
  resolved = 1;
  unsigned prev = 0;
  for (unsigned i = 0; i < work.size (); i++) {
    Red * r = cls2red (clauses[work[i]]);
    const unsigned cur = r->resolved;
    assert (!i || prev <= cur);
    if (!i || prev != cur) resolved++;
    r->resolved = resolved;
    prev = cur;
  }
  release (work);
  resolved++;

  // See comments to 'limits.queuestamp' in 'restamp_queue'.
  // The same arguments apply here.

  while (limits.resolved < (1u<<31) &&
         limits.resolved/2 <= resolved) limits.resolved <<= 1;

  report ('S');
}

/*------------------------------------------------------------------------*/

// Functions to reset flags and stacks used in 'analyze'.

static void reset_tagged () {
  LOG ("reset tagged");
  while (!tagged.empty ())
    var (tagged.back ()).tag = 0, tagged.pop_back ();
}

static void reset_levels () {
  LOG ("reset levels");
  while (!levels.empty ()) {
    Frame & f = control[levels.back ()];
    f.pulled = 0;
    levels.pop_back ();
  }
}

static void reset_analyze () {
  reset_levels ();
  reset_tagged ();
  LOG ("reset clause");
  clause.clear ();
  RESOLVE (resolvent.clear ());
}

/*------------------------------------------------------------------------*/

// The following sections contain a variant of the classical core algorithm
// called 'analyze' for generating a learned clause from a conflict, which
// was pioneered by Joao Marques-Silva and Karem Sakallah in GRASP. As usual
// generating the first UIP clause is best as first explicitly observed by
// Lintao Zhang et.al.  We add some more recent ideas further down in the
// actual 'analyze' function and explain their origin there.

static void check_flags_unset () {
  assert (clause.empty ());
  assert (tagged.empty ());
  assert (levels.empty ());
  RESOLVE (assert (resolvent.empty ()));
#ifdef EXPENSIVECHECKS
  if (!opts::check) return;
  for (int i = 1; i <= maxvar; i++) assert (!var (i).tag);
  for (int i = 1; i <= maxvar; i++) assert (!marked_lit (i));
  for (unsigned l = 0; l < control.size (); l++) assert (!control[l].pulled);
#endif
}

static void resolve_lit (int lit, unsigned & open, unsigned & jump) {
  Var & v = var (lit);
  if (v.tag) return;			// ignore already resolved lits
  assert (val (lit) < 0);		// other resolved lits are false
  unsigned tmp = v.level;
  if (!tmp) return;			// ignore root assigned lits
  mark_seen (lit);			// mark as seen aka resolved
  RESOLVE (resolvent.push_back (lit));
  if (tmp < level) {
    clause.push_back (lit);
    LOG ("learning " << lit);
    if (!control[tmp].pulled++) {	// seen this decision level?
      LOG ("pulling " << tmp);		// new level found!
      levels.push_back (tmp);
      if (tmp > jump) jump = tmp;	// update 'back jump level'?
    }
  } else assert (tmp == level), open++;
}

static void resolve_clause (Cls * c, unsigned & open, unsigned & jump) {
  const unsigned size = c->size ();
  for (unsigned i = 0; i < size; i++)
    resolve_lit (c->lits[i], open, jump);
  bump (c);					// bump resolved clauses
}

// A variant of what was proposed by Han & Somenzi.  We only implemented the
// version where only the reason clause is shrunken (so no test for double
// self-subsumption with the original conflicting clause).

static int on_the_fly_conflict (int lit, int t) {
  assert (t != lit);
  assert (val (t) < 0), assert (var (t).level == level);
  assert (val (lit) > 0), assert (var (lit).level == level);
  const Rsn & r = var (lit).reason;
  assert (r.clausal ());
  Cls * c = r.cls ();
  int o = c->lits[1];
  const unsigned p = (o == lit);
  if (p) o = c->lits[0];
  assert (o != t);
  assert (c->lits[p] == lit);
  assert (val (o) < 0), assert (var (o).level == level);
  const unsigned cidx = r.cidx ();
  otfsconflicts++;
  LOG ("producing on-the-fly conflict by removing " << lit <<
    " from reason clause " << *c);
  copy (unshrunken, *c);
  unsigned new_size = 2;
  c->lits[p] = t;				// replace 'lit' by 't'
  const unsigned old_size = c->size ();
  bool found = false;
  for (unsigned i = 2; i < old_size; i++) {
    int o = c->lits[i];
    if (o == t) { found = true; continue; }
    if (is_fixed_lit (o)) continue;			// ignore roots
    c->lits[new_size++] = o;
  }
  assert (found), (void) found;
  c = realloc_clause (c, old_size, new_size);
  clauses[cidx] = c;
  // Now we have three watches for this code, but this is not harmful after
  // checking that a claimed watched literal during propagate is still one
  // of the first two literals.  However, not fixing the 'blit' of 'o' is
  // really harmfull and breaks lots of invariants, more precisely that
  // propagation is always run to completion and finds all unit clauses.
  // Without that invariant many things breaks (during analyze and during
  // garabge collection for instance, and fixing them turned out to be really
  // hard).  So in essence this means there is a hidden cost of applying
  // on-the-fly subsumption, which is updating the blit here.
  watch_large (t, o, cidx);
  update_large_blit (o, cidx, t);	// might be costly
  trigger_shrunken_or_rewritten (c);
  trace_add_learned_clause (*c);		// again as in
  trace_delete_clause (unshrunken);		// 'add_original_clause'
  reset_analyze ();
  conflict = c;					// no 'conflicts++'!
  conflictcidx = cidx;
  LOG ("new on-the-fly conflict " << *conflict);
  return -1;					// -1 = on-the-fly-conflict
}

static int on_the_fly_driving (int lit, unsigned jump) {
  assert (0 < jump), assert (jump < level);
  assert (val (lit) > 0), assert (var (lit).level == level);
  const Rsn & r = var (lit).reason;
  assert (r.clausal ());
  Cls * c = r.cls ();
  int o = c->lits[1];
  const unsigned p = (o == lit);
  if (p) o = c->lits[0];
  assert (c->lits[p] == lit);
  assert (val (o) < 0), assert (var (o).level == level);
  const unsigned cidx = r.cidx ();
  otfsdrives++;
  LOG ("producing on-the-fly driving clause by removing " << lit <<
    " from reason clause " << *c);
  copy (unshrunken, *c);
  unsigned new_size = 2;
  const unsigned old_size = c->size ();
  int t = 0;
  for (unsigned i = 2; i < old_size; i++) {
    int s = c->lits[i];
    if (is_fixed_lit (s)) continue;		// ignore roots
    if (!t && var (s).level == jump) t = s;	// find & skip 't'
    else c->lits[new_size++] = s;
  }
  assert (t);
  assert (o != t);
  c->lits[p] = t;				// replace 'lit' by 't'
  c = realloc_clause (c, old_size, new_size);
  trigger_shrunken_or_rewritten (c);
  clauses[cidx] = c;
  watch_large (t, o, cidx); 		// see also 'on_the_fly_conflict'
  update_large_blit (o, cidx, t);	// might be costly
  trace_add_learned_clause (*c);	// again as in
  trace_delete_clause (unshrunken);	// 'add_original_clause'
  reset_analyze ();
  backtrack (jump, o);
  LOG ("on-the-fly assignment of " << o << " reason " << *c);
  assign (o, Rsn (c, cidx));
  return 1;					// 1 = on-the-fly-driving
}

// Here is the actual code for on-the-fly subsumption.  If the return value
// is negative then the shrunken reason clause became a new conflict
// (because there were two other false literals on the current decision
// level).  This is the case handled by 'on_the_fly_conflict' above.  If the
// return value is positive the shrunken reason clause has only one other
// literal beside the removed one, actual the other watched literal on the
// current decision level.  Thus we backtrack and assign that one, which is
// handled in 'on_the_fly_driving'.  Otherwise if no on-the-fly subsumption
// is possible zero is returned.

static int on_the_fly_subsumption (int lit, unsigned open) {
  assert (level > 0);
  if (!opts::otfs) return 0;
  const unsigned resolvent_size = open + clause.size ();
  if (resolvent_size < 2) return 0;
  const Rsn & r = var (lit).reason;
  if (!r.clausal ()) return 0;
  Cls * c = r.cls ();
  const unsigned old_size = c->size ();
  unsigned flushed_size = 0;
  for (unsigned i = 0; i < old_size; i++)
    if (!is_fixed_lit (c->lits[i])) flushed_size++;
  if (flushed_size < 2) return 0;
  if (resolvent_size >= flushed_size) return 0;
  unsigned k, jump = 0;
  int t = 0;
  for (k = 2; k < c->size (); k++) {
    unsigned l = var (t = c->lits[k]).level;
    if (l == level) break;
    assert (l < level);
    if (l > jump) jump = l;
  }
  if (k < c->size ()) return on_the_fly_conflict (lit, t);
  else                return on_the_fly_driving (lit, jump);
}

// Resolve in a reason of an assigned literal 'lit'.  In the literature
// this is also often called 'antecedent (clause)' for 'lit'.

static void resolve_reason (int lit, unsigned & open, unsigned & jump) {
  assert (val (lit) > 0);
  const Rsn & r = var (lit).reason;
  assert (!r.decision ());
  LOG ("resolving " << lit << ' ' << r);
  if (r.binary ()) resolve_lit (r.lit (), open, jump);
  else resolve_clause (r.cls (), open, jump);
  update_and_log_resolvent (lit);
  LOG ("open " << open << " jump " << jump);
}

static void resolve_conflict (unsigned & open, unsigned & jump) {
  LOG ("resolving conflicting clause " << *conflict);
  resolve_clause (conflict, open, jump);
  log_initial_resolvent ();
  assert (open > 1);
}

// Here comes the procedure to determine the first UIP clause.

static int deduce_first_uip_clause (unsigned & jump, unsigned & glue) {
RESTART:
  check_flags_unset ();
  unsigned i = trail.size (), open = 0;
  int uip = 0;
  jump = 0;
  resolve_conflict (open, jump);
  for (;;) {
    do uip = trail[--i]; while (!var (uip).tag);
    if (!--open) break;
    resolve_reason (uip, open, jump);
    int otfs = on_the_fly_subsumption (uip, open);
    if (otfs < 0) goto RESTART;				// new conflict
    if (otfs > 0) return 0;				// new assignment
    conflict = 0;
  }
  clause.push_back (-uip);
  LOG ("uip " << uip);
  assert (levels.size () < UINT_MAX);	// since 'maxvar < UINT_MAX'
  glue = levels.size ();
  return uip;
}

/*------------------------------------------------------------------------*/

// See the POS'15 paper on a variant of Glucose restart scheduling.

static void update_glue_emas (unsigned glue) {
  ema.glue.fast.update (glue);
  ema.glue.slow.update (glue);
  LOG ("glue " << glue << " ema fast " << ema.glue.fast.norm ());
  LOG ("glue " << glue << " ema slow " << ema.glue.slow.norm ());
}

static void update_jump_ema (unsigned jump) {
  ema.jump.update (jump);
  LOG ("jump " << jump << " ema " << ema.jump.norm ());
}

static void update_trail_ema () {
  int t = remaining_assigned_variables ();
  ema.trail.fast.update (t);
  ema.trail.slow.update (t);
  LOG ("trail " << t << " ema slow " << ema.trail.slow.norm ());
  LOG ("trail " << t << " ema fast " << ema.trail.fast.norm ());
}

// Update the learned 'glue', 'jump' level, and 'trail' height EMAs, which
// give the averages of these metrics at the point a conflict occurs.

static void update_emas (unsigned jump, unsigned glue) {
  if (simpmode) return;
  update_glue_emas (glue);
  update_jump_ema (jump);
  update_trail_ema ();
}

static void bump_seen () {

  // It is crucual to keep the old order among bumped indices intact.  Not
  // sorting will have a high negative impact on the effectiveness of VMTF.
  // See again our SAT'15 paper on this and related issues.

  // TODO check efficiency of the 'std::sort' implementation here. We are
  // sorting w.r.t. unique stamps of type 'unsigned'.  It might be faster to
  // use a radix sort variant.

  std::sort (tagged.begin (), tagged.end (), bump_less_than ());
  for (unsigned i = 0; i < tagged.size (); i++) bump (tagged[i]);
  if (++resolved >= limits.resolved) rescore_resolved ();
}

/*------------------------------------------------------------------------*/

// Clause minimization through subsuming resolution was originally proposed
// by Niklas Soerensson in MiniSAT.  Our optimized version uses the 'poison'
// idea of Allen Van Gelder (the original MiniSAT implementation was
// quadratic) and the 'pulled twice' idea by Donald Knuth explained below.

static bool minimize_prefix (int other) {
  assert (val (other) < 0);
  const Var & v = var (other);
  if (!v.level) return true;		// ignore roots
  if (v.tag == SEEN) return true;	// check first in clause or cached
  if (v.decision ()) return false;	// then negative base case
  if (v.tag == POISON) return false;	// or cached negative
  if (!control[v.level].pulled) return false;	// level not pulled?
  work.push_back (other);		// otherwise recurse
  return true;
}

// This is a drecursified DFS with early abort.  If at one point it is
// clear that the root can not be minimized, since recursively it depends
// on something that will not be implied by the literals already in the
// first UIP clause, then we abort.  We want to 'cache' as much work as
// possible, i.e., even in this case we do not simply abort, but update
// 'POISON' and 'SEEN' flags while going up the recursion stack (see
// the 'litpoison' and 'litseen' checks which trigger 'mark_poison' and
// 'mark_seen' in the postfix work below).

static bool minimize (int root) {
  bool res = true;
  work.push_back (root);
  while (!work.empty ()) {
    int lit = work.back ();
    work.pop_back ();
    if (lit) {				// prefix work on 'lit'
      assert (val (lit) < 0);
      if (!res) continue;		// early abort, so only backtrack
      const Var & v = var (lit);
      if (v.tag == POISON) continue;			// negative cached
      if (v.tag == SEEN && lit != root) continue;	// positive cached
      assert (lit == root || !v.tag);			// not seen or root
      const Rsn & r = v.reason;
      const unsigned before = work.size ();  // before children are pushed
      if (r.binary ())
	res = minimize_prefix (v.reason.lit ());	// one child
      else {
	Cls * c = r.cls ();
	assert (c);
	const unsigned size = c->size ();
	for (unsigned i = 0; res && i < size; i++) {
	  int other = c->lits[i];
	  if (other == -lit) continue;
	  if (!minimize_prefix (other)) res = false;
	}
      }
      if (res) {
	unsigned i = work.size ();
	work.push_back (before);	// save number of unseen children
	work.push_back (lit);		// of this current literal 'lit'
	work.push_back (0);		// mark for postfix work on 'lit'
	while (i-- > before) {	// copy and then recurse on unseen children
	  int other = work[i];
	  work.push_back (other);
	}
      } else {
	work.resize (before);			// ignore children
	if (lit != root) mark_poison (lit);	// mark negative cached
      }
    } else {					// postfix work on 'lit'
      assert (work.size () >= 2);
      lit = work.back ();
      work.pop_back ();
      assert (lit == root || !var (lit).tag);
      const unsigned before = work.back ();	// start of children
      work.pop_back ();
      bool litpoison = false, litseen = true;
      while (work.size () > before) {
	const int other = work.back ();
	work.pop_back ();
	assert (val (other) < 0);
	if (litpoison) continue;
	const Var & u = var (other);
	if (u.tag == POISON) litpoison = true;
	else if (u.tag != SEEN) litseen = false;
      }
      if (lit == root) assert (work.empty ());	// root is already seen
      else if (litpoison) mark_poison (lit);	// one child negative
      else if (litseen) mark_seen (lit);	// all children positive
    }
  }
  return res;
}

static void minimize_clause (unsigned glue) {
  if (!opts::minimize) return;
  const unsigned orig_size = clause.size ();
  assert (orig_size > glue);
  if (orig_size <= 2) return;
  if (orig_size == glue + 1) return;
  START (minimize);
  unsigned j = 0;
  for (unsigned i = 0; i < clause.size (); i++) {
    int lit = clause[i];
    clause[j++] = lit;
    const Var & v = var (lit);
    if (v.decision ()) continue;	// decisions can not be minimized
    unsigned tmp = v.level;
    if (tmp == level) continue;			// UIP can not be minimized
    assert (0 < tmp), assert (tmp < level);
    assert (control[tmp].pulled > 0);

    // Originally a 'Bloom' filter aka signatures was used here (for
    // instance in MiniSAT still is).  Donald Knuth observed that a literal
    // on a decision level which is only pulled in once can never be
    // minimized away.  In some of our earlier solvers we already used just
    // one bit instead of a counter, so only checked, whether a level was
    // pulled in at least once.  This 'pulled > 0' condition already avoids
    // signatures but Don's 'pulled > 1' check clearly gives more early
    // aborts of the search but needs one more bit per variable at least.
    // We actually simply (but wastefully) use the 32 bit counter 'pulled'
    // instead of two bits.

    if (control[tmp].pulled == 1) continue;
    if (!minimize (lit)) continue;
    LOG ("minimized " << lit);
    minimizedlits++;
    j--;
  }
  clause.resize (j);
  STOP (minimize);
}

/*------------------------------------------------------------------------*/

// As observed by Donald Knuth, there might be situations where a decision
// only clause is much smaller than the minimized first UIP clause.  If the
// glue of the minimized clause is sufficiently large (opts::decoglue), the
// size of the decision only clause is smaller than a fraction of the
// minimized clause ('opts::decsizered' in percent), then the minimized
// first UIP clause is scrapped and replaced by the decision only clause.
// We conservatively include all levels in this decision only clause.  Only
// including the 'pulled' levels is not sound and doing another costly
// conflict analysis did not give benefits (only checked with Lingeling).

// We added the same optimization as in Lingeling though, which uses the
// information that certain levels were only pulled in once. Let 'm' the
// the maximum decision level, which was pulled in multiple times. Between
// that 'm' and 'jump' we can keep all literals and even do not have to
// include the 'decision' but simply use the pulled in literal in 'clause'.

static void
decision_only_clause (int uip, unsigned jump, unsigned & glue) {
  if (!jump) return;
  if (!opts::deco) return;
  if (glue < (unsigned) opts::decoglue) return;

  // All levels up to the 'jump' level plus the current level make up the
  // new glue of the decision only clause, which is also its size.

  const int64_t new_size = jump + 1;
  const int64_t old_size = clause.size ();

  if (new_size * 100 > old_size * opts::decolim) return;

  LOG ("discarding size " << clause.size ()
     << " glue " << glue << " minimized clause in favor of "
     << "maximum size " << new_size << "decision only clause");

  // Get the largest multiple times pulled level 'm'.

  unsigned m = 0;
  for (unsigned i = 0; i < levels.size (); i++) {
    unsigned l = levels[i];
    assert (0 < l), assert (l <= jump);
    if (control[l].pulled > 1 && m < l) m = l;
  }
  LOG ("largest multiple times pulled level " << m);

  // Keep the originally pulled in literals above 'm'.

  unsigned j = 0;
  if (m < jump)
    for (unsigned i = 0; i < clause.size (); i++) {
      int lit = clause[i];
      if (lit == -uip) continue;		// push later
      unsigned l =  var (lit).level;
      if (l > m) clause[j++] = lit;
    }
  clause.resize (j);
  LOG ("keeping " << j << " pulled once level literals");

  // Use decisions for all levels up-to and including 'm'.  We have to
  // include all levels and not just the pulled ones (in 'levels').

  for (unsigned i = m; i; i--)
    clause.push_back (-control[i].decision);	

  clause.push_back (-uip);

  decos++;
  glue = clause.size ();
  assert (glue <= new_size);
  LOG ("actual glue and size " << glue
    << " decision only clause " << clause);
}

/*------------------------------------------------------------------------*/

// Assume we have a clause 'c' of size 'needed' with all its literals
// marked.  Then this function checks whether another clause 'd' is
// subsumed by 'c'.  We keep the literals of 'c' marked, since it is used
// multiple times to find subsumed clauses.

static bool can_subsume_clause (const Cls * d, const unsigned needed) {
  const unsigned size = d->size ();
  if (size < needed) return false;	   	// 'd' smaller than 'c''
  unsigned found = 0, diff = size - needed, i = 0;
  for (;;) {
    int mark = marked_lit ((*d)[i++]);
    if (mark < 0) return false;		// contains opposite literal
    else if (mark > 0) { if (++found == needed) return true; }
    else if (!diff--) return false;	// not enough literals left
  }
}

// Check whether one of the last 'opts::subsumerecent' learned clauses in
// [cidx - opts::subsumerecent ... cidx - 1] is subsumed by the new clause
// 'c' with index 'cidx'.  Collecting such a subsumed clause is delayed
// until the next 'garbage_collection' or 'reduce' (and thus needs some care
// there).  This idea goes back to Donald Knuth and is also implemented
// in Lingeling.

static void subsume_recently_learned (Cls * c, unsigned cidx) {
  assert (clauses[cidx] == c);
  assert (c->redundant);
  unsigned count = opts::subsumerecent;
  if (!count) return;
  START (subrecent);
  mark_clause (*c);
  const unsigned size = c->size ();
  assert (size > 1);
  unsigned i = cidx;
  while (i-- && count--) {
    Cls * d = clauses[i];
    assert (d != c);
    if (d->garbage) continue;
    if (!d->redundant) continue;
    if (!can_subsume_clause (d, size)) continue;
    LOG ("subsumed learned clause " << *d);
    trace_delete_clause (*d);
    assert (!d->garbage);
    assert (!d->traced);
    d->garbage = true;
    d->traced = true;
    subrecent++;
  }
  unmark_clause (*c);
  STOP (subrecent);
}

/*------------------------------------------------------------------------*/

// There are three cases for learned clauses.  Either the learned clause is
// empty and the formula becomes inconsistent, or the learned clause is unit
// and thus we do not have to add it at all, or it is an at least binary
// clause.  The first case occurs separately and thus we factored it out
// into one function.

static void learn_empty_clause () {
  assert (!level);
  assert (!inconsistent);
  trace_add_empty_learned_clause ();
  LOG ("learned empty clause");
  inconsistent = true;
}

static void learn_unit_clause (int unit) {
  LOG ("learned unit " << unit);
  assign (unit, Unit ());
  trace_add_unit_learned_clause (unit);
}

static void learn_additional_binary_clause (int a, int b) {
  LOG ("learned additional binary clause");
  clause.clear ();
  clause.push_back (a);
  clause.push_back (b);
  unsigned cidx;
  (void) add_redundant_clause (0, cidx);
  (void) cidx;
  trace_add_binary_learned_clause (a, b);
  clause.clear ();
}

static void learn_first_uip_clause (int uip, unsigned glue) {
  if (clause.size () == 1) {
    if (!simpmode) iterating = true;	// see 'iteration' below
    learn_unit_clause (-uip);
  } else {
    assert (clause.size () > 1);
    if (clause.size () == 2) {
      LOG ("learned binary clause");
      if (!simpmode) learnedbins++;
    }
    unsigned cidx;
    Cls * c = add_redundant_clause (glue, cidx);
    trace_add_learned_clause (clause);
    subsume_recently_learned (c, cidx);
    assign (-uip, Rsn (c, cidx));
  }
}

/*------------------------------------------------------------------------*/

// This is an attempt to learn more clauses than the (minimized) first UIP
// clause.  For now we focus on additional root level units in the case the
// first UIP clause is a unit clause and on adding hyper binary resolvents
// 'in front' of the first UIP.  The idea is that these additional binary
// clauses will trigger additional propagation after the value of the first
// UIP clause is flipped, which would not happen without them.

static bool resolve_additional_lit (int lit, unsigned & open) {
  Var & v = var (lit);
  unsigned tmp = v.level;
  if (!tmp) return true;		// ignore root assigned lits
  if (tmp < level) {
    LOG ("antecedent literal " << lit << " on level " << tmp);
    return false;
  }
  if (v.tag) return true;		// ignore already resolved lits
  assert (val (lit) < 0);		// other resolved lits are false
  mark_seen (lit);			// mark as seen aka resolved
  assert (tmp == level);
  open++;
  return true;
}

static bool resolve_additional_clause (Cls * c, unsigned & open) {
  const unsigned size = c->size ();
  for (unsigned i = 0; i < size; i++)
    if (!resolve_additional_lit (c->lits[i], open)) return false;
  return true;
}

static bool resolve_additional_reason (int lit, unsigned & open) {
  assert (val (lit) > 0);
  const Rsn & r = var (lit).reason;
  if (r.decision ()) return false;
  LOG ("resolving " << lit << ' ' << r);
  bool res;
  if (r.binary ()) res = resolve_additional_lit (r.lit (), open);
  else res = resolve_additional_clause (r.cls (), open);
  LOG ("open " << open);
  return res;
}

static void deduce_additional_units (int first_uip) {
  LOG ("trying to deduce additional unit clauses");
  assert (units.empty ());
  int uip = first_uip;
  unsigned i = trail.size (), open = 0, maxopen = 0;
  while (trail[--i] != uip)
    ;
  for (;;) {
    if (!resolve_additional_reason (uip, open)) break;
    if (open > maxopen) maxopen = open;
    do uip = trail[--i]; while (!var (uip).tag);
    if (--open) continue;
    if (maxopen > 1) {
      LOG ("deduced additional unit " << -uip);
      units.push_back (-uip);
      additionalunits++;
    } else LOG ("skipping additional unit " << -uip);
    maxopen = 0;
  }
  LOG ("deduced " << units.size () << " additional units");
}

static void deduce_additional_binary_clauses (int first_uip) {
  LOG ("trying to deduce additional binary clauses");
  assert (bins.empty ());
  int uip = first_uip, last = uip;
  unsigned i = trail.size (), open = 0, maxopen = 0;
  while (trail[--i] != uip)
    ;
  for (;;) {
    if (!resolve_additional_reason (uip, open)) break;
    if (open > maxopen) maxopen = open;
    do uip = trail[--i]; while (!var (uip).tag);
    if (--open) continue;
    if (maxopen > 1) {
      LOG ("deduced additional binary clause " << -uip << ' ' << last);
      bins.push_back (-uip);
      bins.push_back (last);
      additionalbins++;
    } else LOG ("skipping binary clause " << -uip << ' ' << last);
    maxopen = 0;
    last = uip;
  }
  LOG ("deduced " << bins.size ()/2 << " additional binary clauses");
}

static void deduce_additional_clauses (int first_uip,
                                       unsigned jump, unsigned glue) {
  if (!opts::alluip) return;
  if (!jump) assert (!glue), deduce_additional_units (first_uip);
  else if (opts::alluip > 1) deduce_additional_binary_clauses (first_uip);
}

static void learn_additional_units () {
  LOG ("learning " << units.size () << " additional unit clauses");
  for (unsigned i = 0; i < units.size (); i++)
    learn_unit_clause (units[i]);
  units.clear ();
}

static void learn_additional_binary_clauses () {
  LOG ("learning " << bins.size ()/2 << " additional binary clauses");
  for (unsigned i = 0; i < bins.size (); i += 2)
    learn_additional_binary_clause (bins[i], bins[i+1]);
  bins.clear ();
}

static void learn_additional_clauses () {
  if (!units.empty ()) learn_additional_units ();
  if (!bins.empty ()) learn_additional_binary_clauses ();
}

/*------------------------------------------------------------------------*/

// Finally, 'analyze' for learning a clause (the 'L' in 'CDCL').

static void analyze () {
  if (!level) { learn_empty_clause (); return; }
  START (analyze);
  unsigned jump, glue;
  int first_uip = deduce_first_uip_clause (jump, glue);
  if (first_uip) {
    bump_seen ();
    minimize_clause (glue);
    update_emas (jump, glue);
    decision_only_clause (first_uip, jump, glue);
    learnedlits += clause.size ();
    deduce_additional_clauses (first_uip, jump, glue);
    backtrack (jump, first_uip);
    learn_first_uip_clause (first_uip, glue);
    learn_additional_clauses ();
  } else {
    // Nothing to learn, since 'on-the-fly-subsumption' is driving an
    // assignment through a shortened clause and no clause is learned.
  }
  reset_analyze ();
  STOP (analyze);
}

/*------------------------------------------------------------------------*/

// Learning a unit clause is lot of progress.  We call this 'iteration'
// inspired by a similar output message in the 'BerkMin' solver (even
// though it is not clear the same was meant back then).  We delay reporting
// such an iteration until propagation completed, because we want to see
// how many variables are actually fixed due to this learned unit clause,
// for instance through 'learn_additional_units' or just because there are
// other implications in the formula implying additional units.

static void iteration () {
  assert (iterating);
  iterating = false;
  iterations++;
  report ('i');
}

/*------------------------------------------------------------------------*/

// On the root-level we might need to flush the root trail.

static void propagate_root_level () {
  assert (!level);
  if (inconsistent) return;
  if (next < trail.size ()) {
    LOG ("root level propagation");
    propagate ();
    if (conflict) learn_empty_clause ();
  }
  if (!inconsistent && trail.size ()) flush_root_trail ();
}

static void disconnect () {
  LOG ("disconnect");
  for (int lit = -maxvar; lit <= maxvar; lit ++)
    if (lit) wchs (lit).clear ();
}

static void flush_clause (Cls * c) {
  trace_delete_clause (*c);
  dealloc_clause (c);
  flushedclauses++;
}

// Remove satisfied and garbage clauses and remove falsified literals.
// Return 'true' if a unit clause was found.  Unit clauses might be found
// since we can not assume that propagation was run until completion for all
// clauses.  For instance during sweeping we disconnect clauses outside the
// blocked set.  Then performing garbage collection after some more back
// bones are found during sweeping of the blocked set we have to garbage
// collect and connect all the other clauses back.

static unsigned collect_clauses_flush_literals () {
  unsigned j = 0, new_units = 0;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (inconsistent) { clauses[j++] = c; continue; }

    unsigned satisfied = c->garbage, falsified = 0;
    const unsigned old_size = c->size ();

    // First test whether this clause contains an assigned literal.

    for (unsigned k = 0; !satisfied && k < old_size; k++) {
      Val tmp = val (c->lits[k]);
      if (tmp > 0) satisfied++;
      if (tmp < 0) falsified++;
    }

    if (satisfied) { flush_clause (c); continue; }

    if (falsified) {

      copy (unshrunken, *c);

      // This loop does the actual shrinking. As in 'add_original_clause' we
      // can not easily merge it with the one above while proof tracing.

      unsigned new_size = 0;
      for (unsigned k = 0; k < old_size; k++) {
	const int lit = c->lits[k];
	if (!val (lit)) c->lits[new_size++] = lit;
	else assert (val (lit) < 0);
      }
      assert (old_size == new_size + falsified);
      flushedlits += falsified;

      if (!new_size) {
	LOG ("found empty clause during garbage collection");
	assert (!inconsistent);
	inconsistent = true;
	trace_add_empty_learned_clause ();
	dealloc_clause (c);
	flushedclauses++;
	c = 0;
      } else if (new_size == 1) {
	assert (!level);
	int unit = c->lits[0];
	LOG ("found unit clause " << unit << " during garbage collection");
	assign (unit, Unit ());
	trace_add_unit_learned_clause (unit);
	dealloc_clause (c);
	flushedclauses++;
	new_units++;
	c = 0;
      } else {
	assert (new_size >= 2);
	c = realloc_clause (c, old_size, new_size);
	trace_add_learned_clause (*c);
      }

      trace_delete_clause (unshrunken);
    }

    if (c) clauses[j++] = c;
  }
  clauses.resize (j);
  return new_units;
}

/*------------------------------------------------------------------------*/

// Here comes the arena based compact memory allocation strategy.  See the
// larger discussion at the declaration of 'arena' above for more details.

static void check_no_wch_in_arena () {
#if !defined(NDEBUG) && !defined(VECTOR)
  for (int idx = 1; idx <= maxvar; idx++) {
    for (int sign = -1; sign <= 1; sign += 2) {
      int lit = sign * idx;
      Wchs & ws = wchs (lit);
      assert (!is_in_arena (ws.begin ()));
    }
  }
#endif
}

static void check_no_cls_in_arena () {
#ifndef NDEBUG
  for (unsigned i = 0; i < clauses.size (); i++)
    assert (!is_in_arena (clauses[i]));
#endif
}

static void check_all_vars_unmarked () {
#ifndef NDEBUG
  for (int i = 1; i <= maxvar; i++) assert (!marks [i]);
#endif
}

static Cls *
copy_and_move_clause (Cls * c, char * start, char * & top, char * end) {
  char * p = (char*) c;
  if (start <= p && p < top) return c;	// Do not copy again ...
  size_t bytes = c->bytes ();
  assert (top + bytes <= end), (void) end;
  if (c->redundant) {
    Red * s = cls2red (c);
    Red * r = (Red*) top;
    memcpy (r, s, bytes);
    deallocate (s, bytes);
    c = &r->cls;
  } else {
    Cls * d = c;
    c = (Cls*) top;
    memcpy (c, d, bytes);
    deallocate (d, bytes);
  }
  top += bytes;
  return c;
}

static void mark_signed_lit (int lit) {
  marks[abs (lit)] |= (lit < 0) ? 1 : 2;
}

static bool marked_signed_lit (int lit) {
  return (marks[abs (lit)] & ((lit < 0) ? 1 : 2)) != 0;
}

static void
generate_binary_implication_graph_roots (sector<int> & roots) {
  assert (roots.empty ());
  check_all_vars_unmarked ();
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->garbage) continue;
    if (c->size () > 2) continue;
    assert (c->size () == 2);
    mark_signed_lit (c->lits[0]);
    mark_signed_lit (c->lits[1]);
  }
  for (int idx = 1; idx <= maxvar; idx++) {
    if (is_active_lit (idx)) {
      bool pos = marked_signed_lit (idx);
      bool neg = marked_signed_lit (-idx);
      if (!pos && neg) roots.push_back (idx);
      if (pos && !neg) roots.push_back (-idx);
    }
    marks[idx] = 0;
  }
}

static void generate_arena_schedule_bfs (int root) {
  if (!is_active_lit (root)) return;
  if (marked_signed_lit (root)) return;
  LOG ("root arena literal " << root);
  unsigned next_lit_pos = work.size ();
  work.push_back (root);
  mark_signed_lit (root);
  while (next_lit_pos < work.size ()) {
    int lit = work[next_lit_pos++];
    const Wchs & ws = wchs (-lit);
    for (unsigned i = 0; i < ws.size (); i++) {
      const int blit = ws[i];
      const bool large = blit & 1;
      if (large) { i++; continue; }
      int other = blit >> 1;
      if (!is_active_lit (other)) return;
      if (marked_signed_lit (other)) continue;
      LOG ("next arena literal " << other);
      work.push_back (other);
      mark_signed_lit (other);
    }
  }
}

static void reset_all_marks () {
  for (int i = 1; i <= maxvar; i++) marks[i] = 0;
}

static void generate_arena_schedule () {
  assert (work.empty ());
  check_all_vars_unmarked ();
  sector<int> roots;
  generate_binary_implication_graph_roots (roots);
  for (unsigned i = 0; i < roots.size (); i++)
    generate_arena_schedule_bfs (roots[i]);
  release (roots);
  for (int idx = 1; idx <= maxvar; idx++)
    generate_arena_schedule_bfs (idx),
    generate_arena_schedule_bfs (-idx);
  reset_all_marks ();
}

static void copy_and_move_arena () {
  if (inconsistent) return;
  if (!opts::arena) return;
  assert (!level);
  assert (next == trail.size ());
  if (trail.size ()) flush_root_trail ();
  START (arena);
  copymovedarenas++;
  LOG ("computing new arena size");
  size_t new_bytes = 0;

  // First determine number of bytes to hold all current clauses.
  // An attempt to ignore certain clauses produced painful bugs!

  unsigned n = 0;
  for (unsigned i = 0; i < clauses.size (); i++)
    new_bytes += clauses[i]->bytes (), n++;

  unsigned w = 0;
#ifndef VECTOR
  if (opts::arena > 1) {

    // Compute number of bytes for all watcher stacks.

    for (int idx = 1; idx <= maxvar; idx++) {
      if (is_active_lit (idx)) {
	for (int sign = 1; sign >= -1; sign -= 2) {
	  const int lit = sign * idx;
	  const Wchs & ws = wchs (lit);
	  unsigned C = ws.capacity ();
	  if (!C) continue;
	  new_bytes += C * sizeof (unsigned);
	  w += C;
	}
      } else release (wchs (idx)), release (wchs (-idx));
    }
  }
#endif
  size_t old_bytes = arena.end - arena.start;
  MSG (1, "arena", copymovedarenas,
    "old " << bytes2mb (old_bytes) << " MB, " << 
    "new " << bytes2mb (new_bytes) << " MB, " << 
    n << " clauses " << w << " watches");

  // Allocate new target arena space.

  char * start;
  ALLOC (char, start, new_bytes);
  char * top = start, * end = start + new_bytes;

  generate_arena_schedule ();

  for (unsigned i = 0; i < work.size (); i++) {
    int lit = work[i];
    Wchs & ws = wchs (lit);

    // Copy watchers too if enabled.

#ifndef VECTOR
    if (opts::arena > 1) {
      unsigned C = ws.capacity ();
      if (!C) continue;
      ws.copy_and_move_to (top);	// not in 'std::vector'
      top += C * sizeof (unsigned);
    }
#endif

    // Move watched clauses in the following loop.

    for (unsigned i = 0; i < ws.size (); i++) {
      const unsigned blit = ws[i];
      const bool large = blit & 1;
      if (!large) continue;
      const unsigned cidx = ws[++i];
      clauses[cidx] =
	copy_and_move_clause (clauses[cidx], start, top, end);
    }
  }

  release (work);

  // Move unwatched clauses like garbage clauses or clauses with
  // only inactive literals (such as fixed literals) as first two
  // literals. Again, this was a very difficult to diagnose bug.

  for (unsigned i = 0; i < clauses.size (); i++)
    clauses[i] = copy_and_move_clause (clauses[i], start, top, end);

  assert (top <= end);
  LOG ("deallocating original space of " << old_bytes << " bytes");

  // Deallocate old arena space, but make sure nothing points into it.

  check_no_cls_in_arena ();
  check_no_wch_in_arena ();
  deallocate_arena ();

  arena.start = start;
  arena.end = end;

  report ('A');
  STOP (arena);
}

static bool arenaing () {
  if (level) return false;
  if (copymovedarenas) return false;	// Once initially for now ...
  if (!opts::arena) return false;
  if (inconsistent) return false;
  return true;
}

/*------------------------------------------------------------------------*/

static void connect_watches () {
  LOG ("connect watches");
  for (unsigned i = 0; i < clauses.size (); i++)
    watch (clauses[i], i);
}

static bool garbage_collection () {
  unsigned found_new_units;
  do {
    propagate_root_level ();
    if (inconsistent) break;

    // Cache last call to 'garbage_collection' in the sense that only
    // perform a garbage collection if there is at least one new root level
    // fixed variable.  Otherwise just skip it.

    if (fixed == limits.fixed.gc) break;
    limits.fixed.gc = fixed;

    START (gc);
    gcs++;
    LOG ("gc " << gcs);
    assert (!level);
    disconnect ();
    found_new_units = collect_clauses_flush_literals ();
    assert (!found_new_units || fixed > limits.fixed.gc);
    connect_watches ();
    if (opts::verbose) report ('g');
    STOP (gc);
  } while (found_new_units);
  return !inconsistent;
}

/*------------------------------------------------------------------------*/

// The assignment of eliminated or substituted variables might need to be
// fixed to satisfy the eliminated clauses on the extension stack.  This
// was pioneered by Niklas Soerensson and for instance described in our
// SAT'10 paper on reconstructing solutions.

static void push_extend_lit (int lit) {
  LOG ("extend " << lit);
  extend.push_back (lit);
}

static void push_extend_bin (int a, int b) {
  push_extend_lit (0);
  push_extend_lit (a);
  push_extend_lit (b);
}

static void push_extend_clause (int first, Cls * c) {
  push_extend_lit (0);
  push_extend_lit (first);
  for (unsigned i = 0; i < c->size (); i++) {
    int lit = c->lits[i];
    if (lit != first) push_extend_lit (lit);
  }
}

static void extend_model () {
  START (extend);
  LOG ("propagating through extension stack of size " << extend.size ());
  int lit = 0;
  bool satisfied = false;
  while (!extend.empty ()) {
    int other = extend.back ();
    extend.pop_back ();
    if (!other) {
      assert (lit);
      if (!satisfied) {
	assert (val (lit) < 0);
	vals[abs (lit)] = sign (lit);
	LOG ("assign " << lit << " flipped");
      } else satisfied = false;
    } else if (!satisfied) {
      Val tmp = val (other);
      assert (tmp);
      if (tmp > 0) satisfied = true;
    }
    lit = other;
  }
  STOP (extend);
}

/*------------------------------------------------------------------------*/

// This section contains the preprocessing / inprocessing algorithm
// 'decompose' for decompositing the binary literal implication graph into
// strongly connected components.  Literals in a strongly connected
// component can then be replaced by a representative.  In the literature
// this algorithm is also called 'equivalent literal substitution' (ELS).

static void connect_binary_implications () {
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->garbage) continue;
    if (c->size () != 2) continue;
    int f = c->lits[0], s = c->lits[1];
    if (val (f)) continue;
    if (val (s)) continue;
    wchs (-f).push_back (s);	// as edges in binary implication graph
    wchs (-s).push_back (f);	// without shifting and one negated sign
  }
}

// This is very similar to 'collect_clauses_flush_literals' except that we
// actually ignore all assigned values.  Instead we replace each literal
// 'lit' by its representative 'repr[lit]'.  This might make some clauses
// trivial which are then collected. Other clauses are shrunken and might
// even become unit clauses.  Assigning these units is postponed though,
// since it is difficult to keep proper watches while replacing literals by
// their representatives. Alternatively one could change 'connect' to resort
// literals in clauses before connecting, to make sure they are properly
// watched. 
//
// TODO investigate whether the latter makes the code simpler.  Also
// in 'collect_clauses_flush_literals' we had to go back to repeat the
// whole procedure if units are found after 'sweep' might find units
// without propagating over all clauses.  It might be best to merge
// these two procedures.

static void flush_repr (int * repr) {
  unsigned j = 0;
  for (unsigned i = 0; i < clauses.size (); i++) {

    Cls * c = clauses[i];

    unsigned trivial = c->garbage, duplicates = 0, rewrite = 0;
    const unsigned old_size = c->size ();

    // First test whether this clause contains clashing literals.

    for (unsigned k = 0; !trivial && k < old_size; k++) {
      const int orig = c->lits[k];
      const int lit = repr[orig];
      if (orig != lit) rewrite++;
      assert (!val (lit));
      Val tmp = marked_lit (lit);
           if (tmp < 0) trivial++;
      else if (tmp > 0) duplicates++;
      else mark_lit (lit);
    }

    for (unsigned k = 0; k < old_size; k++)
      unmark_lit (repr[c->lits[k]]);

    if (trivial) { flush_clause (c); continue; }

    if (duplicates || rewrite) {

      copy (unshrunken, *c);

      // This loop does the actual shrinking. As in 'add_original_clause' we
      // can not easily merge it with the one above while proof tracing.

      unsigned new_size = 0;
      for (unsigned k = 0; k < old_size; k++) {
	const int lit = repr[c->lits[k]];
	Val tmp = marked_lit (lit);
	if (tmp > 0) continue;
	c->lits[new_size++] = lit;
	assert (!tmp);
	mark_lit (lit);
      }
      assert (old_size == new_size + duplicates);
      flushedlits += duplicates;
      assert (new_size >= 1);

      for (unsigned k = 0; k < new_size; k++) unmark_lit (c->lits[k]);

      if (new_size == 1) {
	int unit = c->lits[0];
	units.push_back (unit);
	trace_add_unit_learned_clause (unit);
	dealloc_clause (c);
	flushedclauses++;
	c = 0;			// just delete it
      } else {
	if (!duplicates) assert (old_size == new_size);
	else c = realloc_clause (c, old_size, new_size);
	trace_add_learned_clause (*c);
      }

      // This is only sound since we added all the '-lit repr[lit]'
      // and 'lit -repr[lit]' clause before calling this function.

      trace_delete_clause (unshrunken);
    }

    if (c) clauses[j++] = c;
  }
  clauses.resize (j);
}

static void
decompose_flush_and_find_units (int * repr) {

  if (inconsistent) return;

  assert (units.empty ());

  // Add all equality clauses to eagerly remove clauses in 'flush_repr'.

  for (int lit = 1; lit <= maxvar; lit++) {
    int rlit = repr[lit];
    assert (repr[-lit] == -rlit);
    if (rlit == lit) continue;
    trace_add_binary_learned_clause (-lit, rlit);
    trace_add_binary_learned_clause (lit, -rlit);
  }

  flush_repr (repr);		// replace all 'lit' by 'repr[lit]'
  if (inconsistent) return;

  unsigned count = 0;
  for (int lit = 1; lit <= maxvar; lit++) {
    int rlit = repr[lit];
    assert (repr[-lit] == -rlit);
    if (rlit == lit) continue;

    // Handle equalities through the extension stack instead of
    // maintaining a global union find data structure for 'merged'
    // literals as in Lingeling (there actually for external literals).
    // In general we would need to save two clauses on the extension stack,
    // but with the same trick as discussed above around
    // 'assign_eliminated', we only need to save one.

    push_extend_bin (lit, -rlit);

    // Remove these equality clauses again, since 'lit' is gone.

    trace_delete_binary_learned_clause (-lit, rlit),
    trace_delete_binary_learned_clause (lit, -rlit);

    assert (!level);
    assign_substituted (-lit);
    count++;
  }
  MSG (1, "decompose", decompositions,
    "substituted " << count << " variables");
}

static int * decompose_tarjan () {
  LOG ("decompose tarjan");
  int * repr;	// literal to represenative mapping returned as result
  int * dfsi;	// depth first search index (DFSI)
  int * mrdfsi;	// minimal reachable DFSI (through back-edges)
  ALLOC (int, repr,   maxlit); repr   += maxvar;
  ALLOC (int, dfsi,   maxlit); dfsi   += maxvar;
  ALLOC (int, mrdfsi, maxlit); mrdfsi += maxvar;
  for (int lit = -maxvar; lit <= maxvar; lit++) repr[lit] = lit;
  for (int lit = -maxvar; lit <= maxvar; lit++) dfsi[lit] = 0;
  for (int lit = -maxvar; lit <= maxvar; lit++) mrdfsi[lit] = INT_MAX;
  sector<int> stack;
#ifdef LOGGING
  sector<int> scc;
#endif
  int i = 0;
  for (int ridx = 1; ridx <= maxvar; ridx++) {
    for (int rsign = -1; rsign <= 1; rsign += 2) {
      int root = rsign * ridx;
      if (dfsi[root]) continue;
      assert (work.empty ());
      work.push_back (root);
      while (!work.empty ()) {
	int lit = work.back ();
	if (!lit) {
	  work.pop_back ();
	  assert (!work.empty ());
	  lit = work.back ();
	  work.pop_back ();
	  Wchs & ws = wchs (lit);
	  for (unsigned i = 0; i < ws.size (); i++) {
	    int other = ws[i];
	    if (mrdfsi[lit] > mrdfsi[other])
	      mrdfsi[lit] = mrdfsi[other];
	  }
	  if (mrdfsi[lit] != dfsi[lit]) continue;
	  unsigned start = stack.size (), size = 0;
	  int r = INT_MAX, o;
	  do {
	    o = stack[--start];
	    if (abs (o) < abs (r)) r = o;
	    size++;
	  } while (o != lit);
#ifdef LOGGING
	  scc.clear ();
	  scc.push_back (r);
#endif
	  for (unsigned i = start; i < stack.size (); i++) {
	    int other = stack[i];
	    if (other != r) {
	      repr[other] = r;
#ifdef LOGGING
	      assert (size > 1);
	      LOG ("repr[" << other << "] = " << r);
	      scc.push_back (other);
#endif
	    }
	    mrdfsi[other] = INT_MAX;
	  }
	  if (size > 1) LOG ("scc [ " << scc << " ]");
	  stack.resize (start);
	} else if (!dfsi[lit]) {
	  work.push_back (0);
	  mrdfsi[lit] = dfsi[lit] = ++i;
	  stack.push_back (lit);
	  Wchs & ws = wchs (lit);
	  for (unsigned i = 0; i < ws.size (); i++)
	    work.push_back (ws[i]);
	} else work.pop_back ();
      }
    }
  }
  assert (work.empty ());
  release (stack);
#ifdef LOGGING
  release (scc);
#endif
  mrdfsi -= maxvar; DEALLOC (int, mrdfsi, maxlit);
  dfsi   -= maxvar; DEALLOC (int, dfsi,   maxlit);
  return repr;
}

// Check whether the SCC decomposition is logically consistent.

static void decompose_consistent (int * repr) {
  for (int lit = 1; !inconsistent && lit <= maxvar; lit++) {
    if (repr[lit] != repr[-lit]) continue;
    LOG ("literal " << lit << " and " << -lit << " in same SCC");
    trace_add_unit_learned_clause (lit);
    learn_empty_clause ();
  }
}

static void decompose_delete_repr (int * repr) {
  repr -= maxvar;
  DEALLOC (int, repr, maxlit);
}

static void assign_units_and_propagate () {
  for (unsigned i = 0; !inconsistent && i < units.size (); i++) {
    int unit = units[i];
    Val tmp = val (unit);
    if (tmp > 0) continue;
    else if (tmp < 0) {
      if (is_elim_lit (unit)) continue;
      LOG ("unit " << unit << " already falsified");
      learn_empty_clause ();
    } else assign (unit, Unit ());
  }
  units.clear ();
  propagate_root_level ();
}

static void decompose () {
  assert (!level);
  if (!opts::decompose) return;
  if (!trigger.decompose) return;
  trigger.decompose = false;
  if (!garbage_collection ()) return;
  START (decompose);
  decompositions++;
  LOG ("decompose " << decompositions);
  disconnect ();
  connect_binary_implications ();
  int * repr = decompose_tarjan ();
  disconnect ();
  decompose_consistent (repr);	// SCC with both 'lit' and '-lit'?
  decompose_flush_and_find_units (repr);
  decompose_delete_repr (repr);
  connect_watches ();			// need proper watches first
  assign_units_and_propagate ();	// only then assign units
  report ('d');
  STOP (decompose);
}

/*------------------------------------------------------------------------*/

static void connect_clause (Cls * c, unsigned cidx) {
  assert (clauses[cidx] == c);
  assert (!c->garbage);
  assert (!c->redundant);
  const unsigned size = c->size ();
  for (unsigned i = 0; i < size; i++)
    wchs (c->lits[i]).push_back (cidx);
}

static void connect_all_clauses () {
  for (unsigned i = 0; i < clauses.size (); i++)
    connect_clause (clauses[i], i);
}

static void connect_all_irredundant_clauses () {
  for (unsigned i = 0; i < numirr; i++) {
    Cls * c = clauses[i];
    assert (!c->redundant);
    connect_clause (c, i);
  }
#ifndef NDEBUG
  for (unsigned i = numirr; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    assert (c->redundant);
  }
#endif
}

/*------------------------------------------------------------------------*/

// This is the start of the code for the SAT sweeping approach described in
// our 2013 LPAR paper based on blocked clause decomposition (BCD).  It goes
// all the way down to 'sweep' function.  This is the first inprocessing
// version implemented of it, as far we know.  We experimented with many BCD
// algorithms, but only the pure and simple ordered versions implemented
// below are fast enough to be used during inprocessing.  To implement this
// feature was one of the main motivations to start a new solver, since it
// is hard to add this to Lingeling.  Since this was one of the last added
// features and much code went in and out, documentation is sparse, but we
// want to improve this in the future.

// The blocked set data type is compact and does not need a lot of space.
// However during extending a BCD model in 'extend_bcd_model' it is much
// faster to use a flattened set, similar to the 'extend' stack of the
// main solver.  However, since most of our methods to obtain a blocked
// clause decomposition (BCD) are constructive, that is they approximate
// a blocked set from below by adding blocked clauses, the order is the
// opposite.  This has to be taken into account in 'extend_bcd_model'.
// That is the reason why we have to 'reverse' a 'BlockedSet' some times.

static void
flatten_blocked_set (sector<int> & flat, const BlockedSet & bcd) {
  for (unsigned i = 0; i < bcd.size (); i++) {
    unsigned cidx = bcd[i].cidx;
    int lit = bcd[i].blit;
    Cls * c = clauses[cidx];
    for (unsigned j = 0; j < c->size (); j++) {
      int other = c->lits[j];
      if (other != lit) flat.push_back (other);
    }
    flat.push_back (lit);
    flat.push_back (0);
  }
}

static BPV bpval (int lit) {		// Same as 'val ()' but for 'BPV'
  BPV res = bpvals[abs (lit)];
  if (lit < 0) res = ~res;
  return res;
}

static void random_bcd_model () {
  for (int idx = 1; idx <= maxvar; idx++) {
    if (!is_active_lit (idx)) continue;
    BPV v = rng.rand ();
    v <<= 32;
    v |= rng.rand ();
    // Uncomment to debug failing 'check_bcd_model' test:
    // v = (v&16) ? bptrue : bpfalse;
    bpvals[idx] = v;			// 64 random values
  }
}

// This is the same as 'extend_model' but for bit-parallel values,
// and, as discussed above, assumes reverse order of the blocked clauses.

static void extend_bcd_model (const sector<int> & bcd) {
  for (unsigned i = 0; i < bcd.size (); i++) {
    int lit = 0, other;
    BPV v = 0;
    while ((other = bcd[i]))
      v |= bpval (other), lit = other, i++;
    if ((v = ~v)) bpvals[abs (lit)] ^= v;
  }
}

// One implicit way to check corectness of BCD, which requires that the
// (flattened) blocked set is really blocked, is to simply generate some
// random bit-parallel models and extend them.  The resulting models should
// all be models of the clauses in the blocked set.

static void check_bcd_model (const sector<int> & bcd) {
#ifndef NDEBUG
  for (unsigned i = 0; i < bcd.size (); i++) {
    int lit;
    BPV v = 0;
    while ((lit = bcd[i])) v |= bpval (lit), i++;
    assert (v == bptrue);
  }
#endif
}

// Instead of checking that it is blocked, we check that it allows to
// generate models using the set as reconstruction stack.  This is still
// expensive but also checks 'random_model' and 'extend_bcd_model'.

static void check_blocked_clause_decomposition (BlockedSet & bcd) {
#ifdef EXPENSIVECHECKS
  if (!opts::check) return;
  ALLOC (BPV, bpvals, maxvar + 1);
  sector<int> flat;
  flatten_blocked_set (flat, bcd);
  for (int i = 0; i < 100; i++) {
    random_bcd_model ();
    extend_bcd_model (flat);
    check_bcd_model (flat);
  }
  release (flat);
  DEALLOC (BPV, bpvals, maxvar + 1);
#endif
}

/*------------------------------------------------------------------------*/

static Wchs & bcdwchs (int lit) { return bcdwatches[litidx (lit)]; }

static void clear_and_release_blocked_watches () {
  for (int idx = 1; idx <= maxvar; idx++) {
    for (int sign = -1; sign <= 1; sign += 2) {
      int lit = sign * idx;
      release (bcdwchs (lit));
    }
  }
}

static void mark_clause_blocked (Cls * c, unsigned cidx,
                                 int blit, BlockedSet & bcd) {
  assert (!c->blocked);
  assert (clauses[cidx] == c);
  LOG ("marking blocked " << *c);
  c->blocked = 1;
  unsigned i;
#ifndef NDEBUG
  for (i = 0; i < c->size (); i++)
    if (c->lits[i] == blit) break;
  assert (i < c->size ());
#endif
  for (i = 0; i < c->size (); i++)
    bcdwchs (c->lits[i]).push_back (cidx);
  bcd.push_back (Blocked (blit, cidx));
}

static bool try_bcd_block_clause (Cls * c, unsigned cidx, BlockedSet & bcd) {
  assert (c == clauses[cidx]);
  assert (!c->blocked);
  assert (!c->redundant);
  assert (!c->garbage);
  bool blocked = false;
  int lit = 0;
  mark_clause (*c);
  for (unsigned i = 0; !blocked && i < c->size (); i++) {
    lit = c->lits[i];
    Wchs & ws = bcdwchs (-lit);
    blocked = true;
    for (unsigned j = 0; blocked && j < ws.size (); j++) {
      const unsigned didx = ws[j];
      Cls * d = clauses[didx];
      assert (d->blocked);
      bcdresolutions++;
      bool trivial = false;
      for (unsigned l = 0; !trivial && l < d->size (); l++) {
	int other = d->lits[l];
	if (other != -lit && marked_lit (other) < 0) trivial = true;
      }
      if (!trivial) blocked = false;
    }
  }

  if (blocked) mark_clause_blocked (c, cidx, lit, bcd);
  unmark_clause (*c);

  return blocked;
}

/*------------------------------------------------------------------------*/

static void reset_notblocked_flags () {
  for (unsigned i = 0; i < numirr; i++)
    clauses[i]->notblocked = 0;
}

static void reset_all_blocked_flags () {
  for (unsigned i = 0; i < numirr; i++) {
    Cls * c = clauses[i];
    c->blocked = c->notblocked = 0;
  }
}

/*------------------------------------------------------------------------*/

// This section does a pure blocked clause decomposition as described in our
// 2013 LPAR paper on "Blocked Clause Decomposition".  The heuristic to
// optimize the size of the blocked set is to eagerly block as many clauses
// as possible or if two variables can block the same number to pick the
// variable first which would mark the smallest number of clauses as
// 'notblocked'.  See 'pure_bcd_less_than' for the actual code.

static unsigned pure_bcd_count_not_blocked (int lit) {
  Wchs & ws = wchs (lit);
  unsigned res = 0, j = 0;
  for (unsigned i = 0; i < ws.size (); i++) {
    unsigned cidx = ws[j++] = ws[i];
    Cls * c = clauses[cidx];
    if (c->blocked) { j--; continue; };
    if (c->notblocked) { j--; continue; };
    res++;
  }
  ws.resize (j);
  LOG ("literal " << lit << " actually occurs " << res << " times " <<
    "neither blocked nor not blocked");
  return res;
}

static void pure_bcd_mark_blocked_lit (int lit, BlockedSet & bcd) {
  LOG ("marking remaining clauses with " << lit << " as blocked");
  Wchs & ws = wchs (lit);
  for (unsigned i = 0; i < ws.size (); i++) {
    unsigned cidx = ws[i];
    Cls * c = clauses[cidx];
    if (c->blocked) continue;
    if (c->notblocked) continue;
    mark_clause_blocked (c, cidx, lit, bcd);
  }
  ws.clear ();
}

static void pure_bcd_mark_notblocked_lit (int lit) {
  LOG ("marking remaining clauses with " << lit << " as not blocked");
  Wchs & ws = wchs (lit);
  for (unsigned i = 0; i < ws.size (); i++) {
    unsigned cidx = ws[i];
    Cls * c = clauses[cidx];
    if (c->blocked) continue;
    if (c->notblocked) continue;
    LOG ("marking clause as not blocked " << * c);
    c->notblocked = true;
  }
  ws.clear ();
}

// Sorted backward: first added last. 

struct pure_bcd_less_than {
  bool operator () (unsigned a, unsigned b) const {
    unsigned pa = wchs (a).size ();
    unsigned na = wchs (-a).size ();
    unsigned pb = wchs (b).size ();
    unsigned nb = wchs (-b).size ();

    // Try to block as many as possible.

    unsigned maxa = std::max (pa, na);
    unsigned maxb = std::max (pb, nb);
    if (maxa < maxb) return true;
    if (maxa > maxb) return false;

    // Try to not-block as few as possible.

    unsigned mina = std::min (pa, na);
    unsigned minb = std::min (pb, nb);
    if (mina > minb) return true;
    if (mina < minb) return false;

    return b < a;
  }
};

static void pure_bcd (BlockedSet & bcd, int64_t reslim, unsigned nblim) {
  (void) reslim;
  (void) nblim;
  work.clear ();
  for (int i = 1; i <= maxvar; i++)
    if (is_active_lit (i))
      work.push_back (i);
  std::sort (work.begin (), work.end (), pure_bcd_less_than ());
  while (!work.empty ()) {
    int idx = work.back ();
    LOG ("pure BCD next variable " << idx);
    work.pop_back ();
    unsigned p = pure_bcd_count_not_blocked (idx);
    unsigned n = pure_bcd_count_not_blocked (-idx);
    int lit = (n < p ? idx : -idx);
    pure_bcd_mark_blocked_lit (lit, bcd);
    pure_bcd_mark_notblocked_lit (-lit);
  }
  release (work);
  reverse (bcd);
  reset_notblocked_flags ();
  LOG ("pure bcd of size " << bcd.size ());
}

/*------------------------------------------------------------------------*/

static void
order_bcd (BlockedSet & bcd, unsigned * order,
           int64_t reslim, unsigned nblim) {
  unsigned start = numirr/4, count = 0;
  if (start < 10000) start = 10000;
  unsigned wait = start;
  for (unsigned i = 0; bcdresolutions < reslim && i < numirr; i++) {
    unsigned cidx = order[i];
    Cls * c = clauses[cidx];
    if (c->blocked) continue;
    if (try_bcd_block_clause (c, cidx, bcd)) count++;
    else if (!nblim--) break;
    else if (wait) wait--;
    else if (count*10 < start) break;
  }
}

static void
linear_bcd (BlockedSet & bcd, int64_t reslim, unsigned nblim) {
  unsigned * order;
  ALLOC (unsigned, order, numirr);
  for (unsigned i = 0; i < numirr; i++) order[i] = i;
  order_bcd (bcd, order, reslim, nblim);
  DEALLOC (unsigned, order, numirr);
}

static void
inverse_bcd (BlockedSet & bcd, int64_t reslim, unsigned nblim) {
  unsigned * order;
  ALLOC (unsigned, order, numirr);
  for (unsigned i = 0; i < numirr; i++) order[i] = numirr - i - 1;
  order_bcd (bcd, order, reslim, nblim);
  DEALLOC (unsigned, order, numirr);
}

static void
pure_linear_bcd (BlockedSet & bcd, int64_t reslim, unsigned nblim) {
  pure_bcd (bcd, reslim, nblim);
  linear_bcd (bcd, reslim, nblim);
}

static void
pure_inverse_bcd (BlockedSet & bcd, int64_t reslim, unsigned nblim) {
  pure_bcd (bcd, reslim, nblim);
  inverse_bcd (bcd, reslim, nblim);
}

/*------------------------------------------------------------------------*/

static void check_all_clause_flags_clean () {
#ifndef NDEBUG
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->redundant) assert (i >= numirr);
    else {
      assert (i < numirr);
      assert (!c->garbage);
      assert (!c->blocked);
      assert (!c->notblocked);
    }
  }
#endif
}

static void check_all_variables_untagged () {
#ifndef NDEBUG
  for (int i = 1; i <= maxvar; i++) {
    Var & v = var (i);
    assert (!v.tag);
  }
#endif
}

/*------------------------------------------------------------------------*/

// We allocate all enabled BCD methods statically.  Then during execution we
// remember which one worked best.  During consecutive sweeping rounds we
// then commit to run less and less BCD methods until only the best w.r.t.
// past peformance (size produced BCD, where larger is better) is used.

// TODO not really C++, e.g., should use OOP and polymorphism instead in
// particular for a reentrant version.

typedef void (*BCDFUN)(BlockedSet &, int64_t reslim, unsigned nblim);

struct BCD {
  int idx;
  BCDFUN fun;
  const char * name;
  unsigned size;		// size of last BCD produced
  int64_t score;		// score = sum of past rankings w.r.t. size
};

static const int MAXBCDS = 6;
static BCD bcds[MAXBCDS];
static int active_bcds;

#define INITBCD(ENABLED,FUN,NAME) \
do { \
  if (!ENABLED) break; \
  int idx = active_bcds++; \
  assert (idx < MAXBCDS); \
  bcds[idx].idx = idx; \
  bcds[idx].fun = FUN; \
  bcds[idx].name = NAME; \
  bcds[idx].size = 0; \
  bcds[idx].score = 0; \
} while (0)

/*------------------------------------------------------------------------*/

static void run_bcd (int i, BlockedSet & bcd, BlockedSet & best) {
  if (best.size () + 1 >= numirr) return;
  assert (bcds[i].fun);
  assert (bcds[i].name);
  double start = get_process_time ();
  int64_t before = bcdresolutions, delta;
  check_all_clause_flags_clean ();
  connect_all_irredundant_clauses ();
  if (sweeps > 1) {
    delta = elimresolutions - lastelimres;
    delta *= 20; delta /= 100;
  } else delta = 10000000;
  if (delta < 2*numirr) delta = 2*numirr;
  int64_t reslim = bcdresolutions + delta;
  bcd.clear ();

  assert (best.size () <= numirr);
  unsigned nblim = numirr - best.size ();	// max not-blocked

  bcds[i].fun (bcd, reslim, nblim);
  bcds[i].size = bcd.size ();

  if (bcd.size () > best.size ()) copy (best, bcd);

  reset_all_blocked_flags ();
  clear_and_release_blocked_watches ();
  disconnect ();

  if (opts::verbose > 0) {
    char line[200];
    sprintf (line,
      "c [sweep-%lld] %3.0f%% %-14s %9u in %5.2f sec %9lld res",
      (long long) sweeps, 
      percent (bcd.size (), numirr),
      bcds[i].name,
      (unsigned) bcd.size (),
      get_process_time () - start,
      (long long) (bcdresolutions - before));
    std::cout << line << NL << std::flush;
  }

  check_blocked_clause_decomposition (bcd);
}

struct run_bcd_size_less_than {
  bool operator () (const BCD & a, const BCD & b) const {
    if (a.size > b.size) return true;
    if (a.size < b.size) return false;
    return a.idx < b.idx;
  }
};

struct run_bcd_score_less_than {
  bool operator () (const BCD & a, const BCD & b) const {
    if (a.score < b.score) return true;
    if (a.score > b.score) return false;
    return run_bcd_size_less_than () (a, b);
  }
};

/*------------------------------------------------------------------------*/

static void blocked_clause_decomposition (BlockedSet & best) {
  START (bcd);

  if (!bcds[0].name) {
    INITBCD (opts::bcdlinear,	   linear_bcd,       "linear");
    INITBCD (opts::bcdpurelinear,  pure_linear_bcd,  "pure-linear");
    INITBCD (opts::bcdinverse,     inverse_bcd,      "inverse");
    INITBCD (opts::bcdpureinverse, pure_inverse_bcd, "pure-inverse");
    if (!active_bcds) INITBCD (1,  linear_bcd,       "linear");
  }

  disconnect ();
  NEW (Wchs, bcdwatches, maxlit + 1);

  BlockedSet bcd;
  for (int i = 0; i < active_bcds; i++) run_bcd (i, bcd, best);
  release (bcd);

  std::sort (bcds, bcds + active_bcds, run_bcd_size_less_than ());
  for (int i = 0; i < active_bcds; i++) bcds[i].score += i;
  std::sort (bcds, bcds + active_bcds, run_bcd_score_less_than ());

  if (sweeps > 3 && active_bcds > 1) {
    active_bcds--;
    MSG (1, "sweep", sweeps,
      "BCD method " << bcds[active_bcds].name << " becomes inactive");
  }

  DEALLOC (Wchs, bcdwatches, maxlit + 1);
  connect_watches ();
  STOP (bcd);
}

/*------------------------------------------------------------------------*/

static void
sweep_simulate_random_model (int round, const sector<int> & bcd) {
  LOG ("sweep simulating random model round " << round);
  random_bcd_model ();
  extend_bcd_model (bcd);
  check_bcd_model (bcd);
  int pos = 0, neg = 0, count = 0;
  for (int idx = 1; idx <= maxvar; idx++) {
    Var & v = vars[idx];
    if (v.status) continue;
    count++;
    if (!v.pbb) continue;
    BPV tmp = bpvals[idx];
    if (tmp != bptrue &&  (v.pbb & 1)) {
      LOG ("discarding literal " << idx << " as potential backbone");
      v.pbb &= 2;
    }
    if (tmp != bpfalse && (v.pbb & 2)) {
      LOG ("discarding literal " << -idx << " as potential backbone");
      v.pbb &= 1;
    }
    assert (v.pbb != 3);
    if (v.pbb & 1) neg++;
    if (v.pbb & 2) pos++;
  }
  MSG (1, "sweep", sweeps, round << ": " <<
    pos + neg << " potential backbones " <<
    percent (pos + neg, count) << "% left");
}

static void
sweep_initialize_or_reuse_potential_backbones () {
  int opbb = 0, count = 0;
  for (int i = 1; i <= maxvar; i++) {
    if (!is_active_lit (i)) continue;
    count++;
    if (!vars[i].pbb) continue;
    opbb++;
    if (vars[i].pbb & 1) LOG ("using backbone " << i);
    if (vars[i].pbb & 2) LOG ("using backbone " << -i);
  }
  if (opbb) {
    MSG (1, "sweep", sweeps,
      "reusing " << opbb <<
      " potential backbones " << percent (opbb, count) << '%');
  } else {
    for (int i = 1; i <= maxvar; i++)
      if (is_active_lit (i)) vars[i].pbb = 3, count++;
    MSG (1, "sweep", sweeps,
      "all " << count << " active variables are potential backbones");
  }
}

static bool at_least_one_potential_backbone_left () {
  for (int i = 1; i <= maxvar; i++)
    if (is_active_lit (i) && vars[i].pbb) return true;
  return false;
}

/*------------------------------------------------------------------------*/

static unsigned sweep_flush_partitions () {
  unsigned i = 0;
  for (unsigned j = 0; j < parts.size (); j++) {
    Part * p = parts[j];
    if (p->garbage) { dealloc_part (p); continue; }
    assert (part.empty ());
    for (unsigned k = 0; k < p->size; k++) {
      int lit = p->lits[k];
      if (is_active_lit (lit)) part.push_back (lit);
    }
    if (part.size () < p->size) {
      dealloc_part (p);
      if (part.size () > 1) parts[i++] = alloc_part (part);
    } else parts[i++] = p;
    part.clear ();
  }
  parts.resize (i);
  return i;
}

static void
sweep_initialize_or_reuse_potential_equivalence_classes () {
  unsigned count = sweep_flush_partitions ();
  if (count && !at_most_two_bits_set (sweeps)) {
    MSG (1, "sweep", sweeps,
      "reusing " << count << " potential equivalence classes");
  } else {
    reset_parts ();
    MSG (1, "sweep", sweeps,
      "empty set of potential equivalence classes");
  }
}

/*------------------------------------------------------------------------*/

struct lit2val {
  inline static Val map (int i) { return val (i); }
};

struct lit2bpval {
  inline static BPV map (int i) { return bpval (i); }
};

template<class V>
struct sweep_less_than {
  bool operator () (int a, int b) { return V::map (a) < V::map (b); }
};

struct less_than_part {
  bool operator () (Part * a, Part * b) { return a->size < b->size; }
};

static void sweep_bit_parallel_initialize_partition () {
  assert (parts.empty ());
  sector<int> sorted;
  for (int i = 1; i <= maxvar; i++) {
    if (!is_active_lit (i)) continue;
    int lit = bpval (-i) < bpval (i) ? -i : i;
    sorted.push_back (lit);
  }
  std::sort (sorted.begin (), sorted.end (), sweep_less_than<lit2bpval> ());
  int prev = 0;
  assert (part.empty ());
  for (unsigned i = 0; i < sorted.size (); i++) {
    int lit = sorted[i];
    if (!prev || bpval (prev) == bpval (lit))
      part.push_back (lit);
    else {
      if (part.size () > 1)
	parts.push_back (alloc_part (part));
      part.clear ();
      part.push_back (lit);
    }
    prev = lit;
  }
  if (part.size () > 1)
    parts.push_back (alloc_part (part));
  part.clear ();
  release (sorted);
  MSG (1, "sweep", sweeps, "0: " <<
    parts.size () << " potential equivalence classes " <<
    percent (parts.size (), remaining_variables ()) << "%");
  std::sort (parts.begin (), parts.end (), less_than_part ());
}

template<class V>
static void sweep_refine_partition () {
  sector<Part *> newparts;
  unsigned i = 0;
  for (unsigned j = 0; j < parts.size (); j++) {
    Part * p = parts[j];
    if (p->garbage) { dealloc_part (p); continue; }
    std::sort (p->lits, p->lits + p->size, sweep_less_than<V> ());
    assert (part.empty ());
    unsigned splits = 0;
    int prev = 0;
    for (unsigned k = 0; k < p->size; k++) {
      int lit = p->lits[k];
      if (!is_active_lit (lit)) continue;
      if (!prev || V::map (prev) == V::map (lit))
	part.push_back (lit);
      else {
	assert (part.size () < p->size);
	if (part.size () > 1) {
	  Part * q = alloc_part (part);
	  if (!splits++) parts[i++] = q;
	  else newparts.push_back (q);
	}
	part.clear ();
	part.push_back (lit);
      }
      prev = lit;
    }
    if (part.size () == p->size) parts[i++] = p;
    else {
      dealloc_part (p);
      if (part.size () > 1) {
	Part * q = alloc_part (part);
	if (!splits++) parts[i++] = q;
	else newparts.push_back (q);
      }
    }
    part.clear ();
  }
  parts.resize (i);
  for (unsigned j = 0; j < newparts.size (); j++)
    parts.push_back (newparts[j]);
  release (newparts);
  LOG ("refined to " << parts.size () << " potential equivalence classes");
}

/*------------------------------------------------------------------------*/

static void sweep_bit_parallel_refine_partition (int round) {
  sweep_refine_partition<lit2bpval> ();
  MSG (1, "sweep", sweeps, round << ": " <<
    parts.size () << " potential equivalence classes " <<
    percent (parts.size (), remaining_variables ()) << "%");
}

static void sweep_initialize (const sector<int> & bcd) {
  START (rand);
  LOG ("sweep initialize");
  sweep_initialize_or_reuse_potential_backbones ();
  sweep_initialize_or_reuse_potential_equivalence_classes ();
  bool pbbleft = true;
  ALLOC (BPV, bpvals, maxvar + 1);
  for (int i = 0; i < opts::sweeprand; i++) {
    sweep_simulate_random_model (i, bcd);
    if (pbbleft) pbbleft = at_least_one_potential_backbone_left ();
    if (parts.empty ()) sweep_bit_parallel_initialize_partition ();
    else sweep_bit_parallel_refine_partition (i);
    if (!pbbleft && parts.empty ()) break;
  }
  DEALLOC (BPV, bpvals, maxvar + 1);
  STOP (rand);
}

static void sweep_reset () {
  unsigned count = sweep_flush_partitions ();
  MSG (1, "sweep", sweeps,
    "keeping " << count << " potential equivalence classes");
}

static void sweep_refine_backbone () {
  int removed = 0, count = 0;
  for (int other = 1; other <= maxvar; other++) {
    if (!is_active_lit (other)) continue;
    count++;
    Var & u = vars[other];
    if (!u.pbb) continue;
    Val tmp = val (other);
    if (tmp < 0 && (u.pbb & 1)) u.pbb &= 2, removed++;
    if (tmp > 0 && (u.pbb & 2)) u.pbb &= 1, removed++;
  }
  LOG ("removed " << removed << " potential backbone candidates " <<
    percent (removed, count) << "% out of " << count);
}

static int limited_assume_solve (int decision, int64_t limit);

static bool sweep_backbones () {
  LOG ("sweep backbones");
  int count = 0, tried = 0, refined = 0;

  int64_t clim = conflicts + 100000;
  int64_t dlim = decisions + 1000000 + remaining_variables ();
  int64_t plim = propagations + 10000000 + 10*remaining_variables ();

  int start = 0;
  bool changed = false;

  while (!inconsistent) {

    if (conflicts > clim) break;
    if (decisions > dlim) break;
    if (propagations > plim) break;

    if (!nextbb || nextbb > maxvar) nextbb = 1;

    int idx = nextbb;

    if (!start) start = nextbb;
    else if (start == nextbb) {
      if (!changed) break;
      else changed = false;
    }
    nextbb++;

    if (!is_active_lit (idx)) continue;

    Var & v = var (idx);
    if (!v.pbb) continue;
    tried++;
    assert (v.pbb != 3);
    int pbb = (v.pbb & 1) ? idx : -idx;
    backbonechecks++;
    LOG ("checking potential backbone " << pbb);
    int tmp = limited_assume_solve (-pbb, opts::sweepconflim);
    if (inconsistent) break;
    if (tmp == 20) {
      LOG ("backbone " << pbb);
      assert (val (pbb) > 0);
      backbones++;
      changed = true;
      count++;
    } else if (tmp == 10) {
      LOG ("check failed thus refining");
      sweep_refine_backbone ();
      sweep_refine_partition<lit2val> ();
      backtrack ();
      refined++;
    } else {
      assert (!tmp);
      LOG ("conflict limit reached checking " << pbb);
    }
  }
  MSG (1, "sweep", sweeps,
    count << " backbones, " <<
    refined << " refined " << percent (refined, tried));
  return count > 0;
}

static int limited_assume2_solve (int d1, int d2, int64_t limit);

static bool sweep_equivalences () {
  LOG ("sweep equivalences");
  int count = 0, tried = 0, refined = 0;

  unsigned checks = 0;

  int64_t clim = conflicts + 100000;
  int64_t dlim = decisions + 1000000 + remaining_variables ();
  int64_t plim = propagations + 10000000 + 10*remaining_variables ();

  while (!inconsistent && !parts.empty ()) {

    if (conflicts > clim) break;
    if (decisions > dlim) break;
    if (propagations > plim) break;

    unsigned i;
    if (!(checks++ & 1)) {
      if (partlo >= parts.size ()) {
	if (!sweep_flush_partitions ()) break;
	partlo = 0;
      }
      i = partlo++;
      LOG ("picking lower end partition at position " << i);
    } else {
      if (!parthi) parthi = parts.size ();
      else if (parthi > parts.size ()) parthi = parts.size ();
      i = --parthi;
      LOG ("picking upper end partition at position " << i);
    }
    Part * p = parts[i];
    if (p->garbage) continue;
    unsigned pos1 = rng.rand () % p->size;
    unsigned pos2 = rng.rand () % p->size;
    while (pos2 == pos1) pos2 = rng.rand () % p->size;
    if (pos1 > pos2) std::swap (pos1, pos2);
    int lit1 = p->lits[pos1];
    int lit2 = p->lits[pos2];
    assert (pos1 < pos2);
    assert (lit1 != lit2);
    LOG ("checking equivalence of " << lit1 << " <-> " << lit2 <<
      " in potential equivalence class of size " << p->size);
    equivchecks++;
    implieschecks++;
    int lim = opts::sweepconflim;
    int tmp = limited_assume2_solve (lit1, -lit2, lim);
    if (tmp == 20) {
      LOG ("implies " << lit1 << " -> " << lit2);
      implies++;
      if (level) backtrack ();
      implieschecks++;
      tmp = limited_assume2_solve (lit2, -lit1, lim);
      if (tmp == 20) {
	implies++;
	LOG ("implies " << lit2 << " -> " << lit1);
	equivs++;
	count++;
	LOG ("equivalent " << lit1 << " <-> " << lit2);

	// One could rely on 'HBR' to derive those clauses.  Without 'HBR'
	// it clearly will not be always the case and then the following
	// 'decompose' will miss this equivalence.  On the other hand this
	// is not costly at all and thus we always eagerly add these.

	learn_additional_binary_clause (-lit2, lit1);
	learn_additional_binary_clause (lit2, -lit1);

	if (p->size == 2) p->garbage = true;
	else parts[i] = realloc_part (p, lit2);
      }
    }
    if (tmp == 10) {
      LOG ("implication check failed thus refining");
      sweep_refine_backbone ();
      sweep_refine_partition<lit2val> ();
      partlo = 0;
      parthi = parts.size ();
      backtrack ();
      refined++;
    } 
    if (!tmp) {
      assert (!tmp);
      LOG ("conflict limit reached checking");
    }
  }
  MSG (1, "sweep", sweeps,
    count << " equivalences, " <<
    refined << " refined " << percent (refined, tried) << '%');
  return count > 0;
}

/*------------------------------------------------------------------------*/

static void watch_bcd_clauses (const BlockedSet & bcd) {
  bool more_aggressive_watching = !(sweeps & 1);
#ifndef NDEBUG
  for (unsigned i = 0; i < clauses.size (); i++)
    assert (!clauses[i]->watched);
#endif
  for (unsigned i = 0; i < bcd.size (); i++) {
    unsigned cidx = bcd[i].cidx;
    Cls * c = clauses[cidx];
    assert (!c->redundant);
    watch (clauses[cidx], cidx);
    if (more_aggressive_watching) c->watched = 1;
  }
  if (!more_aggressive_watching) return;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->watched) continue;
    unsigned size = c->size ();
    if (size > 2) continue;
    watch (c, i);
  }
  for (unsigned i = 0; i < bcd.size (); i++) {
    unsigned cidx = bcd[i].cidx;
    Cls * c = clauses[cidx];
    assert (c->watched);
    c->watched = 0;
  }
}

static void sweep () {
  assert (!level);
  if (!opts::sweep) return;
  if (opts::sweepwait && simplifications <= opts::sweepwait) return;
  if (delays.sweep.delayed ()) return;
  if (!garbage_collection ()) return;
  START (sweep);
  sweeps++;
  LOG ("sweep " << sweeps);
  BlockedSet bcd;
  blocked_clause_decomposition (bcd);
  check_blocked_clause_decomposition (bcd);
  disconnect ();
  watch_bcd_clauses (bcd);
  sector<int> flat;
  flatten_blocked_set (flat, bcd);
  release (bcd);
  sweep_initialize (flat);
  release (flat);
  bool success = false;
  if (sweep_backbones ()) success = true;
  if (sweep_equivalences ()) success = true;
  disconnect ();
  sweep_reset ();
  connect_watches ();
  delays.sweep.update (success);
  report ('w');
  STOP (sweep);
}

/*------------------------------------------------------------------------*/

static int all_variables_assigned () { return trail.size () == active; }

static int next_decision () {
  check_queue ();
  assert (!all_variables_assigned ());
  assert (next == trail.size ());
  assert (!inconsistent);
  assert (!conflict);
  unsigned count = 0;
  while (val (queue.searched))
    queue.searched = lnks[queue.searched].prev, count++;
  LOG ("searched next " << lnk (queue.searched).stamp);
  searches += count;
  int idx = queue.searched;
  int phase = phases[idx];
  return phase * idx;		// see 'phase saving' above
}

static void assume_decision (int lit) {
  assert (next == trail.size ());
  if (!level) flush_root_trail ();
  LOG ("decide " << lit);
  decisions++;
  level++;
  control.push_back (Frame (lit));
  assign (lit, Decision ());
}

static void decide () { assume_decision (next_decision ()); }

/*------------------------------------------------------------------------*/

struct probe_less_than {
  bool operator () (int i, int j) {
    int64_t pi = wchs (i).size (), ni = wchs (-i).size ();
    int64_t pj = wchs (j).size (), nj = wchs (-j).size ();
    int64_t si = (pi + 1) * ni;
    int64_t sj = (pj + 1) * nj;
    if (si > sj) return true;
    if (si < sj) return false;
    return abs (i) < abs (j);
  }
};

static void increment_probe_stamp (unsigned & last, unsigned * stamps) {
  if (++last) return;
  for (int lit = -maxvar; lit <= maxvar; lit++) stamps[lit] = 0;
  last = 1;
}

static void probe () {
  if (!opts::probe) return;
  if (!trigger.probe) return;
  if (opts::probewait && simplifications <= opts::probewait) return;
  if (delays.probe.delayed ()) return;

  START (probe);

  probings++;

  unsigned failedlits_this_time = 0;

  unsigned * stamps;
  ALLOC (unsigned, stamps, 2*maxvar+1);
  stamps += maxvar;

  int64_t delta = propagations - lastprops;
  delta *= opts::probeproplim; delta /= 1000;
  if (delta < opts::probepropminlim) delta = opts::probepropminlim;
  int64_t limit = propagations + delta;

  int round = 0;

  for (;;) {

    if (propagations >= limit) {
      MSG (1, "probe", probings,
        "incomplete probing " << 
	probes.size () << " probes left after "  << round << " rounds ");
      break;
    }

    round++;
    int64_t before = failedlits;
    unsigned last = 0;

    if (!garbage_collection ()) {
      MSG (1, "probe", probings,
        "empty clause in garbage collection after " << round << " rounds");
      break;
    }

    // Reset stamps.

    for (int lit = -maxvar; lit <= maxvar; lit++) stamps[lit] = 0;

    // Calculate binary clause occurrence for all literals.

    for (unsigned i = 0; i < clauses.size (); i++) {
      Cls * c = clauses[i];
      if (c->garbage) continue;
      if (c->size () > 2) continue;
      stamps[c->lits[0]] = 1;
      stamps[c->lits[1]] = 1;
    }

    // Try reusing previous probes.

    if (!probes.empty ()) {
      unsigned j = 0;
      for (unsigned i = 0; i < probes.size (); i++) {
	int lit = probes[i];
	if (val (lit)) continue;
	if (stamps[lit]) continue;
	if (!stamps[-lit]) continue;
	stamps[-lit] = 0;			// mark as already used
	probes[j++] = lit;
      }
      probes.resize (j);
      MSG (1, "probe", probings,
        "reusing " << probes.size () << " old probes");
    } else
      MSG (1, "probe", probings, "probes generated from scratch");

    // Generate list of roots in binary implication graph as probes.

    for (int idx = 1; !inconsistent && idx <= maxvar; idx++) {
      if (val (idx)) continue;
      int pos = stamps[idx], neg = stamps[-idx];
      if (pos && !neg) probes.push_back (-idx);
      if (!pos && neg) probes.push_back (idx);
    }
    MSG (1, "probe", probings,
      "using " << probes.size () << " roots as probes");
#ifndef NDEBUG
    for (unsigned i = 0; i < probes.size (); i++)
      assert (!stamps[probes[i]]);
#endif

    std::sort (probes.begin (), probes.end (), probe_less_than ());

    last = 1;
    before = failedlits;

    for (unsigned i = 0;
	 !inconsistent && propagations < limit && i < probes.size ();
	 i++) {
      int probe = probes[i];
      if (val (probe)) continue;
      if (stamps[probe] == last) continue;
      LOG ("probe " << probe);
      assert (simpmode);
      assert (!level);
      unsigned before = trail.size ();
      assume_decision (probe);
      failedlitprobes++;
      if (propagate ()) { 
	LOG ("literal first phase " << probe << " not failed");
	for (unsigned h = before; h < trail.size (); h++)
	  stamps[trail[h]] = last;
	backtrack ();
      } else {
	failedlits_this_time++;
	failedlits++;
	LOG ("failed literal " << probe);
	analyze ();
	increment_probe_stamp (last, stamps);
	if (!propagate ()) learn_empty_clause ();
      }
    }

    if (inconsistent) {
      MSG (1, "probe", probings,
        "probing finds empty clause after " << round << " rounds");
      break;
    }

    if (propagations < limit && before == failedlits) {
      MSG (1, "probe", probings,
        "completed probing all roots in " << round << " rounds");
      trigger.probe = false;
      release (probes);
      break;
    }

    unsigned j = 0;
    for (unsigned i = 0; i < probes.size (); i++) {
      int probe = probes[i];
      if (stamps[probe] == last) continue;
      probes[j++] = probe;
    }
    probes.resize (j);
    MSG (1, "probe", probings,
      "keeping " << probes.size () << " probes for next round");
    assert (trigger.probe);
  }

  stamps -= maxvar;
  DEALLOC (unsigned, stamps, 2*maxvar+1);

  delays.probe.update (failedlits_this_time > 0);
  MSG (1, "probe", probings,
    "produced " << failedlits_this_time << " failed ");
  report ('p');
  STOP (probe);
}

/*------------------------------------------------------------------------*/

// This section implements a variant of bounded variable elimination (BVE)
// as described in our SAT 2005 paper.  We do not eagerly disconnect
// eliminated clauses but run garbage collection explicitly in intervals.
// This allows us to use much larger limits on the number of occurrences as
// in the original implementation.  We also keep the subsumption outside of
// this part and run it as different preprocessing during inprocessing.

static unsigned elim_score (int i) {
  unsigned pos = wchs (i).size ();
  unsigned neg = wchs (-i).size ();
  assert (UINT_MAX - pos >= neg);
  return pos + neg;
}

struct elim_less_than {
  bool operator () (int i, int j) const {
    unsigned a = elim_score (i);
    unsigned b = elim_score (j);
    if (a > b) return true;
    if (a < b) return false;
    assert (i > 0), assert (j > 0);
    return i > j;
  }
};

static void sort_elim_score () {
  std::sort (work.begin (), work.end (), elim_less_than ());
}

static void generate_elimination_schedule () {
  check_all_variables_untagged ();
  assert (work.empty ());
  assert (resched.empty ());
  for (int idx = 1; idx <= maxvar; idx++)
    if (is_active_lit (idx)) work.push_back (idx);
  sort_elim_score ();
  for (unsigned i = 0; i < work.size (); i++)
    var (work[i]).tag = 1;
  LOG ("generate schedule of size " << work.size ());
}

static void save_redundant (sector<Cls*> & redundant) {
  unsigned j = 0;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->garbage) flush_clause (c);
    else if (c->redundant) redundant.push_back (c);
    else clauses[j++] = c;
  }
  clauses.resize (j);
  LOG ("saved " << redundant.size () << " redundant clauses");
}

static int actual_occurrences (int lit, int lim1) {
  Wchs & ws = wchs (lit);
  if (ws.size ()/10 > (unsigned) lim1) return lim1;
  unsigned j = 0;
  int res = 0;
  for (unsigned i = 0; i < ws.size (); i++) {
    unsigned cidx = ws[i];
    Cls * c = clauses[cidx];
    if (c->garbage) continue;
    if (c->blocked) continue;
    if (c->notblocked) continue;
    ws[j++] = cidx;
    res++;
  }
  ws.resize (j);
  LOG ("literal " << lit << " with " << res << " occurrences");
  return res;
}

static bool variable_can_be_eliminated (int idx) {
  if (!is_active_lit (idx)) return false;
  LOG ("try to eliminate variable " << idx);
  int lim1 = opts::elim1lim;
  int pos = actual_occurrences (idx, lim1);
  if (pos >= lim1) {
    LOG ("one sided limit " << lim1 << " hit for " << idx);
    return false;
  }
  int neg = actual_occurrences (-idx, lim1);
  if (neg >= lim1) {
    LOG ("one sided limit " << lim1 << " hit for " << -idx);
    return false;
  }
  if (pos * (int64_t) neg > (int64_t) opts::elim2lim) {
    LOG ("two sided limit " << opts::elim2lim << " hit for " << idx);
    return false;
  }
  if (pos > neg) idx = -idx;
  assert (pos <= opts::elim1lim);
  assert (neg <= opts::elim1lim);
  assert (pos*neg <= opts::elim2lim);
  const int trivial_resolvents_needed = pos*neg - pos - neg;
  const int maximum_non_trivial_resolvents = pos + neg;
  int trivial_resolvents = 0, non_trivial_resolvents = 0;
  Wchs & pws = wchs (idx);
  Wchs & nws = wchs (-idx);
  bool done = false;
  for (unsigned i = 0; !done && i < pws.size (); i++) {
    unsigned cidx = pws[i];
    Cls * c = clauses[cidx];
    assert (!c->garbage);
    mark_clause (*c);
    for (unsigned j = 0; !done && j < nws.size (); j++) {
      elimresolutions++;
      unsigned didx = nws[j];
      Cls * d = clauses[didx];
      assert (!d->garbage);
      bool trivial = false;
      for (unsigned k = 0; !trivial && k < d->size (); k++) {
	int other = d->lits[k];
	if (other == -idx) continue;
	Val tmp = marked_lit (other);
	if (tmp > 0) continue;
	if (tmp < 0) trivial = true;
      }
      LOG ("1st antecendent " << *c);
      LOG ("2nd antecendent " << *d);
      if (trivial) {
	LOG ("resolvent would be trivial");
	if (++trivial_resolvents >= trivial_resolvents_needed)
	  done = true;
      } else if (c->size () + d->size () > (unsigned) opts::elimclslim + 2) {
	LOG ("non-trivial resolvent too large");
	non_trivial_resolvents = maximum_non_trivial_resolvents + 1;
	done = true;
      } else {
	LOG ("resolvent would be non-trivial");
	if (++non_trivial_resolvents > maximum_non_trivial_resolvents)
	  done = true;
      }
    }
    unmark_clause (*c);
  }
  return non_trivial_resolvents <= maximum_non_trivial_resolvents;
}

static void add_resolved_clause () {
  assert (!level);
  trace_add_learned_clause (clause);
  assert (!clause.empty ());
  if (clause.size () == 1) {
    int unit = clause[0];
    LOG ("resolved unit clause " << unit);
    if (is_fixed_lit (unit)) units.push_back (unit);
    else assign (unit, Unit ());
  } else {
    assert (clause.size () >= 2);
    Cls * c = allocate_irredundant_clause (clause.size ());
    LOG ("resolved length " << c->size () << " clause " << *c);
    assert (clauses.size () < UINT_MAX);
    unsigned cidx = clauses.size ();
    clauses.push_back (c);
    connect_clause (c, cidx);
  }
}

static void touch_again_literal (int idx) {
  Var & v = var (idx);
  if (v.tag) return;
  resched.push_back (idx);
  v.tag = 1;
  LOG ("touch again " << idx);
}

static void touch_again_clause (Cls * c, int except) {
  for (unsigned i = 0; i < c->size (); i++) {
    int other = c->lits[i];
    if (other == except) continue;
    touch_again_literal (abs (other));
  }
}

static int next_elim_work () {
  if (work.empty ()) {
    if (resched.empty ()) return 0;
    copy (work, resched);
    resched.clear ();
    sort_elim_score ();
  }
  int res;
  if (resched.empty () ||
      elim_score (resched.back ()) >= elim_score (work.back ())) {
    assert (!work.empty ());
    res = work.back ();
    work.pop_back ();
    LOG ("next elim work " << res << " score " << elim_score (res));
  } else {
    assert (!resched.empty ());
    res = resched.back ();
    resched.pop_back ();
    LOG ("next elim resched " << res << " score " << elim_score (res));
  }
  assert (res);
  Var & v = var (res);
  assert (v.tag);
  v.tag = 0;
  return res;
}

static void untouch_all () {
  for (unsigned i = 0; i < resched.size (); i++) var (resched[i]).tag = 0;
  for (unsigned i = 0; i < work.size (); i++) var (work[i]).tag = 0;
  release (resched);
  release (work);
}

static void eliminate_variable (int idx) {
  assert (!inconsistent);
  LOG ("eliminate variable " << idx);
  Wchs & pws = wchs (idx);
  Wchs & nws = wchs (-idx);
  unsigned plits = 0, nlits = 0;
  for (unsigned i = 0; i < pws.size (); i++) {
    unsigned cidx = pws[i];
    Cls * c = clauses[cidx];
    assert (!c->garbage);
    assert (clause.empty ());
    for (unsigned k = 0; k < c->size (); k++) {
      int other = c->lits[k];
      if (other == idx) continue;
      clause.push_back (other);
      mark_lit (other);
    }
    for (unsigned j = 0; j < nws.size (); j++) {
      elimresolutions++;
      unsigned didx = nws[j];
      Cls * d = clauses[didx];
      assert (!d->garbage);
      bool trivial = false;
      assert (clause.size () == c->size () - 1);
      for (unsigned k = 0; !trivial && k < d->size (); k++) {
	int other = d->lits[k];
	if (other == -idx) continue;
	Val tmp = marked_lit (other);
	if (tmp > 0) continue;
	if (tmp < 0) trivial = true;
	else clause.push_back (other);
      }
      LOG ("1st antecendent " << *c);
      LOG ("2nd antecendent " << *d);
      if (trivial) LOG ("resolvent trivial");
      else {
	LOG ("resolved clause " << clause);
	add_resolved_clause ();
      }
      clause.resize (c->size () - 1);
    }
    unmark_clause (*c);
    clause.clear ();
    trace_delete_clause (*c);
    assert (!c->garbage);
    assert (!c->traced);
    c->garbage = true;
    c->traced = true;
    touch_again_clause (c, idx);
    plits += c->size () + 1;
  }
  for (unsigned j = 0; j < nws.size (); j++) {
    unsigned didx = nws[j];
    Cls * d = clauses[didx];
    trace_delete_clause (*d);
    assert (!d->garbage);
    assert (!d->traced);
    d->garbage = true;
    d->traced = true;
    touch_again_clause (d, -idx);
    nlits += d->size () + 1;
  }

  // The following is an optimization based on an idea by Niklas Soerensson,
  // who also originally came up with the whole idea of the solution
  // reconstruction stack, we call 'extend'.  After eliminating a variable
  // we save the eliminated clauses on this stack,  with the eliminated
  // literal first.  During extending an assignment to the eliminated
  // literal we have to check whether any of the eliminated clauses is
  // false.  If so we flip the value of the eliminated variable.  In order
  // to save space and for faster extension, Niklas observed that we only
  // have to save the clauses for the positive or negative occurrence of the
  // eliminated variable.  For the other phase a unit clause can be pushed.
  // The reconstruction algorithm will then always assign the variable first
  // to the other phase and then check whether the saved smaller set of
  // clauses are also satisfied.  We use a simple extension, which assigns
  // the eliminated variable immediately, which means that we do not have to
  // save the unit clause on the extension stack.

  if (plits > nlits) {
    for (unsigned j = 0; j < nws.size (); j++)
      push_extend_clause (-idx, clauses [nws[j]]);
    assign_eliminated (idx);
  } else {
    for (unsigned i = 0; i < pws.size (); i++)
      push_extend_clause (idx, clauses [pws[i]]);
    assign_eliminated (-idx);
  }
}

// We collect garbage clauses in intervals instead of eagerly detaching
// them.  This has the drawback that the actual occurrence count does not
// necessarily match the size of the watcher stacks and that potentially
// more memory is used.  This however avoids a hidden quadratic algorithm
// for removing the references to clauses, which really happens.  Because of
// using this delayed garbage collection we can use much higher limits on
// the size of the occurrence clauses than in other solvers.

static void collect_garbage_clauses () {
  unsigned j = 0, collected = 0;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->garbage) flush_clause (c), collected++;
    else clauses[j++] = c;
  }
  clauses.resize (j);
  LOG ("collected " << collected << " garbage clauses");
}

static void collect_elimination_garbage () {
  disconnect ();
  collect_garbage_clauses ();
  connect_all_clauses ();
  if (opts::verbose) report ('c');	// also shows progress ...
}

static unsigned next_elim_gc_limit () {
  assert (units.empty ());
  // Allow a little bit more garbage clauses than current clauses.
  unsigned limit = 5*(clauses.size ()/4) + 10000;
  LOG ("next variable elimination garbage collection limit " << limit);
  return limit;
}

static void elim_gc_if_necessary (unsigned & limit) {
  if (clauses.size () < limit) return;
  collect_elimination_garbage (),
  limit = next_elim_gc_limit ();
}

static bool run_variable_elimination_schedule () {
  check_all_clause_flags_clean ();	// 'actual_occurrences' requirement
  unsigned limit = next_elim_gc_limit ();
  LOG ("run bounded variable elimination");
  unsigned count = 0;
  int idx;
  while ((idx = next_elim_work ())) {
    elim_gc_if_necessary (limit);
    if (variable_can_be_eliminated (idx))
      eliminate_variable (idx), count++;
  }
  untouch_all ();	// redundant as long the above loop is unlimited ...
  MSG (1, "elim", eliminations, "eliminated " << count << " variables");
  if (count) trigger.subsume = trigger.block = true;
  return count > 0;
}

static bool contains_eliminated_variable (Cls * c) {
  const unsigned size = c->size ();
  for (unsigned i = 0; i < size; i++)
    if (is_elim_lit (c->lits[i])) return true;
  return false;
}

static void copy_back_redundant (const sector<Cls*> & redundant) {
  unsigned copied = 0;
  for (unsigned i = 0; i < redundant.size (); i++) {
    Cls * c = redundant[i];
    if (c->garbage || contains_eliminated_variable (c)) flush_clause (c);
    else clauses.push_back (c), copied++;
  }
  LOG ("copied back "
    << copied << " redundant clauses"
    << percent (copied, redundant.size ()));
}

static void elim () {
  assert (!level);
  if (!opts::elim) return;
  if (!trigger.elim) return;
  if (delays.elim.delayed ()) return;
  if (!garbage_collection ()) return;
  START (elim);
  eliminations++;
  LOG ("elim " << eliminations);
  trigger.elim = false;
  lastelimres = elimresolutions;
  disconnect ();
  sector<Cls*> redundant;
  save_redundant (redundant);
  connect_all_clauses ();
  generate_elimination_schedule ();
  bool success = run_variable_elimination_schedule ();
  disconnect ();
  collect_garbage_clauses ();
  copy_back_redundant (redundant);
  release (redundant);
  connect_watches ();			// need proper watches first
  assign_units_and_propagate ();
  delays.elim.update (success);
  report ('e');
  STOP (elim);
}

/*------------------------------------------------------------------------*/

// Bloom filter for variables or signatures, see our SAT'05 paper.

static const int sigmask = (8*sizeof (Sig)) - 1;

static Sig varsig (int lit) { return ((Sig)1) << (abs (lit) & sigmask); }

static bool subsig (Sig a, Sig b) { return !(a & ~b); }

/*------------------------------------------------------------------------*/

// This is the blocked clause elimination (BCE) part as described in our
// TACAS'10 paper and in follow up journal articles (JAR + JAIR).  As in
// other implementations scheduling the blocked clause removal is done
// identical to bounded variable elimination, by trying to eliminate clauses
// with positive and negative occurrence of that variable.  Variables with
// a small number (sum) of positive and negative occurrences are tried
// first. If a clause is removed the other variables in that clause are
// scheduled to be tried again.  What is new here is the use of a 'nontriv'
// flag, which remembers, whether a clause produced a non-trivial resolvent,
// while checking the clauses with one phase of occurrence of the scheduled
// variable.  We also move failing (that is witness for non-trivial
// resolvents) to the front of the watcher stacks. The main function for
// this inprocessing technique is 'block' further down.

// Blocked (irredundant) clauses can be turned into 'redundant' clauses.  We
// call this 'moving' a (blocked) clause.  This allows bounded variable
// eliminination to ignore them. But they can still be used for propagation
// and learning.  See our IJCAR'12 paper for a formal argument, why this is
// sound (as long model changing clauses are not added for instance through
// blocked clause addition).  The tricky part is that redundant clauses can
// in principle be always thrown away (either in 'reduce' or during
// eliminating some variable which occurs in the moved clause.  This does
// not give a huge benefit, but seems to slightly increase performance (and
// counteracts the argument that BCE should be harmful in theory).

static void move_clause (Cls * c) {
  moved++;
  assert (!c->redundant);
  assert (opts::blockmove);
  assert (c->size () <= (unsigned) opts::blockmovesize);
  assert (clause.empty ());
  assert (!level);
  for (unsigned i = 0; i < c->size (); i++)
    clause.push_back (c->lits[i]);
  LOG ("moving clause " << *c);
  unsigned cidx;
  (void) add_but_do_not_watch_redundant_clause (0, cidx);
  assert (c->garbage);
  assert (!c->traced);
  c->traced = true;
  clause.clear ();
}

// This is trying to find blocked clauses with 'lit'.  If the second
// argument is non-zero those clause (indices) are stored in the
// 'BlockedSet' instead of being marked as garbage.  Instead they are marked
// as 'blocked'.  This allows to compute 'BCE(S)' of a set of clauses
// without actually removing them.  If 'bs' is zero, then the classical BCE
// is performed.

static unsigned block_lit (int lit) {
  LOG ("trying to block clauses on " << lit);
  Wchs & pws = wchs (lit);
  Wchs & nws = wchs (-lit);
  unsigned i = 0, res = 0;;
  for (unsigned j = 0; j < pws.size (); j++) {
    unsigned cidx = pws[i++] = pws[j];
    Cls * c = clauses[cidx];
    if (c->garbage || c->blocked || c->notblocked) { i--; continue; }
    if (c->nontriv) continue;
    unsigned pos = 0;
    while (c->lits[pos] != lit) pos++, assert (pos < c->size ());
    bool all_trivial;
    if (!nws.size ()) all_trivial = true;
    else {
      mark_clause (*c);
      unsigned didx = UINT_MAX, k;
      all_trivial = true;
      for (k = 0; all_trivial && k < nws.size (); k++) {
	std::swap (nws[k], didx);
	Cls * d = clauses[didx];
	if (d->garbage) continue;
	if (d->blocked) continue;
	if (d->notblocked) continue;
	blockresolutions++;
	bool trivial = false;
	for (unsigned l = 0; !trivial && l < d->size (); l++) {
	  int other = d->lits[l];
	  if (other == -lit) continue;
	  if (marked_lit (other) < 0) trivial = true;
	}
	if (trivial) continue;
	d->nontriv = 1;
	all_trivial = false;
      }
      assert (nws[0] == UINT_MAX);
      if (all_trivial) {
	assert (k == nws.size ());
	do std::swap (nws[k-1], didx); while (--k);
      } else {
	assert (didx != UINT_MAX);
	nws[0] = didx;			// move witness to front
      }
      unmark_clause (*c);
    }
    if (!all_trivial) continue;

    LOG ("blocked clause " << *c);
    c->garbage = true;
    trigger.elim = true;
    push_extend_clause (lit, c);
    touch_again_clause (c, lit);
    if (opts::blockmove &&
	c->size () <= (unsigned) opts::blockmovesize) move_clause (c);
    else trace_delete_clause (*c), c->traced = true;
    assert (c->traced);
    blocked++;
    res++;
    i--;
  }
  pws.resize (i);	// on-the-fly shrinking of watchers stacks
  LOG ("eliminated " << res << " blocked clauses with " << lit);
  return res;
}

static void reset_has_non_trivial_resolvents (int lit) {
  Wchs & ws = wchs (lit);
  for (unsigned i = 0; i < ws.size (); i++) {
    unsigned cidx = ws[i];
    Cls * c = clauses[cidx];
    c->nontriv = 0;
  }
}

static unsigned block_var (int idx) {
  unsigned count = 0;
  LOG ("blocked clause elimination for variable " << idx);
  blockvars++;
  int lim1 = opts::block1lim;
  int pos = actual_occurrences (idx, lim1);
  if (pos >= lim1) return 0;
  int neg = actual_occurrences (-idx, lim1);
  if (neg >= lim1) return 0;
  if (pos * (int64_t) neg >= opts::block2lim) return 0;
  int lit = (pos < neg ? -idx : idx);
  count += block_lit (lit);
  count += block_lit (-lit);
  reset_has_non_trivial_resolvents (lit);
  reset_has_non_trivial_resolvents (-lit);
  return count;
}

static bool run_blocked_clause_elimination () {
  check_all_clause_flags_clean ();	// 'actual_occurrences' requirement
  unsigned gc_limit = next_elim_gc_limit ();
  LOG ("run blocked clause elimination");
#ifdef NDEBUG
  for (unsigned i = 0; i < numirr; i++)
    assert (!clauses[i]->redundant);
  for (unsigned i = numirr; i < clauses.size (); i++)
    assert (clauses[i]->redundant);
#endif
  unsigned count = 0;
  int idx;
  while ((idx = next_elim_work ())) {
    elim_gc_if_necessary (gc_limit);
    count += block_var (idx);
  }
  untouch_all (); // redundant
  MSG (1, "block", blocks, "blocked " << count << " clauses");
  return count > 0;
}

static void block () {
  assert (!level);
  if (!opts::block) return;
  if (!trigger.block) return;
  if (delays.block.delayed ()) return;
  if (!garbage_collection ()) return;
  START (block);
  blocks++;
  LOG ("block " << blocks);
  trigger.block = false;
  disconnect ();
  connect_all_irredundant_clauses ();
  generate_elimination_schedule ();
  bool success = run_blocked_clause_elimination ();
  disconnect ();
  collect_garbage_clauses ();
  connect_watches ();
  assert (next == trail.size ());
  delays.block.update (success);
  report ('b');
  STOP (block);
}

/*------------------------------------------------------------------------*/

// The search state is mainly given through the order of the variables in
// the decision queue and the saved phases in 'phases'.  We save both and
// restore it after 'simplify'.  Clauses learned during 'simplify' are just
// handled the same way as during search.  So we separated 'simpreds' from
// actual search reductions 'reductions'.  The restart EMAs are not updated
// during 'simplify' either and we simply restart with the base interval
// frequency.  Otherwise there is only one global 'conflicts' counter, both
// for search and simplification.  Beside 'reduce' and 'restart' this
// counter influences only simplification scheduling, which however is
// updated anyhow after simplification.  So we hope that this saves and
// restores enough of the important state of search.

static void save_search_state () {
  flush_root_trail ();
  LOG ("saving search space");
  ALLOC (Val, saved.phases, maxvar+1);
  for (int i = 1; i <= maxvar; i++) saved.phases[i] = phases[i];
  int count = 0;
  for (int i = queue.first; i; i = lnks[i].next)
    saved.queue.push_back (i), count++;
  LOG ("saved " << count << " variables on queue");
}

static void copy_back_and_release_saved_search_state () {
  flush_root_trail ();
  LOG ("restoring saved search space");
  for (int i = 1; i <= maxvar; i++) phases[i] = saved.phases[i];
  DEALLOC (Val, saved.phases, maxvar+1);
#ifndef NDEBUG
  for (int i = 1; i <= maxvar; i++) lnks[i].prev = lnks[i].next = 0;
#endif
  queue.first = queue.last = queue.searched = 0;
  queue.stamp = 0;
  int count = 0;
  for (unsigned i = 0; i < saved.queue.size (); i++) {
    int idx = saved.queue[i];
    if (!is_active_lit (idx)) continue;
    queue.enqueue (idx);
    count++;
  }
  queue.searched = queue.last;
  release (saved.queue);
  LOG ("enqueued back " << count << " variables");
}

/*------------------------------------------------------------------------*/

// Even though the backward subsumption techniques described in our SAT'05
// paper on bounded variable elimination can be considered as the first one
// reported to make subsumption work on a large number of clauses, even in
// combination with self-subsuming resolution, it was not applied in
// state-of-the-art solvers for learned clauses as far we know (see however
// 'subsume_recently_learned').  Further, as observed by Bayardo and Panda
// in their SDM'11 paper, our original algorithm is still not the fastest
// possible one, particularly when applied to the slightly more general
// problem of finding extremal sets.  They give two improved algorithms.
// The first one uses a cardinality argument and actually performs and
// reduce the number of watchers to one per clause.  This is similar to
// Lintao Zhang's forward-subsumption technique also using one watcher.  The
// Baryardo and Panda algorithm also sorts literals in clauses based on the
// total number of occurrences instead of using marking algorithms.  We
// implemented a variant of their first cardinality based algorithm here.
// It is both applied to redundant and irredundant clauses.  One of the main
// motivations was to be able remove subsumed sticky clauses.  The point is
// that the original MiniSAT implementation without sticky clauses (say glue
// <= 2), would implicitly remove the subsumed redundant clauses, since they
// would become inactive with high probability very soon due to the fact
// that the subsuming clauses would be used in propagations and conflict
// analysis instead of the subsumed clauses.  Then these subsumed clauses
// will neever be bumped (again) and at one point recycled during 'reduce'.
// As soon we use sticky clauses, this mechanism stops working for sticky
// clauses and subsumed sticky clauses have to be removed explicitly.  We
// also hoped to be able to shrink learned clauses this way eagerly and
// maybe even get a faster subsumption algorithm for irredundant clauses.

static unsigned * occs;

struct occs_less_than {
  bool operator () (int a, int b) {
    assert (a != -b);
    unsigned p = occs[abs (a)], q = occs[abs (b)];
    if (p < q) return true;
    if (p > q) return false;
    return abs (a) < abs (b);
  }
};

struct subsume_less_than {
  bool operator () (unsigned cidx, unsigned didx) {
    Cls * c = clauses[cidx], * d = clauses[didx];
    if (c->size () > d->size ()) return true;
    if (c->size () < d->size ()) return false;
    if (!c->redundant && d->redundant) return true;
    if (c->redundant && !d->redundant) return false;
    return cidx > didx;
  }
};

static bool
subsume_include_clause (Cls * c,
                       bool irr, bool red, 
		       const unsigned maxsize) {
  assert (irr || red);
  if (c->garbage) return false;
  if (!irr && !c->redundant) return false;
  if (!red && c->redundant) return false;
  if (c->size () > maxsize) return false;
  return true;
}

static void generate_subsume_schedule (bool irr, bool red) {
  CALLOC (unsigned, occs, maxvar+1);
  const unsigned maxsize = opts::subsumeclslim;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (!subsume_include_clause (c, irr, red, maxsize)) continue;
    for (unsigned j = 0; j < c->size (); j++)
      occs[abs (c->lits[j])]++;
  }
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (!subsume_include_clause (c, irr, red, maxsize)) continue;
    std::sort (c->lits, c->lits + c->size (), occs_less_than ());
  }
  assert (work.empty ());
  unsigned maxi = 0;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (!subsume_include_clause (c, irr, red, maxsize)) continue;
    work.push_back (i);
    maxi = i;
  }
  std::sort (work.begin (), work.end (), subsume_less_than ());
  szclsigs = maxi + 1;
  CALLOC (Sig, clsigs, szclsigs);
  MSG (2, "subsume", subsumes,
    "scheduled " << work.size () << " clauses");
}

static bool subsume_clauses_match (Cls * c, Cls * d, int & r) {
  assert (!c->garbage);
  assert (!d->garbage);
  assert (c->size () == d->size ());
  r = 0;
  for (unsigned i = 0; i < c->size (); i++) {
    int l = c->lits[i], k = d->lits[i];
    if (l == k) continue;
    if (l != -k || r) { r = 0; return false; }
    r = k;
  }
  return r ? false : true;
}

// This does the actual subsumption check between two clauses.  It assumes
// the two clauses are sorted w.r.t. to the same order and thus the check is
// linear in the size of the two clauses. Note that if a literal from the
// smaller clause is not found in the larger clauses (or its negation) the
// whole larger clause has to be traversed.  We could use binary search as
// well in the case the larger clauses is really way larger than the smaller
// one.  However, in that case, since we schedule subsumption checks from
// smaller to larger clauses, the larger clause is the main considered
// clause in 'subsume_clause' anyhow, thus marked and we can use the
// faster marking based algorithm below, which is linear in the smaller
// clause.  This last comment is actually invalid for rescheduled shrunken
// clauses. For those on the other hand, at least if they are just
// rescheduled once, the difference in size is just one literal.

static bool
subsume_clause_subsumes_clause_sorted (Cls * c, Cls * d, int & r) {
  assert (!c->garbage);
  assert (!d->garbage);
  assert (c->size () <= d->size ());
  const unsigned s = c->size (), t = d->size ();
  unsigned i = 0, j = 0, except = d->size () - c->size ();
  r = 0;
  for (;;) {
    if (i == s) return !r;
    int lit = c->lits[i++];
    // TODO: binary search if 'd->size ()' way bigger than 'c->size ()'
    // or 't - j' big enough ... needs to keep occurrences around ...
    for (;;) {
      if (j == t) { r = 0; return false; }
      int other = d->lits[j++];
      if (other == lit) break;
      if (other == -lit) {
	if (r) { r = 0; return false; }
	r = other;
	break;
      }
      if (occs_less_than () (lit, other)) { r = 0; return false; }
      if (!except--) { r = 0; return false; }
    }
  }
}

static Cls * subsume_marked_clause;

// This is a marking based forward subsumption algorithm.  So in contrast to
// 'can_subsume_clause' the larger clause is assumed to be marked, which
// however gives nice complexity linear in the size of the smaller clause.

static bool
subsume_clause_subsumes_clause_marked (Cls * c, int & r) {
  const unsigned size = c->size ();
  assert (!c->garbage);
  assert (subsume_marked_clause);
  assert (c != subsume_marked_clause);
  r = 0;
  for (unsigned i = 0; i < size; i++) {
    int lit = c->lits[i];
    int tmp = marked_lit (lit);
    if (tmp > 0) continue;
    if (!tmp) { r = 0; return false; }
    assert (tmp < 0);
    if (r) { r = 0; return false; }
    r = -lit;				// flipped semantics!!!
  }
  return !r;
}

static bool
subsume_clause_subsumes_clause (Cls * c, Cls * d, int & r) {
  assert (!c->garbage);
  assert (!d->garbage);
  assert (c->size () < d->size ());
  if (d == subsume_marked_clause) {
    return subsume_clause_subsumes_clause_marked (c, r);
  } else {
    assert (!subsume_marked_clause || c == subsume_marked_clause);
    return subsume_clause_subsumes_clause_sorted (c, d, r);
  }
}

// Turn a redundant clause into a irredundant one.  This might break some
// invariants, which can be checked by monitoring the 'makeirrs' counter.
// See 'move_illplaced_irredundant_clauses' for how to fix this problem.

static Cls * make_irredundant (Cls * c, unsigned cidx) {
  assert (!c->garbage);
  assert (c->redundant);
  assert (clauses[cidx] == c);
  const unsigned size = c->size ();
  size_t bytes = sizeof (Cls) + (size - CLSLITS) * sizeof (int);
  Cls * d = (Cls *) allocate (bytes);
  init_clause_header (d, false, size, 0);
  for (unsigned i = 0; i < size; i++) d->lits[i] = c->lits[i];
  if (subsume_marked_clause == c) subsume_marked_clause = d;
  dealloc_redundant_clause (c);
  clauses[cidx] = d;
  makeirrs++;
  numirr++;
  return d;
}

// This Bloom filter for clauses is an overapproximation of the set of
// variables in a clause.  In our SAT'05 paper we called them 'signatures'
// which is also a common name for such Bloom filters.

static Sig sig_clause (Cls * c) {
  assert (!c->garbage);
  Sig res = 0;
  for (unsigned i = 0; i < c->size (); i++)
    res |= varsig (c->lits[i]);
  return res;
}

static void subsume_shrink_clause (Cls * c, unsigned cidx, int r) {
  assert (!c->garbage);
  assert (clauses[cidx] == c);
  if (subsume_marked_clause == c) {
    unmark_clause (*subsume_marked_clause);
    subsume_marked_clause = 0;
  }
  LOG ("shrinking removes " << r << " from " << *c);
  copy (unshrunken, *c);
  unsigned old_size = c->size (), new_size = 0;
  for (unsigned l = 0; l < old_size; l++) {
    int other = c->lits[l];
    if (other == r) continue;
    c->lits[new_size++] = other;
  }
  assert (new_size + 1 == old_size);
  trigger.elim = trigger.block = true;
  if (new_size > 1) {
    c = realloc_clause (c, old_size, new_size);
    trace_add_learned_clause (*c);
    clauses[cidx] = c;
    if (new_size == 2) trigger_binary ();
    resched.push_back (cidx);
    clsigs[cidx] = sig_clause (c);
  } else {
    assert (new_size == 1);
    int unit = c->lits[0];
    c->garbage = true;
    c->traced = true;
    trace_add_unit_learned_clause (unit);
    LOG ("shrunken to unit clause " << unit);
    if (is_fixed_lit (unit)) units.push_back (unit);
    else assign (unit, Unit ());
  }
  trace_delete_clause (unshrunken);
  shrunken++;
  subsumed++;
}

static void subsume_clause_by_clause (unsigned cidx, unsigned didx) {
  Sig csig = clsigs[cidx];
  Sig dsig = clsigs[didx];
  subsumechecks++;
  if (!subsig (csig, dsig) && !subsig (dsig, csig)) return;
  Cls * c = clauses[cidx];
  Cls * d = clauses[didx];
  if (c->garbage) return;
  if (d->garbage) return;
  if (c->size () > d->size () ||
      (c->size () == d->size () && c->redundant && !d->redundant)) {
    std::swap (c, d);
    std::swap (cidx, didx);
    std::swap (csig, dsig);
  }
  assert (c->size () <= d->size ());
  assert (clauses[cidx] == c);
  assert (clauses[didx] == d);
  if (!subsig (csig, dsig)) return;
  bool tmp;
  int r;
  if (c->size () == d->size ()) tmp = subsume_clauses_match (c, d, r);
  else tmp = subsume_clause_subsumes_clause (c, d, r);
  if (tmp) {
    if (d->redundant) subsumedred++;
    else if (c->redundant && !d->redundant) {
      assert (c->size () < d->size ());
      c = make_irredundant (c, cidx);
      subsumedred++;
    } else subsumedirr++;
    LOG ("subsuming clause " << *c);
    LOG ("subsumed clause " << *d);
    trace_delete_clause (*d);
    assert (!d->garbage);
    assert (!d->traced);
    d->garbage = true;
    d->traced = true;
    subsumed++;
  } else if (r && c->size () == d->size ()) {
    assert (!c->redundant || d->redundant);
    LOG ("double shrinking antecedent " << *d);
    subsume_shrink_clause (c, cidx, -r);
    trace_delete_clause (*d);
    assert (!d->garbage);
    assert (!d->traced);
    d->garbage = true;
    d->traced = true;
    if (d->redundant) subsumedred++; else subsumedirr++;
    subsumed++;
  } else if (r && c->size () < d->size ()) {
    LOG ("shrinking antecedent " << *c);
    subsume_shrink_clause (d, didx, r);
  }
}

static void subsume_clause (unsigned cidx) {
  Cls * c = clauses[cidx];
  assert (!c->garbage);
  assert (c->size () <= (unsigned) opts::subsumeclslim);
  assert (!subsume_marked_clause);
  if (c->size () > 2) {
    subsume_marked_clause = c;
    mark_clause (*c);
  }
  unsigned minsize = UINT_MAX;
  int minlit = 0;
  LOG ("considering size " << c->size () << " clause " << *c);
  clsigs[cidx] = sig_clause (c);
  unsigned csize = c->size ();
  bool done = false;
  for (unsigned i = 0; !done && i < c->size (); i++) {
    for (int sign = 1; !done && sign >= -1; sign -= 2) {
      int lit = sign*c->lits[i];
      Wchs & ws = wchs (lit);
      if (ws.size () > (unsigned) opts::subsumeocclim) continue;
      unsigned j = 0;
      for (unsigned k = 0; k < ws.size (); k++) {
	unsigned didx = ws[j++] = ws[k];
	if (cidx == didx) { j--; continue; }
	if (done) continue;
	if (clauses[didx]->garbage) { j--; continue; }
	subsume_clause_by_clause (cidx, didx);
	c = clauses[cidx];
        if (c->garbage || (c->size () < csize)) done = true;
      }
      ws.resize (j);
      if (!done && sign > 0 && ws.size () < minsize)
	minlit = lit, minsize = ws.size ();
    }
  }
  if (minlit && !done) {
    LOG ("watching literal " << minlit << " with " <<
      minsize << " occurrences in size " << c->size () << " clause " << *c);
    wchs (minlit).push_back (cidx);
  }
  if (subsume_marked_clause) {
    unmark_clause (*subsume_marked_clause);
    subsume_marked_clause = 0;
  }
}

// At several places we use the invariant that irredundant clauses have
// smaller clause indices than redundant ones.  This invariant is usually
// broken if 'make_irredundant' is called and thus we have to fix again the
// order if this was necessary.

static void move_illplaced_irredundant_clauses () {
  unsigned r;
  for (r = 0; r < clauses.size (); r++)
    if (clauses[r]->redundant) break;
  unsigned i = r;
  sector<Cls *> redundant;
  for (unsigned j = r; j < clauses.size (); j++) {
    Cls * c = clauses[i++] = clauses[j];
    if (!c->redundant) continue;
    redundant.push_back (c);
    i--;
  }
  clauses.resize (i);
  for (unsigned j = 0; j < redundant.size (); j++)
    clauses.push_back (redundant[j]);
  release (redundant);
}

/*------------------------------------------------------------------------*/

// Our new subsumption 'subsume' algorithm, which is also run during search
// to remove subsumed and shrink learned clauses, sorts literals in clauses
// for fast subsumption checking instead of using marking.  In order to
// avoid influencing the search and particularly to counteract the effect
// sorting would have on the effectivenes of the 'arena' idea we save the
// first two watched literals of all clauses before sorting in 'subsume' and
// restore them afterwards.

static int * saved_watches;
static unsigned size_saved_watches;

static void save_watches () {
  size_saved_watches = clauses.size ();
  assert (size_saved_watches < (1u<<31));
  ALLOC (int, saved_watches, 2*size_saved_watches);
  int * p = saved_watches;
  for (unsigned i = 0; i < size_saved_watches; i++) {
    Cls * c = clauses[i];
    *p++ = c->lits[0];
    *p++ = c->lits[1];
  }
  assert (p == saved_watches + 2*size_saved_watches);
}

static void restore_watches () {
  assert (size_saved_watches == clauses.size ());
  const int * p = saved_watches;
  for (unsigned i = 0; i < size_saved_watches; i++, p += 2) {
    Cls * c = clauses[i];
    if (c->garbage) continue;
    const unsigned size = c->size ();
    for (unsigned i = 0; i < 2; i++) {
      unsigned j;
      for (j = 0; j < size; j++)
	if (c->lits[j] == p[i])
	  break;
      if (j == size) continue;
      if (i == j) continue;
      std::swap (c->lits[i], c->lits[j]);
    }
  }
  assert (p == saved_watches + 2*size_saved_watches);
  DEALLOC (int, saved_watches, 2*size_saved_watches);
}

/*------------------------------------------------------------------------*/

static bool next_scheduled_cidx (unsigned & cidx) {
  if (!resched.empty ()) {
    cidx = resched.back ();
    resched.pop_back ();
    return true;
  }
  if (!work.empty ()) {
    cidx = work.back ();
    work.pop_back ();
    return true;
  }
  return false;
}

/*------------------------------------------------------------------------*/

static bool subsume_clauses (bool irr, bool red) {
  assert (irr || red);
  check_all_vars_unmarked ();

  int64_t makeirrs_before = makeirrs;		// see below ...

  // We allow 1000 times more subsumption checks than the number of
  // scheduled clauses, so for each scheduled clause 1000 checks.

  int64_t scheduled = work.size (), delta = 1000*scheduled;
  if (simpmode) {
    // But during 'simplification' (inbetween 'elim' calls) the number of
    // subsumption checks is limited by the number of BVE resolutions.
    int64_t delta_res = elimresolutions - lastelimres;
    delta *= opts::subsumechecklim; delta /= 1000;
    if (delta_res < delta) delta = delta_res;
  }
  if (delta < opts::subsumecheckminlim) delta = opts::subsumecheckminlim;
  int64_t limit = subsumechecks + delta;

  const unsigned subsumedirr_before = subsumedirr;
  const unsigned subsumedred_before = subsumedred;
  const unsigned shrunken_before = shrunken;

  unsigned tried = 0;

  assert (units.empty ());
  assert (resched.empty ());

  unsigned cidx;
  while (subsumechecks < limit && next_scheduled_cidx (cidx))
    if (!clauses[cidx]->garbage)
      subsume_clause (cidx), tried++;

  release (work);
  release (resched);

  DEALLOC (unsigned, occs, maxvar+1);
  DEALLOC (Sig, clsigs, szclsigs);

  // Subsumption of an irredundant clause by a redundant clause is tricky to
  // handle since we want to maintain the invariant, that all irredundant
  // clauses have smaller clause indices than redundant ones.  We use the
  // global counter 'makeirrs' to detect when this happen and then fix the
  // invariant explicitly (see 'make_irredundant' too).

  if (makeirrs_before < makeirrs) move_illplaced_irredundant_clauses ();

  unsigned subirr = subsumedirr - subsumedirr_before;
  unsigned subred = subsumedred - subsumedred_before;
  unsigned str = shrunken - shrunken_before;

  if (subirr) trigger.elim = trigger.block = true;

  if (subsumechecks < limit)
    MSG (2, "subsume", subsumes,
      "completed after " << tried << " clauses");
  else
    MSG (2, "subsume", subsumes, 
      "incomplete after " << tried << " clauses " <<
      percent (tried, scheduled) << '%');
  const char * complete =
    (subsumechecks < limit) ? "complete" : "incomplete";
  if (irr && red) {
    MSG (1, "subsume", subsumes,
      "subsumed " << subirr + subred << " = " << 
      subirr << " (irr) + " <<
      subred << " (red), " <<
      str << " shrunken (" << complete << ')');
  } else if (irr) {
    assert (!subred);
    MSG (1, "subsume", subsumes,
      "subsumed " << subirr << " irredundant clauses, " <<
      str << " shrunken (" << complete << ')');
  } else {
    assert (red);
    assert (!subirr);
    MSG (1, "subsume", subsumes,
      "subsumed " << subred << " redundant clauses, " <<
      str << " shrunken (" << complete << ')');
  }
  check_all_vars_unmarked ();
  return subirr > 0;
}

static void subsume (bool redtoo_in_simpmode = false) {
  if (!opts::subsume) return;
  if (simpmode && !redtoo_in_simpmode) {
    if (!trigger.subsume) return;
    if (delays.subsume.delayed ()) return;
  }
  if (level) backtrack ();
  if (!garbage_collection ()) return;
  START (subsume);
  subsumes++;
  disconnect ();
  bool irr = simpmode;
  bool red = !simpmode || redtoo_in_simpmode;
  assert (irr || red);
  if (irr && red) LOG ("subsumption checking all clauses " << subsumes);
  else if (irr) LOG ("subsumption checking irredundant clauses " << subsumes);
  else LOG ("subsumption checking redundant clauses " << subsumes);
  generate_subsume_schedule (irr, red);
  save_watches ();
  bool success = subsume_clauses (irr, red);
  restore_watches ();
  disconnect ();
  collect_garbage_clauses ();
  connect_watches ();
  assign_units_and_propagate ();
  if (simpmode) delays.subsume.update (success || redtoo_in_simpmode);
  else inc.conflicts.subsume += opts::subsumeinc;
  limits.subsume.conflicts = conflicts + inc.conflicts.subsume; 
  limits.subsume.reductions = 0;
  report ('s');
  STOP (subsume);
  if (!simpmode) copy_and_move_arena ();
}

static bool subsuming () {
  if (!opts::subsume) return false;
  if (!limits.subsume.conflicts)
    limits.subsume.conflicts = inc.conflicts.subsume = opts::subsumeint;
  if (conflicts < limits.subsume.conflicts) return false;
  if (conflicts > limits.subsume.conflicts + 100000) return true;
  if (level) return false;
  if (!limits.subsume.reductions) limits.subsume.reductions = reductions + 1;
  return reductions >= limits.subsume.reductions;
}

/*------------------------------------------------------------------------*/

// As in Lingeling we schedule all the inprocessors from one 'simplify'
// function which is scheduled based on three criteria beside waiting that
// the current decision level is zero anyhow (to avoid artificial restarts).
// First based on a (in principle) geometrically increasing conflict
// interval we wait for a conflict limit. Otherwise if a number of unit
// clauses or binary clauses has been learned we schedule additional
// simplifications.  The number of required learned units and binary clauses
// between such events grows geometrically as well.  The actual conflict
// interval between two consecutive simpliciation can be substantially
// smaller based on the success of the previous simplification measured in
// terms of the amount of reduction (removed variables and clauses).  This
// is described in more detail in 'simplify_update_limits'.

static bool simplifying () {
  if (!opts::simplify) return false;
  if (!level) return false;
  if (simplifications > 1 && !opts::inprocessing) return false;
  if (conflicts >= limits.simplify.conflicts) {
    report ('*');
    MSG (1, "simplify", simplifications+1,
      "limit hit after " << conflicts << " conflicts");
    simpreason = 0;
    return true;
  }
  if (!inc.simplify.iterations) {
    inc.simplify.iterations = opts::simplifyitliminit;
    limits.simplify.iterations = iterations + inc.simplify.iterations;
  }
  if (iterations >= limits.simplify.iterations) {
    report ('*');
    MSG (1, "simplify", simplifications+1,
      "limit hit after " << iterations << " learned units");
    inc.simplify.iterations *= opts::simplifyitlimfactor;
    limits.simplify.iterations = iterations + inc.simplify.iterations;
    simpreason = 1;
    return true;
  }
  if (!inc.simplify.learnedbins) {
    inc.simplify.learnedbins = opts::simplifybinliminit;
    limits.simplify.learnedbins = learnedbins + inc.simplify.learnedbins;
  }
  if (learnedbins >= limits.simplify.learnedbins) {
    report ('*');
    MSG (1, "simplify", simplifications+1,
	"limit hit after " << learnedbins << " learned binary clauses ");
    inc.simplify.learnedbins *= opts::simplifybinlimfactor;
    limits.simplify.learnedbins = learnedbins + inc.simplify.learnedbins;
    simpreason = 2;
    return true;
  }
  return false;
}

// This is in essence the Lingeling policy described in our POS'14 talk.  It
// is pretty important to get this right.  Otherwise inprocessing becomes
// useless, since too much time or not too much (if important preprocessors
// are preempted) is spent in preprocessors/inprocessors.

static void
simplify_and_update_limits_only (int64_t var_before, int64_t cls_before) {
  if (!var_before || !cls_before) {
    MSG (1, "simplify", simplifications,
      "no change in conflict limit since formula was trivial");
    return;
  }
  if (simpreason) {
    MSG (1, "simplify", simplifications,
      "keeping old conflict limit " << limits.simplify.conflicts);
    return;
  }
  int64_t var_after = remaining_variables ();
  int64_t cls_after = numirr;
  int64_t var_rem = var_before - var_after;
  int64_t cls_rem = cls_before - cls_after;
  MSG (1, "simplify", simplifications, "removed " <<
    var_rem << " variables " << percent (var_rem, var_before) << '%' <<
    " (" << var_after << " left " << percent (var_after, maxvar) << "%)");
  MSG (1, "simplify", simplifications, "removed " <<
    cls_rem << " clauses " << percent (cls_rem, cls_before) << '%' <<
    " (" << cls_after << " left " << percent (cls_after, maxcls) << "%)");
  int64_t var_rel_rem = (100*var_rem + 50)/var_before;
  int64_t cls_rel_rem = (100*cls_rem + 50)/cls_before;
  int64_t relrem = std::max (var_rel_rem, cls_rel_rem);
  int64_t initial_delta = inc.simplify.conflicts;
  int64_t delta = initial_delta;
  if (relrem > 1) {
    delta /= relrem;
    MSG (1, "simplify", simplifications,
      "initial delta " << initial_delta << " reduced by " << relrem <<
      " to " << delta);
  } else {
    MSG (1, "simplify", simplifications,
      "initial delta " << initial_delta << " not reduced");
  }
  if (!delta) delta++;
  if (delta < opts::simplifyminint) {
    delta = opts::simplifyminint;
    MSG (1, "simplify", simplifications, "using minimum delta " << delta);
  } else if (delta > opts::simplifyincsoft &&
             (var_after < var_before || cls_after < cls_before)) {
    delta = opts::simplifyincsoft;
    MSG (1, "simplify", simplifications, "using soft delta limit " << delta);
  } else if (delta > opts::simplifyinchard) {
    delta = opts::simplifyinchard;
    MSG (1, "simplify", simplifications, "using hard delta limit " << delta);
  } else {
    int64_t old_inc = inc.simplify.conflicts;
    int64_t new_inc = old_inc + delta;
    MSG (1, "simplify", simplifications,
      "increasing delta increment from " << old_inc << " to " << new_inc);
    inc.simplify.conflicts = new_inc;
  }
  limits.simplify.conflicts = conflicts + delta;
  lastprops = propagations;
  MSG (1, "simplify", simplifications,
    "next conflict limit " << limits.simplify.conflicts <<
    " delta " << delta);
}

static void dump_to_output (int64_t var_before, int64_t cls_before) {
  assert (output);
  if (outputs &&
      remaining_variables () == var_before &&
      numirr == cls_before) {
    MSG (2, "output", outputs,
      "no change (keeping '" << output << "')");
    return;
  }
  outputs++;
  unsigned m = maxvar, n = 0;
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->garbage || c->redundant) continue;
    n++;
  }
  assert (n == numirr);
  size_t len = strlen (output) + 2;
  char * backup;
  if (file_exists (output)) {
    backup = (char*) allocate (len);
    strcpy (backup, output);
    strcat (backup, "~");
    delete_file (backup);
    rename_file (output, backup);
  } else backup = 0;
  FILE * file = fopen (output, "w");
  if (!file) DIE ("failed to write to '" << output << "'");
  MSG (1, "output", outputs, "writing '" << output << "'");
  fprintf (file, "p cnf %u %u\n", m, n);
  for (unsigned i = 0; i < clauses.size (); i++) {
    Cls * c = clauses[i];
    if (c->garbage || c->redundant) continue;
    for (unsigned j = 0; j < c->size (); j++)
      fprintf (file, "%d ", c->lits[j]);
    fputs ("0\n", file);
  }
  fclose (file);
  if (backup) {
    delete_file (backup);
    deallocate (backup, len);
  }
  MSG (1, "output", outputs,
    "wrote 'p cnf " << m << ' ' << n << "' (" <<
    file_bytes (output)/(double)(1<<20) << " MB)");
}

static void simplify_update_limits (int64_t var_before, int64_t cls_before) {
  simplify_and_update_limits_only (var_before, cls_before);
  if (output) dump_to_output (var_before, cls_before);
}

static void simplify () {
  int var_before = remaining_variables (), cls_before = (int) numirr;
  if (!something_triggered ()) {
    simplifications++;
    MSG (1, "simplify", simplifications, "nothing triggered");
    simplify_update_limits (var_before, cls_before);
    return;
  }
  STOP (search);
  START (simplify);
  assert (!simpmode);
  simpmode = true;
  simplifications++;
  LOG ("simplify " << simplifications);
  if (level) backtrack ();

  save_search_state ();

  garbage_collection ();	// one before if necessary

  probe ();		// 'p' = probe roots in binary implication graph
  sweep ();		// 'w' = blocked clause decomposition based sweeping
  decompose ();		// 'd' = strongly connected component decomposition
  			//       and equivalent literal substitution
  subsume (true);	// 's' = subsume both redundant and irredundant

  for (int round = 1;
       !inconsistent &&
       round <= opts::elimrounds &&
       (trigger.elim || trigger.subsume || trigger.block);
       round++) {
    elim ();		// 'e' = bounded variable elimination
    subsume ();		// 's' = subsumption checking irredundant only
    block ();		// 'b' = blocked clause elimination
  }
  garbage_collection ();	// and one after all if necessary
  copy_and_move_arena ();	// compact memory layout 

  copy_back_and_release_saved_search_state ();

  simplify_update_limits (var_before, cls_before);
  simpmode = false;
  STOP (simplify);
  START (search);
}

/*------------------------------------------------------------------------*/

// Variant of the Glucose inspired restart policy, which only uses
// exponential moving averages, as described in our POS'15 paper.

static bool restarting () {

  if (!opts::restart) return false;

  if (!level) return false;	   // might otherwise give infinite looping

  // The 'restart_interval ()' interval is a lower bound.

  if (conflicts < limits.conflicts.restart)  
    return false;

  if (opts::restartforce &&
      ema.glue.slow.cmpema (ema.glue.fast, opts::restartforcelim) > 0) {
    report ('N');
    LOG ("restart not forced even though limit hit");
    limits.conflicts.restart = conflicts + opts::restartint;
    restartsnoforce++;
    restartskipped++;
    return false;
  }

  // This idea was first used in Lingeling and is similar in spirit as
  // the blocking idea of Glucose.  Search for 'restartsblocked' above.
  // It checks whether we have a reasonably high decision level before
  // restarting.  A typical value is 50% for 'opts::restartdelaylim' of the
  // (moving) average of decisions levels at conflicts, for which the
  // average jump level is actually a good approximation and maybe
  // intuitively anyhow the better metric to compare to.  The decision or
  // jump level averages should move slow, e.g., 'opts::jumpema == 20'.

  if (opts::restartdelay &&
      conflicts > opts::restartdelaybound &&
      ema.jump.cmpu32 (level, opts::restartdelaylim) > 0) {
    report ('D');
    LOG ("restart delayed even though limit hit and forced");
    limits.conflicts.restart = conflicts + opts::restartint;
    restartsdelayed++;
    restartskipped++;
    return false;
  }

  // Note that these EMAs are 64 bit fixed point values with 32 bit
  // fractional part and thus dividing by 100 should only initially result
  // in a zero value. So we do not have to multiply as at other places where
  // we use a percentage option.

  return true;
}

// Marijn Heule observed that sometimes restarts would not change some part
// of the trail.  So instead of completely restarting it is possible to
// keep those decision levels with decisions, which would be picked anyhow
// before the current next decision before restarting is taken.

static int reuse_trail () {
  if (!opts::reusetrail) return 0;
  int lit = next_decision ();
  unsigned stamp = lnk (lit).stamp, res;
  for (res = 1; res < control.size (); res++) {
    if (lnk (control[res].decision).stamp < stamp) break;
    LOG ("reusing trail " << res);
    reusedlevels++;
  }
  res--;
  assert (0 <= res), assert (res <= level);
  if (res) reusedtrails++, notreusedlevels += level - res;
  if (res == level) {
    // Redundant, but keep it here as comment:
    // limits.conflicts.restart = conflicts + opts::restartint;
    restartsreused++;
    restartskipped++;
  }
  return res;
}

static void restart () {
  START (restart);
  restarts++;
  LOG ("restart " << restarts);
  backtrack (reuse_trail ());
  limits.conflicts.restart = conflicts + opts::restartint;
  report ('R');
  STOP (restart);
}

/*------------------------------------------------------------------------*/

// The learned clause data base reduction 'reduce' removes learned clauses
// predicted not to be beneficial.  Since the original Glucose paper
// from Gilles Audemard and Laurent Simon in IJCAI'09 it became standard to
// use the 'glucose level' or short 'glue' (in the paper they call it 'LBD')
// as predication.  It counts the number of decision levels in the learned
// clause, which allow it to become propagating (flip the first UIP).  Small
// glue is better.  In contrast to MiniSAT, which only uses activities
// (kind of inverse exponential averages), this scheme tends to give better
// prediction.  Small glue clauses can actually be kept forever.  This was
// discussed for instance recently in Chanseok Oh's SAT'15 paper, who
// suggested to use a large glue limit for instance for cryptographic
// instances.  In principle our scheme is similar, but instead of using a
// fixed glue limit we compute the average glue of resolved clauses and
// their size in conflict analysis.  If a clause has better glue and is
// smaller than this average it is kept unconditionally for the given
// reduction.  From the remaining clauses we remove half of the clauses
// using a least-recently used 'resolved' counter and further also keep
// clauses which have been resolved since the last reduction ('active').
// This seems to give a pretty robust heuristic.

static bool reducing () {
  if (!opts::reduce) return false;
  if (conflicts < limits.conflicts.reduce) return false;
  return true;
}

// The same comments as for 'resolved_less_than' apply here, e.g.,
// smaller working stack and stable sorting are favored over speed.

struct reduce_less_than {
  bool operator () (const unsigned & i, const unsigned & j) const {
    Red * r = cls2red (clauses[i]);
    Red * s = cls2red (clauses[j]);
    if (r->resolved < s->resolved) return true;
    if (r->resolved > s->resolved) return false;
    if (r->cls.size () > s->cls.size ()) return true;
    if (r->cls.size () < s->cls.size ()) return false;
    return i < j;;
  }
};

static unsigned first_redundant_clause () {
#ifndef NDEBUG
  const unsigned s = clauses.size ();
  unsigned o = 0;
  while (o < s && !clauses[o]->redundant) o++;
  assert (o == numirr);
#else
  unsigned o = numirr;
#endif
  assert (o + numred == s);
  return o;
}

static void reduce_protect_reasons () {
  for (unsigned i = 0; i < trail.size (); i++) {	// protect reasons
    Var & v = var (trail[i]);
    if (!v.level) continue;
    const Rsn & r = v.reason;		// of lits on trail
    if (r.binary ()) continue;
    if (r.decision ()) continue;
    assert (r.clausal ());
    Cls * c = r.cls ();
    if(c->garbage) c->garbage = 0;
  }
}

// This is a new idea in this solver.  We consider clauses to be important,
// if their size and glue is both below the average of resolved clauses.
// Those are not thrown away.  In the original Glucose implementation only
// clauses with glue 2 (but beware of off by one descripancies in
// definitions here) were not reduced. Chanseok Oh and also others tries
// also much larger fixed glue values (like 5).  It seems this new scheme
// here is robust and works as least as good, while saving on memory for
// some instance, where keeping glue 5 clauses would be really bad.

static bool sticky_clause (Cls * c) {
  unsigned glue = ema.resolved.glue.nval ();
  unsigned size = ema.resolved.size.nval ();
  assert (c->redundant);
  Red * r = cls2red (c);
  return c->size () < size && r->glue < glue;
}

// Mark clauses to be collected during 'reduce'.

static void reduce_mark () {
  assert (work.empty ());
  const unsigned s = clauses.size ();
  const unsigned o = first_redundant_clause ();
  for (unsigned i = o; i < s; i++) {		// collect candidates
    Cls * c = clauses[i];
    if (!c->redundant) continue;	// only reduce redundant clauses
    if (c->garbage) continue;		// will be collected anyhow
    if (c->size () <= (unsigned) opts::keep) continue;
    if (sticky_clause (c)) continue;
    Red * r = cls2red (c);
    if (r->active) r->active = 0; // resolved / bumped since last reduce?
    else work.push_back (i);
  }
  std::sort (work.begin (), work.end (), reduce_less_than ());
  const unsigned target = work.size () / 2;
  for (unsigned i = 0; i < target; i++)		// mark half as garbage
    clauses[work[i]]->garbage = true;
  work.clear ();
  reduce_protect_reasons ();
}

// Move and collect phase of 'reduce'.

static unsigned reduce_move () {
  const unsigned s = clauses.size ();
  const unsigned o = first_redundant_clause ();

  // Use 'work' as map from old indices to new indices.  We only need
  // to map redundant clauses and thus can restrict this procedure to work
  // on the subset of clauses starting from the first redundant clause to
  // the end of all clauses.

  assert (work.empty ());
  while (s - o > work.size ()) work.push_back (0);

  // Moving garbage collector, moves non-garbage clauses to lower indices.

  unsigned i, j, t = 0;
  for (i = t = o; i < s; i++) {
    Cls * c = clauses[i];				// src 'i'
    work[i - o] = (c->garbage ? UINT_MAX : t++);	// dst 'work[i - o]'
  }
  assert (t < UINT_MAX);			// UINT_MAX == garbage

  // Update reason by mapping clause indices of assigned literals on trail.

  for (i = 0; i < trail.size (); i++) {
    Var & v = var (trail[i]);
    if (!v.level) continue;
    Rsn & r = v.reason;
    if (r.binary ()) continue;
    if (r.decision ()) continue;
    assert (r.clausal ());
    unsigned cidx = r.cidx ();
    if (cidx < o) continue;
    r.update (work [cidx - o]);
  }

  // Update clause references on watcher stacks too.

  for (int lit = -maxvar; lit <= maxvar; lit++) {
    if (!lit) continue;
    Wchs & ws = wchs (lit);
    const unsigned t = ws.size ();
    for (i = j = 0; i < t; i++) {
      const int blit = ws[j++] = ws[i];		// by default copy 'blit'
      if (!(blit & 1)) continue;		// ignore binary watches
      unsigned cidx = ws[++i];
      if (cidx >= o &&
	  (cidx = work[cidx - o]) == UINT_MAX) j--; // undo 'blit' copy
      else ws[j++] = cidx;			    // copy 'cidx' too
    }
    ws.resize (j);
  }

  work.clear ();

  // Finally actually collect all garbage clauses.

  unsigned res = 0;		// Number of reduced clauses.

  for (i = j = o; i < s; i++) {
    Cls * c = clauses[i];

    // Removing binary clauses would be very dangerous, since we do not
    // update the binary watchers here.  Removing them might keep stale
    // binary watchers around for removed binary clauses.  A binary clause
    // might become garbage during learned clause subsumption.  Collecting
    // such a clause is post-poned until the next garbage collection, where
    // binary watchers are anyhow flushed and regenerated.

    if (c->garbage && c->size () > 2) {
      assert (c->redundant);
      trace_delete_clause (*c);
      dealloc_clause (c);
      reduced++;
      res++;
    } else clauses[j++] = c;				// actually move
  }

  assert (j == t);
  clauses.resize (t);

  return res;
}

// We experimented with many other schemes, but the original Glucose scheme
// using an arithmetic series seems to work best.  Note however that we
// simplify the condition for scheduling a reduction by using conflict
// intervals instead of monitoring and limiting the number of currently
// existing learned clauses.  See also 'reducing' below.

static void reduce_update_limits (unsigned reduced_this_time) {
  if (!simpmode) inc.conflicts.reduce += opts::reduceinc;
  limits.conflicts.reduce = conflicts + inc.conflicts.reduce;
}

static void reduce () {
  START (reduce);
  assert (!conflict);
  if (simpmode) {
    simpreds++;
    reductions++;
    LOG ("reduce simp " << simpreds);
  } else {
    searchreds++;
    reductions++;
    LOG ("reduce search " << searchreds);
  }
  reduce_mark ();
  unsigned reduced_this_time = reduce_move ();
  reduce_update_limits (reduced_this_time);
  report ('-');
  STOP (reduce);
}

/*------------------------------------------------------------------------*/

static bool reached_limit () {
  return opts::clim >= 0 && conflicts >= opts::clim;
}

static int solve () {
  START (search);
  int res = 0;
  while (!res)
         if (inconsistent) res = 20;
    else if (!propagate ()) analyze ();
    else if (iterating) iteration ();
    else if (all_variables_assigned ()) res = 10;
    else if (arenaing ()) copy_and_move_arena ();
    else if (simplifying ()) simplify ();
    else if (subsuming ()) subsume ();
    else if (reached_limit ()) break;
    else if (restarting ()) restart ();
    else if (reducing ()) reduce ();
    else decide ();
  STOP (search);
  return res;
}

static int limited_assume_solve (int assumption, int64_t limit) {
  assert (!level);
  int res = 0;
  limit += conflicts;
  while (!res)
         if (inconsistent) return res = 20;
    else if (!propagate ()) analyze ();
    else if (val (assumption) < 0) assert (!level), res = 20;
    else if (all_variables_assigned ()) res = 10;
    else if (conflicts > limit) break;
    else if (!val (assumption)) assume_decision (assumption);
    else if (restarting ()) restart ();
    else if (reducing ()) reduce ();
    else decide ();
  if (res != 10 && level) backtrack ();
  return res;
}

static int limited_assume2_solve (int d1, int d2, int64_t limit) {
  assert (!level);
  int res = 0;
  limit += conflicts;
  while (!res)
         if (inconsistent) return res = 20;
    else if (!propagate ()) analyze ();
    else if (val (d1) < 0) assert (!level), res = 20;
    else if (val (d2) < 0) assert (level <= 1), res = 20;
    else if (all_variables_assigned ()) res = 10;
    else if (conflicts > limit) break;
    else if (!val (d1)) assume_decision (d1);
    else if (!val (d2)) assume_decision (d2);
    else if (restarting ()) restart ();
    else if (reducing ()) reduce ();
    else decide ();
  if (res != 10 && level) backtrack ();
  return res;
}

/*------------------------------------------------------------------------*/

static void print_profile (const char * name, double t, double total) {
  char line[80];
  double p = percent (t, total);
  if (!opts::verbose && p + 0.5 < 1.0) return;
  sprintf (line, "c %9.3f %3.0f%% %s", t, p, name);
  std::cout << line << NL;
}

static double print_profile () {
  double total = flush_and_update_all_active_profs ();
  const unsigned size = profs::table.size ();
  std::cout << "c" << NL;

  // Use selection sort here to avoid any memory allocation.  Otherwise
  // calling this function during catching a signal triggered by out of
  // memory, e.g., segmentation fault, might trigger another segmentation
  // fault, which would prevent printing any profile information (exactly
  // when this should be avoid while debugging such a problem).

  for (unsigned i = 0; i < size; i++) {
    Prof * p = profs::table[i];
    for (unsigned j = i + 1; j < size; j++)
      if (p->time < profs::table[j]->time)
	std::swap (p, profs::table[j]);
    print_profile (p->name, p->time, total);
  }
  std::cout << "c ===========================" << NL;
  print_profile ("total", total, total);
  std::cout << "c" << NL << std::flush;
  return total;
}

/*------------------------------------------------------------------------*/

// Input string is supposed to be 'c <A> <B>s', where 'A' is a string
// without space character and '<B>' makes up the rest of the 'line'.  There
// is a trailing single character 's' expected at the end of the 'line' too.
// The space character directly after '<A>' is replace by ':' and the
// trailing 's' by ')'. Further all occurrences of '@' in '<A>' and '<B>'
// are replaced by a space character.  Note, '<B>' might contain spaces.

static void fix_stats_line (char * line) {
  int i;
  assert (line[0] == 'c');
  assert (line[1] == ' ');
  for (i = 2; line[i] != ' '; i++)
    if (line[i] == '@') line[i] = ' ';
  for (line[i++] = ':';line[i]; i++)
    if (line[i] == '@') line[i] = ' ';
  assert (i > 0);
  assert(line[i-1] == 's');
  line[i-1] = ')';
}

// A: average (a, b) (per B)

static void print_relative_stats (const char * A, const char * B,
                                  double a, double b) {
  if (A[0] == '@' && !opts::verbose) return;
  char line[200];
  sprintf (line,
    "c %-25s  %13.0f %11.2f    (per %s",
    A, a, average (a, b), B);
  fix_stats_line (line);
  std::cout << line << NL;
}

// A: average (b, a) (B/C)  where C is a (short) version of A

static void print_reverse_relative_stats (const char * A,
                                          const char * B,
                                          const char * C,
                                          double a, double b) {
  if (A[0] == '@' && !opts::verbose) return;
  char line[200];
  sprintf (line,
    "c %-25s  %13.0f %11.2f    (%s/%s",
    A, a, average (b, a), B, C);
  fix_stats_line (line);
  std::cout << line << NL;
}

// A: percent (a, b) % (per B)

static void
print_percentage_stats (const char * A, const char * B,
                        double a,  double b) {
  if (A[0] == '@' && !opts::verbose) return;
  char line[200];
  sprintf (line,
    "c %-25s  %13.0f %11.2f %%  (per %s",
    A, a, percent (a, b), B);
  fix_stats_line (line);
  std::cout << line << NL;
}

// A: a U

static void
print_resource_stats (const char * A, double a, const char * U) {
  if (A[0] == '@' && !opts::verbose) return;
  char line[200];
  sprintf (line,
    "c %-37s  %13.2f    (%ss",
    A, a, U);
  fix_stats_line (line);
  std::cout << line << NL;
}

static void stats () {
  double seconds = print_profile ();
  print_reverse_relative_stats (
    "simplifications", "conflicts", "simps", simplifications, conflicts);
  print_percentage_stats (
    "@@eliminated-variables", "variables", eliminated, maxvar);
  print_relative_stats (
    "@@elim-resolutions", "elimvars", elimresolutions, eliminated);
  print_percentage_stats (
    "@@bcd-resolutions", "elimress", bcdresolutions, elimresolutions);
  print_percentage_stats (
    "@@blocked-clauses", "variables", blocked, maxvar);
  print_percentage_stats (
    "@@@@moved-clauses", "blockeds", moved, blocked);
  print_relative_stats (
    "@@@@block-resolutions", "blockeds", blockresolutions, blocked);
  print_reverse_relative_stats (
    "@@@@block-vars", "res", "vars", blockvars, blockresolutions);
  print_relative_stats (
    "@@subsume-checks", "subsumeds", subsumechecks, subsumed);
  print_percentage_stats (
    "@@@@subsumed-irredundant", "subsumeds", subsumedirr, subsumed);
  print_percentage_stats (
    "@@@@subsumed-redundant", "subsumeds", subsumedred, subsumed);
  print_percentage_stats (
    "@@@@shrunken", "subsumeds", shrunken, subsumed);
  print_percentage_stats (
    "@@substituted-variables", "variables", substituted, maxvar);
  print_reverse_relative_stats (
    "@@failed-literals", "probes", "faileds", failedlits, failedlitprobes);
  print_reverse_relative_stats (
    "@@lifted-literals", "probes", "lifteds", liftedlits, failedlitprobes);
  print_reverse_relative_stats (
    "@@back-bones", "checks", "founds", backbones, backbonechecks);
  print_reverse_relative_stats (
    "@@implied-literals", "checks", "founds", implies, implieschecks);
  print_reverse_relative_stats (
    "@@equivalent-literals", "checks", "founds", equivs, equivchecks);
  print_reverse_relative_stats (
    "reductions", "conflicts", "reds", reductions, conflicts);
  print_percentage_stats (
    "@@simp-reductions", "reductions", simpreds, reductions);
  print_percentage_stats (
    "@@search-reductions", "reductions", searchreds, reductions);
  print_percentage_stats (
    "@@reduced-clauses", "conflicts", reduced, conflicts);
  print_reverse_relative_stats (
    "restarts", "confs", "restarts", restarts, conflicts);
  print_reverse_relative_stats (
    "@@blocked-restarts", "confs", "blockeds", restartsblocked, conflicts);
  print_percentage_stats (
    "@@skipped", "skip+rsts", restartskipped, restarts + restartskipped);
  print_percentage_stats (
    "@@@@not-forced", "skippeds", restartsnoforce, restartskipped);
  print_percentage_stats (
    "@@@@delayed", "skippeds", restartsdelayed, restartskipped);
  print_percentage_stats (
    "@@@@reused", "skippeds", restartsreused, restartskipped);
  print_percentage_stats (
    "@@reused-trails", "restarts", reusedtrails, restarts);
  print_percentage_stats (
    "@@reused-levels", "levels",
    reusedlevels, reusedlevels + notreusedlevels);
  print_relative_stats (
    "conflicts", "seconds", conflicts, seconds);
  print_percentage_stats (
    "@@simp-conflicts", "conflicts", simpconflicts, conflicts);
  print_percentage_stats (
    "@@search-conflicts", "conflicts",
    conflicts - simpconflicts, conflicts);
  print_relative_stats (
    "@@learned-literals", "conflicts", learnedlits, conflicts);
  print_percentage_stats (
    "@@updated-glues", "conflicts", glueupdates, conflicts);
  print_percentage_stats (
    "@@minimized-literals", "originals",
    minimizedlits, minimizedlits + learnedlits);
  print_percentage_stats (
    "@@on-the-fly-subs", "otfs+confs",
    otfsconflicts + otfsdrives, otfsconflicts + otfsdrives + conflicts);
  print_percentage_stats (
    "@@@@conflicts", "otfss",
    otfsconflicts, otfsconflicts + otfsdrives);
  print_percentage_stats (
    "@@@@drives", "otfss",
    otfsdrives, otfsconflicts + otfsdrives);
  print_percentage_stats (
    "@@subsumed-recent", "conflicts", subrecent, conflicts);
  print_percentage_stats (
    "@@decision-only-clauses", "conflicts", decos, conflicts);
  print_reverse_relative_stats (
    "@@iterations", "conflicts", "its", iterations, conflicts);
  print_percentage_stats (
    "@@additional-units", "iterations", additionalunits, iterations);
  print_reverse_relative_stats (
    "@@learned-binary", "conflicts", "bins", learnedbins, conflicts);
  print_percentage_stats (
    "@@additional-bins", "conflicts", additionalbins, conflicts);
  print_relative_stats (
    "decisions", "conflicts", decisions, conflicts);
  print_relative_stats (
    "@@next-searches", "decisions", searches, decisions);
  print_relative_stats (
    "@@propagations", "decisions", propagations, decisions);
  if (opts::verbose) {
    print_reverse_relative_stats (
      "hyper-bin-res", "propagations", "hbrs", hbrs, propagations);
    print_percentage_stats (
      "@@hbr-subsumed", "hbrs", hbrsubs, hbrs);
    print_reverse_relative_stats (
      "garbage-collections", "conflicts", "gcs", gcs, conflicts);
    print_relative_stats (
      "@@flushed-clauses", "gcs", flushedclauses, gcs);
    print_relative_stats (
      "@@flushed-literals", "gcs", flushedlits, gcs);
    print_relative_stats (
      "time-samples", "seconds", getimes, seconds);
  }
  print_relative_stats (
    "million-propagations", "seconds", propagations/1e6, seconds);
  print_resource_stats (
    "maximally-allocated-memory", maxspace (), "MB");
  print_resource_stats (
    "process-time", seconds, "seconds");
}

/*------------------------------------------------------------------------*/

static bool catchedsig = false;

static void (*sig_int_handler)(int);
static void (*sig_segv_handler)(int);
static void (*sig_abrt_handler)(int);
static void (*sig_term_handler)(int);
static void (*sig_bus_handler)(int);

static void reset_signal_handlers (void) {
  (void) signal (SIGINT, sig_int_handler);
  (void) signal (SIGSEGV, sig_segv_handler);
  (void) signal (SIGABRT, sig_abrt_handler);
  (void) signal (SIGTERM, sig_term_handler);
  (void) signal (SIGBUS, sig_bus_handler);
}

static void caughtsigmsg (int sig) {
  std::cout << "c CAUGHT SIGNAL " << sig << NL << std::flush;
}

static void catchsig (int sig) {
  if (!catchedsig) {
    catchedsig = true;
    caughtsigmsg (sig);
    std::cout  << "c s UNKNOWN" << NL << std::flush;
    stats ();
  }
  reset_signal_handlers ();
  raise (sig);
}

static void init_signal_handlers (void) {
  sig_int_handler = signal (SIGINT, catchsig);
  sig_segv_handler = signal (SIGSEGV, catchsig);
  sig_abrt_handler = signal (SIGABRT, catchsig);
  sig_term_handler = signal (SIGTERM, catchsig);
  sig_bus_handler = signal (SIGBUS, catchsig);
}

/*------------------------------------------------------------------------*/

static void print_witness () {
  // TODO nicer format, e.g., always fill 75 columns or so ...
  for (int idx = 1; idx <= maxvar; idx++) {
    if ((idx % 8) == 1) {
      if (idx > 1) std::cout << NL;
      std::cout << "v";
    }
    std::cout << ' ' << vals[idx] * idx;
  }
  if (maxvar > 0) std::cout << NL;
  std::cout << "v 0" << NL << std::flush;
}

/*------------------------------------------------------------------------*/

int main (int argc, char ** argv) {

  // Handle options which print something and exit before banner is printed.

  const char * path = 0;
  bool checking = false; // Actually only enabled '#ifdef EXPENSIVECHECKS'.
  bool logging = false;  // Actually only enabled '#ifdef LOGGING'.
  bool witness = true;
  int verbose = 0;

  const char * trace = 0;

  for (int i = 1; i < argc; i++) {
         if (!strcmp (argv[i], "-h")) { usage (); exit (0); }
    else if (!strcmp (argv[i], "-e")) { embedded_opts (); exit (0); }
    else if (!strcmp (argv[i], "-r")) { range_opts (); exit (0); }
    else if (!strcmp (argv[i], "--version")) {
      std::cout << VERSION << NL; exit (0);
    } else if (!strcmp (argv[i], "-n")) witness = false;
    else if (!strcmp (argv[i], "-v")) verbose++;
#ifdef LOGGING
    else if (!strcmp (argv[i], "-l")) logging = true;
#endif
#ifdef EXPENSIVECHECKS
    else if (!strcmp (argv[i], "-c")) checking = true;
#endif
    else if (!strcmp (argv[i], "-t")) {
      if (++i == argc)
	DIE ("argument to '-t' missing");
      if (trace)
	DIE ("multiple trace options '-t " << trace <<
	  "' and '-t " << argv[i] << "'");
      trace = argv[i];
    } else if (!strcmp (argv[i], "-o")) {
      if (++i == argc)
	DIE ("argument to '-o' missing");
      if (output)
	DIE ("multiple output options '-o " << output <<
	  "' and '-t " << argv[i] << "'");
      output = argv[i];
    } else if (hasopt (argv[i])) ;		// long option see 'opts.h'
    else if (*argv[i] == '-')
      DIE ("invalid option '" << argv[i] << "' (try '-h')");
    else if (path)
      DIE ("multiple input files '" <<
        path << "' and '" << argv[i] << "'");
    else if (!file_exists (path = argv[i]))
      DIE ("file '" << path << "' does not exist");
  }

  // Now print banner.

  std::cout
  << "c Splatz SAT Solver" << NL
  << "c Version " << VERSION << NL
  << "c Copyright (C) 2015 - 2016 Armin Biere" << NL
  << "c Johannes Keppler University Linz" << NL
  << "c " << COMPILE << NL << std::flush;

  // Global initialization before options are either set through command
  // line option parsing or during parsing the header of a DIMACS file.

  init_global ();
#ifdef LOGGING
  if (logging) opts::verbose = MAXVERBOSE;
#endif
  if (verbose) setopt ("verbose", verbose);
  if (checking) setopt ("check", checking);
  if (logging) setopt ("log", logging);

  // Now parse again the long options among the command line options which
  // actually set real 'opts::...' options, which then is printed.
  // Note that long options for 'verbose', 'check' and 'log' take
  // precedence over short options.

  for (int i = 1; i < argc; i++)
    if (argv[i][0] == '-' && argv[i][1] == '-')
      parseopt (argv[i]);

  if (trace) {
    if (!opts::trace) setopt ("trace", 1);
    std::cout << "c tracing to '" << trace << "'" << NL;
    tout = new std::ofstream (trace, std::ofstream::out);
    delete_and_close_tout = true;
  }
  
  if (output)
    std::cout << "c output '" << output << "'" << NL;

  if (opts::verbose)
    std::cout
    << "c sizeof (Cls) == " << sizeof (Cls) << NL
    << "c sizeof (Red) == " << sizeof (Red) << NL
    << "c sizeof (Var) == " << sizeof (Var) << NL
    << "c sizeof (Lnk) == " << sizeof (Lnk) << NL
    << "c sizeof (Frame) == " << sizeof (Frame) << NL
    << std::flush;

  init_signal_handlers ();
  profs::init ();		// function profilig setup

  parse (path);			// might set options
  				// will also initialize variables

  init_solve ();		// solving specific parameter setup

  int res = solve ();

  // Report result.

  if (opts::trace) std::cout << std::flush;
  if (res == 10) {
    report ('1');
    std::cout << 'c' << NL;
    std::cout << "s SATISFIABLE" << NL << std::flush;
    extend_model ();
    if (witness) print_witness ();
  } else if (res == 20) {
    report ('0');
    std::cout << 'c' << NL;
    std::cout << "s UNSATISFIABLE" << NL << std::flush;
  } else {
    report ('?');
    std::cout << 'c' << NL;
    std::cout << "c UNKNOWN" << NL << std::flush;
  }

#ifndef NDEBUG
  if (res == 10) check_model ();
#endif

  reset ();			// and checks for leaks ...
  reset_signal_handlers ();

  stats ();			// no heap etc needed

  return res;
}
